#include "config.h"
#include "fupermod_cblas.h"
#include <stdio.h>
#include <string.h>

#ifdef HAVE_LIBCUBLAS
#include "cuda_auxiliary.hpp"
#include "cutil.h"
#include "cutil_inline_runtime.h"

/*! memory transfer every time, use all devices  */
void cuda_gemm_basic(fupermod_gemm* gemm,
          const enum CBLAS_ORDER Order, const enum CBLAS_TRANSPOSE TransA, const enum CBLAS_TRANSPOSE TransB,
          const int M, const int N, const int K, const fupermod_float alpha, const fupermod_float *A, const int lda,
          const fupermod_float *B, const int ldb, const fupermod_float beta, fupermod_float *C, const int ldc)
{
  cublasxgemm_params* params = (cublasxgemm_params*)gemm->params;
  int nd, w = N / K;
  if (w == 1) {
    params->num_device_use = nd = 1;
  } else {
    params->num_device_use = nd = params->num_devices;
  }
  int width[nd];
  int j;
  for (j = 0; j < nd; j++) {
    if (w % nd == 0) {
      width[j] = w / nd;
    } else {
      width[j] = (j < w % nd) ? (w / nd + 1) : (w / nd);
    }
  }
  
  // create cublas handles
  params->handles = (cublasHandle_t*)malloc(sizeof(cublasHandle_t) * nd);
  int device_id = params->device_id;
  for (j = 0; j < nd; j++) {
    CUDA_SAFE_CALL(cudaSetDevice(device_id++));
    CUBLAS_SAFE_CALL(cublasCreate((cublasHandle_t*)&params->handles[j]));
  }
  // memory allocation
  params->dA = (fupermod_float**)malloc(sizeof(fupermod_float*) * nd);
  params->dB = (fupermod_float**)malloc(sizeof(fupermod_float*) * nd);
  params->dC = (fupermod_float**)malloc(sizeof(fupermod_float*) * nd);

  device_id = params->device_id;
  for (j = 0; j < nd; j++) {
    // set device
    CUDA_SAFE_CALL(cudaSetDevice(device_id++));
    // alloc memory
    CUDA_SAFE_CALL(cudaMalloc((void**)(&params->dA[j]), sizeof(fupermod_float) * M * K));
    CUDA_SAFE_CALL(cudaMalloc((void**)(&params->dB[j]), sizeof(fupermod_float) * K * width[j] * K));
    CUDA_SAFE_CALL(cudaMalloc((void**)(&params->dC[j]), sizeof(fupermod_float) * M * width[j] * K));
//  CUDA_SAFE_CALL(cudaMemset((void*)params->dC[j],  0, sizeof(fupermod_float) * M * width[j] * K));
  }
  // C = alpha*AxB + beta*C
  if (beta > 1e-4) {
    device_id = params->device_id;
    for (j = 0; j < nd; j++) {
      // set device
      CUDA_SAFE_CALL(cudaSetDevice(device_id++));
      // synchronous call, control is not returned to host before the device has completed the requested task
      CUDA_SAFE_CALL(cudaMemcpy2D(params->dC[j], sizeof(fupermod_float) * width[j] * K, addr(C,0,j,NULL,width,N,K,3), sizeof(fupermod_float) * ldc,
                sizeof(fupermod_float) * width[j] * K, M, cudaMemcpyHostToDevice));
    }
  }

  // send A and B[j], and do gemm, and copy results back
  device_id = params->device_id;
  for (j = 0; j < nd; j++) {
    // set device
    CUDA_SAFE_CALL(cudaSetDevice(device_id++));
    // data transfer
    CUDA_SAFE_CALL(cudaMemcpy2D(params->dA[j], sizeof(fupermod_float) * K, A, sizeof(fupermod_float) * lda, sizeof(fupermod_float) * K, M, cudaMemcpyHostToDevice));
    CUDA_SAFE_CALL(cudaMemcpy2D(params->dB[j], sizeof(fupermod_float) * width[j] * K, addr(B,0,j,NULL,width,N,K,2), sizeof(fupermod_float) * ldb,
              sizeof(fupermod_float) * width[j] * K, K, cudaMemcpyHostToDevice));
           
    // Note: cublas is column-major, so have to swap matrices A and B
    CUBLAS_SAFE_CALL(CUBLAS_GEMM(params->handles[j], trans_cblas2cublas(TransB), trans_cblas2cublas(TransA), 
                  width[j] * K, M, K, &alpha, params->dB[j], width[j] * K, params->dA[j], K, &beta, params->dC[j], width[j] * K));
                                    
    // synchronous call, control is not returned to host before the device has completed the requested task
    CUDA_SAFE_CALL(cudaMemcpy2D(addr(C,0,j,NULL,width,N,K,3), sizeof(fupermod_float) * ldc, params->dC[j], sizeof(fupermod_float) * width[j] * K,
              sizeof(fupermod_float) * width[j] * K, M, cudaMemcpyDeviceToHost));
  }
  
  // cleanup
  device_id = params->device_id;
  for (j = 0; j < nd; j++) {
    // set device
    CUDA_SAFE_CALL(cudaSetDevice(device_id++));
    // free device memory
    CUDA_SAFE_CALL(cudaFree(params->dA[j])); params->dA[j] = NULL;
    CUDA_SAFE_CALL(cudaFree(params->dB[j])); params->dB[j] = NULL;
    CUDA_SAFE_CALL(cudaFree(params->dC[j])); params->dC[j] = NULL;
    // destroy handle
    CUBLAS_SAFE_CALL(cublasDestroy(params->handles[j])); params->handles[j] = NULL;
  }
  free(params->dA); params->dA = NULL;
  free(params->dB); params->dB = NULL;
  free(params->dC); params->dC = NULL;
  free(params->handles); params->handles = NULL;
}

// non-basic algorithm
void cuda_gemm(fupermod_gemm* gemm,
          const enum CBLAS_ORDER Order, const enum CBLAS_TRANSPOSE TransA, const enum CBLAS_TRANSPOSE TransB,
          const int M, const int N, const int K, const fupermod_float alpha, const fupermod_float *A, const int lda,
          const fupermod_float *B, const int ldb, const fupermod_float beta, fupermod_float *C, const int ldc)
{
  cublasxgemm_params* params = (cublasxgemm_params*)gemm->params;
  if (params->num_devices == 1) {
    // decide kernel type at the beginning
    if (params->idx_outer == 0) {
      size_t mem_size_device = gama * getDeviceMemory(params->device_id)/sizeof(fupermod_float);
      if ((M * N + M * K + K * N) > mem_size_device) {
        params->kernel_type = 1; // exceed one device
      } else {
        params->kernel_type = 0; // fit in one device
      }
      // number of devices to use for this problem size
      params->num_device_use = 1;
      // create cublas handles, according to the number of devices to use
      params->handles = (cublasHandle_t*)malloc(sizeof(cublasHandle_t));
      CUDA_SAFE_CALL(cudaSetDevice(params->device_id));
      CUBLAS_SAFE_CALL(cublasCreate((cublasHandle_t*)&params->handles[0]));
    }
    switch (params->kernel_type) {
      case 0:
        gemm_kernel_gpu_single_fit(gemm, Order, TransA, TransB, M, N, K, alpha, A, lda, B, ldb, beta, C, ldc);
        break;
      case 1:
        gemm_kernel_gpu_single_exceed(gemm, Order, TransA, TransB, M, N, K, alpha, A, lda, B, ldb, beta, C, ldc);
        break;
      default:
        fprintf(stderr, "%s: %s: Error gpu_single kernel\n", __FILE__, __func__);
        break; 
    } 
  } else {
    int nd = params->num_devices;
    // decide kernel type at the beginning
    if (params->idx_outer == 0) {
      size_t mem_size_device = gama * getDeviceMemory(params->device_id)/sizeof(fupermod_float);
      if ((M * N + M * K * nd + K * N) > nd * mem_size_device) { // 1D partitioning for multiple gpus
        params->kernel_type = 1; // exceed gpu server
      } else {
        params->kernel_type = 0; // fit in gpu server
      }
      params->num_device_use = nd;
      // create cublas handles, according to the number of devices to use
      params->handles = (cublasHandle_t*)malloc(sizeof(cublasHandle_t) * params->num_device_use);
      int i, device_id = params->device_id;
      for (i = 0; i < params->num_device_use; i++) {
        CUDA_SAFE_CALL(cudaSetDevice(device_id++));
        CUBLAS_SAFE_CALL(cublasCreate((cublasHandle_t*)&params->handles[i]));
      }
    }
    switch (params->kernel_type) {
      case 0:
        gemm_kernel_gpu_server_fit(gemm, Order, TransA, TransB, M, N, K, alpha, A, lda, B, ldb, beta, C, ldc);
        break;
      case 1:
        gemm_kernel_gpu_server_exceed(gemm, Order, TransA, TransB, M, N, K, alpha, A, lda, B, ldb, beta, C, ldc);
        break;
      default:
        fprintf(stderr, "%s: %s: Error gpu_server kernel\n", __FILE__, __func__);
        break;
    }
  }  
}

void gemm_kernel_gpu_single_exceed(fupermod_gemm* gemm,
          const enum CBLAS_ORDER Order, const enum CBLAS_TRANSPOSE TransA, const enum CBLAS_TRANSPOSE TransB,
          const int M, const int N, const int K, const fupermod_float alpha, const fupermod_float *A, const int lda,
          const fupermod_float *B, const int ldb, const fupermod_float beta, fupermod_float *C, const int ldc)
{
  cublasxgemm_params* params = (cublasxgemm_params*)gemm->params;
  switch (params->algorithm) {
    case 0: 
      gemm_kernel_gpu_single_exceed_0(gemm, Order, TransA, TransB, M, N, K, alpha, A, lda, B, ldb, beta, C, ldc);
      break;
    case 1:
      gemm_kernel_gpu_single_exceed_1(gemm, Order, TransA, TransB, M, N, K, alpha, A, lda, B, ldb, beta, C, ldc);
      break;    
    case 2:
      gemm_kernel_gpu_single_exceed_2(gemm, Order, TransA, TransB, M, N, K, alpha, A, lda, B, ldb, beta, C, ldc);
      break;
    case 3:
      gemm_kernel_gpu_single_exceed_3(gemm, Order, TransA, TransB, M, N, K, alpha, A, lda, B, ldb, beta, C, ldc);
      break;
 /*   case 4:
      gemm_kernel_gpu_single_exceed_4(gemm, Order, TransA, TransB, M, N, K, alpha, A, lda, B, ldb, beta, C, ldc);
      break;*/
    default:
      fprintf(stderr, "%s: %s: Error gpu_single_exceed kernel\n", __FILE__, __func__);
      break;      
  }
}

void gemm_kernel_gpu_server_exceed(fupermod_gemm* gemm,
          const enum CBLAS_ORDER Order, const enum CBLAS_TRANSPOSE TransA, const enum CBLAS_TRANSPOSE TransB,
          const int M, const int N, const int K, const fupermod_float alpha, const fupermod_float *A, const int lda,
          const fupermod_float *B, const int ldb, const fupermod_float beta, fupermod_float *C, const int ldc)
{
  cublasxgemm_params* params = (cublasxgemm_params*)gemm->params;
  switch (params->algorithm) {
    case 0: 
      gemm_kernel_gpu_server_exceed_nd2_1(gemm, Order, TransA, TransB, M, N, K, alpha, A, lda, B, ldb, beta, C, ldc);
      break;
    case 1:
      gemm_kernel_gpu_server_exceed_np2_1(gemm, Order, TransA, TransB, M, N, K, alpha, A, lda, B, ldb, beta, C, ldc);
      break;
    case 2:
      gemm_kernel_gpu_server_exceed_nd2_2(gemm, Order, TransA, TransB, M, N, K, alpha, A, lda, B, ldb, beta, C, ldc);
      break;
    case 3:
      gemm_kernel_gpu_server_exceed_np2_2(gemm, Order, TransA, TransB, M, N, K, alpha, A, lda, B, ldb, beta, C, ldc);
      break;
    case 4:
      gemm_kernel_gpu_server_exceed_nd2_3(gemm, Order, TransA, TransB, M, N, K, alpha, A, lda, B, ldb, beta, C, ldc);
      break;
    default:
      fprintf(stderr, "%s: %s: Error gpu_server_exceed kernel\n", __FILE__, __func__);
      break;      
  }
}

// split matrix C into smaller blocks
// there should be some relation between value M, N, K and the time for data transfer and gemm excution time
// ideally can use an experimental funtion, to determin the m1, and n1, so that the data transfer could be covered 
void matrix_split_single_fit(int M, int N, int K, int *h_, int *w_, int **height, int **width)
{
  int h = M / K;
  int w = N / K;

  // 10 is used as default values
  // note: what if h, or n is smaller than 10? 
  *h_ = *w_ = 10;
  *height = (int*)malloc(sizeof(int) * (*h_));
  *width  = (int*)malloc(sizeof(int) * (*w_));
  
  int i;
  for (i = 0; i < *h_; i++) {
    if (h % *h_ == 0) {
      (*height)[i] = h / *h_;
    } else {
      if (i < h % *h_) {
        (*height)[i] = h / *h_ + 1;
      } else {
        (*height)[i] = h / *h_;
      }
    }
  }
  for (i = 0; i < *w_; i++) {
    if (w % *w_ == 0) {
      (*width)[i] = w / *w_;
    } else {
      if (i < w % *w_) {
        (*width)[i] = w / *w_ + 1;
      } else {
        (*width)[i] = w / *w_;
      }
    }
  }
}

void matrix_split_single_exceed(int M, int N, int K, int mem_size, int *h_, int *w_, int **height, int **width)
{
  size_t unit = K * K;
  size_t g = (mem_size / unit);
  size_t h = M / K;
  size_t w = N / K;
  int _n1, _m1, n1;
  
  if ((2 + 3 * w) > g) {
    _n1 = (g - 2) / 3;
    *w_ = ceil((double)w / _n1); // width
  } else {
    *w_ = 1;
  }
  n1 = (w % *w_ == 0) ? (w / *w_) : (w / *w_ + 1);
  _m1 = 0.5 * ((g - n1) / (n1 + 1));
  *h_ = ceil((double)h / _m1); // height

  *height = (int*)malloc(sizeof(int) * (*h_));
  *width  = (int*)malloc(sizeof(int) * (*w_));

  int i;
  for (i = 0; i < *h_; i++) {
    if (h % *h_ == 0) {
      (*height)[i] = h / *h_;
    } else {
      if (i < h % *h_) {
        (*height)[i] = h / *h_ + 1;
      } else {
        (*height)[i] = h / *h_;
      }
    }
  }
  for (i = 0; i < *w_; i++) {
    if (w % *w_ == 0) {
      (*width)[i] = w / *w_;
    } else {
      if (i < w % *w_) {
        (*width)[i] = w / *w_ + 1;
      } else {
        (*width)[i] = w / *w_;
      }
    }
  }
}

void matrix_split_server_exceed(int M, int N, int K, int mem_size_device, int num_device_use, int *h_, int *w_, int **height, int **width)
{
  size_t unit = K * K;
  size_t g = (mem_size_device / unit);
  size_t h = M / K;
  size_t w = N / K;
  int nd = num_device_use;
  int _n1, _m1, n1;
  
  if ((nd * 2 + 3 * w) > nd * g) {
    _n1 = (g - 2) / 3;
    *w_ = ceil((double)w / _n1);
    if (*w_ % nd != 0) {
      *w_ = (*w_ / nd + 1) * nd;
    }
  } else {
    *w_ = nd;
  }
  n1 = (w % *w_ == 0) ? (w / *w_) : (w / *w_ + 1);
  _m1 = 0.5 * (g - n1) / (n1 + 1);
  *h_ = ceil((double)h / _m1); // height
  
  *height = (int*)malloc(sizeof(int) * (*h_));
  *width  = (int*)malloc(sizeof(int) * (*w_));

  int i;
  for (i = 0; i < *h_; i++) {
    if (h % *h_ == 0) {
      (*height)[i] = h / *h_;
    } else {
      if (i < h % *h_) {
        (*height)[i] = h / *h_ + 1;
      } else {
        (*height)[i] = h / *h_;
      }
    }
  }
  for (i = 0; i < *w_; i++) {
    if (w % *w_ == 0) {
      (*width)[i] = w / *w_;
    } else {
      if (i < w % *w_) {
        (*width)[i] = w / *w_ + 1;
      } else {
        (*width)[i] = w / *w_;
      }
    }
  }
}
#endif
