#include "config.h"
#include "fupermod_cblas.h"
#include "fupermod_memory.h"

#ifdef HAVE_LIBCUBLAS
#include "cuda_auxiliary.hpp"
#include "cutil.h"
#include "cutil_inline_runtime.h"
#include "cuda_profiler_api.h"
#endif

#ifdef HAVE_LIBMKL_CORE
#include <mkl_blas.h>
#include <mkl_service.h>
#endif

#if defined HAVE_LIBACML || defined HAVE_LIBACML_MP
#include <omp.h>
#include <acml.h>
#endif

#ifdef HAVE_LIBESSLBG
#include <essl.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

extern "C"
fupermod_gemm* fupermod_gemm_alloc(fupermod_process_conf* conf) {
  fupermod_gemm* gemm = NULL;
  if (conf) {
    gemm = (fupermod_gemm*)malloc(sizeof(fupermod_gemm));
    gemm->conf = conf;
    gemm->time = 0;
    // process binding, also suitable for a process with multiple threads
    if (strcmp(conf->bind, "na")) {
      fupermod_bind_process(conf->bind);
    }
    // parse suboptions
    char* subopts = strdup(conf->subopts);
    char* subopts_0 = subopts;
    if (!strcmp(conf->device_type, "cpu")) {
      char* tokens[] = {"t", NULL};
      char* value;
      while (*subopts != '\0') {
        switch (getsubopt(&subopts, tokens, &value)) {
          case 0:
            int threads = atoi(value);
            #ifdef HAVE_LIBMKL_CORE
            mkl_set_num_threads(threads);
            #elif defined HAVE_LIBACML_MP
            if (strcmp(conf->bind, "na")){
              omp_set_num_threads(threads);
            }
            #endif
            break;
        }
      }
      gemm->params = NULL;
    } else {
      if (!strcmp(conf->device_type, "gpu")) {
        /*
          id is the initial device id of the gpu server
          nd is the number of devices
          md is the memory mode when allocating device memory
          al is the cuda_gemm algorithm
          ol is a flag indicating if use overlapping or not
        */
        #ifdef HAVE_LIBCUBLAS
        cublasxgemm_params* params = (cublasxgemm_params*)malloc(sizeof(cublasxgemm_params));
        // default settings
        params->device_id   = 0;
        params->num_devices = 1;
        params->memory_mode = 0;
        params->algorithm   =-1;
        params->is_overlap  = 1;
        // get suboptions
        char* tokens[] = {"id", "nd", "md", "al", "ol", NULL};
        char* value;
        while (*subopts != '\0') {
          switch (getsubopt(&subopts, tokens, &value)) {
            case 0:
              params->device_id = atoi(value);
              break;
            case 1:
              params->num_devices = atoi(value);;
              break;
            case 2:
              params->memory_mode = atoi(value);
              break;
            case 3:
              params->algorithm = atoi(value);
              break;
            case 4:
              params->is_overlap = atoi(value);
              break;
          }
        }
        if (params->algorithm > -1) {
          params->idx_outer = 0;
          // Note: when building fpms, the device memory etc is destroyed in fupermod_gemm_free other than the cuda_gemm functions
          // when running routines, device memory and handles are destroyed in cuda_gemm functions
          params->num_iteration = 0;
        }
        //params->dma_WB = NULL;
        gemm->params = params;
        #else
        fprintf(stderr, "%s: %s: Error no cuda library\n", __FILE__, __func__);
        #endif
      }
    }
    free(subopts_0);
  }
  return gemm;
}

char CBLAS_TRANSPOSE_CHAR[] = {'N', 'T', 'C'};
char* cblas_transpose(CBLAS_TRANSPOSE Trans) {
  switch(Trans) {
    case 111: return &CBLAS_TRANSPOSE_CHAR[0];
    case 112: return &CBLAS_TRANSPOSE_CHAR[1];
    case 113: return &CBLAS_TRANSPOSE_CHAR[2];
  }
  return NULL;
}

extern	"C"
void fupermod_gemm_execute(struct fupermod_gemm* gemm,
		const enum CBLAS_ORDER Order, const enum CBLAS_TRANSPOSE TransA, const enum CBLAS_TRANSPOSE TransB,
		const int M, const int N, const int K,
		const fupermod_float alpha, const fupermod_float *A, const int lda,
		const fupermod_float *B, const int ldb,  const fupermod_float beta,
		fupermod_float *C, const int ldc) 
{
  if (!gemm || !strcmp(gemm->conf->device_type, "cpu")) { // !gemm: in case conf = NULL and then gemm = NULL
    #ifdef ENABLE_BLAS_SP
      #ifdef WITH_BLAS_GSL
      cblas_sgemm(Order, TransA, TransB, M, N, K, alpha, A, lda, B, ldb, beta, C, ldc);
      #elif defined HAVE_LIBGOTO2
      cblas_sgemm(Order, TransA, TransB, M, N, K, alpha, A, lda, B, ldb, beta, C, ldc);
      #elif defined HAVE_LIBMKL_CORE
      sgemm(cblas_transpose(TransB), cblas_transpose(TransA), &N, &M, &K, &alpha, B, &ldb, A, &lda, &beta, C, &ldc);
      #elif defined HAVE_LIBACML || defined HAVE_LIBACML_MP
      //note: BLAS is colum-major, so have to swap matrices A and B
      sgemm(*cblas_transpose(TransB), *cblas_transpose(TransA), N, M, K, alpha, (fupermod_float *)B, ldb, (fupermod_float *)A, lda, beta, C, ldc);
      #elif defined HAVE_LIBESSLBG
      //note: BLAS is colum-major, so have to swap matrices A and B
      sgemm((char*)TransB, (char*)TransA, N, M, K, alpha, B, ldb, A, lda, beta, C, ldc);
      #else
      sgemm_((char)TransB, (char)TransA, N, M, K, alpha, B, ldb, A, lda, beta, C, ldc);
      #endif
    #else
      #ifdef WITH_BLAS_GSL
      cblas_dgemm(Order, TransA, TransB, M, N, K, alpha, A, lda, B, ldb, beta, C, ldc);
      #elif defined HAVE_LIBGOTO2
      cblas_dgemm(Order, TransA, TransB, M, N, K, alpha, A, lda, B, ldb, beta, C, ldc);
      #elif defined HAVE_LIBMKL_CORE
      dgemm(cblas_transpose(TransB), cblas_transpose(TransA), &N, &M, &K, &alpha, B, &ldb, A, &lda, &beta, C, &ldc);
      #elif defined HAVE_LIBACML || defined HAVE_LIBACML_MP
      //note: BLAS is colum-major, so have to swap matrices A and B
      dgemm(*cblas_transpose(TransB), *cblas_transpose(TransA), N, M, K, alpha, (fupermod_float *)B, ldb, (fupermod_float *)A, lda, beta, C, ldc);
	    #elif defined HAVE_LIBESSLBG
      //note: BLAS is colum-major, so have to swap matrices A and B
      dgemm((char*)TransB, (char*)TransA, N, M, K, alpha, B, ldb, A, lda, beta, C, ldc);
      #else
      dgemm_((char)TransB, (char)TransA, N, M, K, alpha, B, ldb, A, lda, beta, C, ldc);
      #endif
    #endif
  } else {
    if (!strcmp(gemm->conf->device_type, "gpu")) {
      #ifdef HAVE_LIBCUBLAS
      switch(((cublasxgemm_params*)gemm->params)->algorithm) {
        case -1:
          cuda_gemm_basic(gemm, Order, TransA, TransB, M, N, K, alpha, A, lda, B, ldb, beta, C, ldc);
        break;
        default:
          cudaProfilerStart();
          cuda_gemm(gemm, Order, TransA, TransB, M, N, K, alpha, A, lda, B, ldb, beta, C, ldc);
          cudaProfilerStop();
        break;
      }
      #else
      fprintf(stderr, "%s: %s: Error no cuda library\n", __FILE__, __func__);
      #endif
    }
  }
}

extern "C"
void fupermod_gemm_free(fupermod_gemm* gemm) {
  if (gemm) {
    if (gemm->params) {
      if (!strcmp(gemm->conf->device_type, "gpu")) {
        #ifdef HAVE_LIBCUBLAS
        cublasxgemm_params* params = (cublasxgemm_params*)gemm->params;
        int j, device_id, nd = params->num_device_use;
        // free device memory (when building fpms memory is freed here)
        // fprintf(stderr, "* free device memory etc. in fupermod_gemm_free (when running builder)\n");
        if (params->dA) {
          device_id = params->device_id;
          for (j = 0; j < nd; j++) {
            CUDA_SAFE_CALL(cudaSetDevice(device_id++));
            if (params->dA[j])
              CUDA_SAFE_CALL(cudaFree(params->dA[j]));
          }
          free(params->dA);
        }
        if (params->dB) {
          device_id = params->device_id;
          for (j = 0; j < nd; j++) {
            CUDA_SAFE_CALL(cudaSetDevice(device_id++));
            if (params->dB[j])
              CUDA_SAFE_CALL(cudaFree(params->dB[j]));
          }
          free(params->dB);
        }
        if (params->dC) {
          device_id = params->device_id;
          for (j = 0; j < nd; j++) {
            CUDA_SAFE_CALL(cudaSetDevice(device_id++));
            if (params->dC[j])
              CUDA_SAFE_CALL(cudaFree(params->dC[j]));
          }
          free(params->dC);
        }
/*        if (params->dma_WB) {
          fupermod_free(params->dma_WB, gemm->conf);
        }*/
        // destroy cublas handles
        if (params->handles) {
          device_id = params->device_id;
          for (j = 0; j < nd; j++) {
            CUDA_SAFE_CALL(cudaSetDevice(device_id++));
            if (params->handles[j])
              CUBLAS_SAFE_CALL(cublasDestroy(params->handles[j]));
          }
          free(params->handles);
        }
        // device reset
        device_id = params->device_id;
        for (j = 0; j < nd; j++) {
          CUDA_SAFE_CALL(cudaSetDevice(device_id++));
          CUDA_SAFE_CALL(cudaDeviceReset());
        }
        #else
        fprintf(stderr, "%s: %s: Error no cuda library\n", __FILE__, __func__);
        #endif
      }
      free(gemm->params);
    }
    free(gemm);
  }
}
