#ifndef FUPERMOD_MEMORY_H_
#define FUPERMOD_MEMORY_H_

#include <stddef.h>
#include "config.h"
#include "fupermod/fupermod_conf.h"

/*!
 * \defgroup fupermod_memory Wrappers for memory management
 * API to handle data that can be allocated either in main or in GPU memory:
 * for example, blocks of matrices in hybrid (CPU/GPU) matrix multiplication.
 * \{
 */
#ifdef __cplusplus
extern "C" {
#endif

  /*!
   * Memory allocation modes, including C standard malloc and CUDA memory allocation.
   * CUDA flags can be combined with "|"
   */
  typedef enum fupermod_mem_mode {
    /*! C standard malloc */
    FUPERMOD_MEM_DEFAULT = -0x01,
    /*! Cuda memory allocation, allocate host memory as page-locked */
    FUPERMOD_MEM_PAGELOCKED = 0x00,
    /*! Cuda memory allocation, allocate host memory as page-locked by all CUDA contexts, not just the one that performed the allocation */
    FUPERMOD_MEM_PORTABLE = 0x01,
    /*! Cuda memory allocation, maps the allocation into the CUDA address space*/
    FUPERMOD_MEM_MAPPED = 0x02,
    /*! Cuda memory allocation, allocate host memory as write-combined, WC memory can be transferred quickly on PCI but cannot be read efficiently by most CPU */
    FUPERMOD_MEM_WRITECOMBINED = 0x04

  } fupermod_mem_mode;

  /*!
   * Allocate a block of size bytes of memory. If NULL is returned, then there is something wrong
   * \param size size of memory to be allocated
   * \param conf process configuration
   * \return a pointer to the beginning of the memory block
   */
  void* fupermod_malloc(size_t size, fupermod_process_conf* conf);

  /*! 
   * Free memory
   * \param ptr the beginning of the memory block to be freed
   * \param conf process configuration
   */
  void fupermod_free(void* ptr, fupermod_process_conf* conf);

#ifdef __cplusplus
}
#endif
/*!
 * \}
 */

#endif /* FUPERMOD_MEMORY_H_ */
