#include "config.h"
#include "fupermod_cblas.h"
#include <stdio.h>
#include <string.h>

#ifdef HAVE_LIBCUBLAS
#include "cuda_auxiliary.hpp"
#include "cutil.h"
#include "cutil_inline_runtime.h"

// synchronous data transfers, accumulate intermediate results in devices
void gemm_kernel_gpu_server_fit(fupermod_gemm* gemm,
          const enum CBLAS_ORDER Order,
          const enum CBLAS_TRANSPOSE TransA,
          const enum CBLAS_TRANSPOSE TransB,
          const int M, const int N, const int K,
          const fupermod_float alpha,
          const fupermod_float *A, const int lda,
          const fupermod_float *B, const int ldb,
          const fupermod_float beta,
          fupermod_float *C, const int ldc)
{
  cublasxgemm_params* params = (cublasxgemm_params*)gemm->params;

  int nd = params->num_device_use;
  int w = N / K; // if w = 1?
  int width[nd];
  int j;
  for (j = 0; j < nd; j++) {
    if (w % nd == 0) {
      width[j] = w / nd;
    } else {
      width[j] = (j < w % nd) ? (w / nd + 1) : (w / nd);
    }
  }
  
  // initialization
  int device_id = params->device_id;
  if (params->idx_outer == 0) {
    fprintf(stdout, "Kernel: gemm_kernel_gpu_server_fit\n");
    params->dA = (fupermod_float**)malloc(sizeof(fupermod_float*) * nd);
    params->dB = (fupermod_float**)malloc(sizeof(fupermod_float*) * nd);
    params->dC = (fupermod_float**)malloc(sizeof(fupermod_float*) * nd);
    
    for (j = 0; j < nd; j++) {
      // set device
      CUDA_SAFE_CALL(cudaSetDevice(device_id++));
      // alloc memory
      CUDA_SAFE_CALL(cudaMalloc((void**)(&params->dA[j]), sizeof(fupermod_float) * M * K));
      CUDA_SAFE_CALL(cudaMalloc((void**)(&params->dB[j]), sizeof(fupermod_float) * K * width[j] * K));
      CUDA_SAFE_CALL(cudaMalloc((void**)(&params->dC[j]), sizeof(fupermod_float) * M * width[j] * K));
//    CUDA_SAFE_CALL(cudaMemset((void*)params->dC[j],  0, sizeof(fupermod_float) * M * width[j] * K));
    }
    // C = alpha*AxB + beta*C
    if (beta > 1e-4) {
      device_id = params->device_id;
      for (j = 0; j < nd; j++) {
        // set device
        CUDA_SAFE_CALL(cudaSetDevice(device_id++));
        // synchronous call, control is not returned to host before the device has completed the requested task
        CUDA_SAFE_CALL(cudaMemcpy2D(params->dC[j], sizeof(fupermod_float) * width[j] * K, addr(C,0,j,NULL,width,N,K,3), sizeof(fupermod_float) * ldc,
                                                   sizeof(fupermod_float) * width[j] * K, M, cudaMemcpyHostToDevice));
      }
    }
  }

  // send A and B[j], and do gemm
  device_id = params->device_id;
  for (j = 0; j < nd; j++) {
    // set device
    CUDA_SAFE_CALL(cudaSetDevice(device_id++));
    // data transfer
    CUDA_SAFE_CALL(cudaMemcpy2D(params->dA[j], sizeof(fupermod_float) * K, A, sizeof(fupermod_float) * lda, sizeof(fupermod_float) * K, M, cudaMemcpyHostToDevice));
    CUDA_SAFE_CALL(cudaMemcpy2D(params->dB[j], sizeof(fupermod_float) * width[j] * K, addr(B,0,j,NULL,width,N,K,2), sizeof(fupermod_float) * ldb,
                                               sizeof(fupermod_float) * width[j] * K, K, cudaMemcpyHostToDevice));
    
    //Note: cublas is column-major, so have to swap matrices A and B
    CUBLAS_SAFE_CALL(CUBLAS_GEMM(params->handles[j], trans_cblas2cublas(TransB), trans_cblas2cublas(TransA), 
                  width[j] * K, M, K, &alpha, params->dB[j], width[j] * K, params->dA[j], K, &beta, params->dC[j], width[j] * K));    
  }
  
  // index increases by 1
  if (PRINT)
    fprintf(stdout, "idx_outer %d\n", params->idx_outer);
  params->idx_outer++;
  
  // at the last iteration, copy result back, and free device memory, and destroy handles
  if ((params->num_iteration) && (params->idx_outer >= params->num_iteration)) {
    fprintf(stdout, "idx_outer %d, params->num_iteration %d, Copy result back, and free dA/B/C\n", params->idx_outer, params->num_iteration);
    device_id = params->device_id;
    for (j = 0; j < nd; j++) {
      // set device
      CUDA_SAFE_CALL(cudaSetDevice(device_id++));
      // synchronous call, control is not returned to host before the device has completed the requested task
      CUDA_SAFE_CALL(cudaMemcpy2D(addr(C,0,j,NULL,width,N,K,3), sizeof(fupermod_float) * ldc, params->dC[j], sizeof(fupermod_float) * width[j] * K,
                                                                sizeof(fupermod_float) * width[j] * K, M, cudaMemcpyDeviceToHost));
      // free memory
      CUDA_SAFE_CALL(cudaFree(params->dA[j])); params->dA[j] = NULL;
      CUDA_SAFE_CALL(cudaFree(params->dB[j])); params->dB[j] = NULL;
      CUDA_SAFE_CALL(cudaFree(params->dC[j])); params->dC[j] = NULL;
      // destroy handles
      CUBLAS_SAFE_CALL(cublasDestroy(params->handles[j])); params->handles[j] = NULL;
    }
    free(params->dA); params->dA = NULL;
    free(params->dB); params->dB = NULL;
    free(params->dC); params->dC = NULL;
    free(params->handles); params->handles = NULL;
  }
}
#endif
