#include "config.h"
#include "fupermod_cblas.h"
#include <stdio.h>
#include <string.h>

#ifdef HAVE_LIBCUBLAS
#include "cuda_auxiliary.hpp"
#include "cutil.h"
#include "cutil_inline_runtime.h"

void gemm_kernel_gpu_single_fit(fupermod_gemm* gemm,
          const enum CBLAS_ORDER Order,
          const enum CBLAS_TRANSPOSE TransA,
          const enum CBLAS_TRANSPOSE TransB, 
          const int M, const int N, const int K, 
          const fupermod_float alpha,
          const fupermod_float *A, const int lda,
          const fupermod_float *B, const int ldb,
          const fupermod_float beta,
          fupermod_float *C, const int ldc)
{
/*  cudaEvent_t start, end;
  CUDA_SAFE_CALL(cudaEventCreate(&start));
  CUDA_SAFE_CALL(cudaEventCreate(&stop));
  CUDA_SAFE_CALL(cudaEventRecord(start, 0));*/
  
  cublasxgemm_params* params = (cublasxgemm_params*)gemm->params;

  // allocate device memory at the beginning
  if (params->idx_outer == 0) {
    // only one device to be used in this case, so the size of dA, dB, dC is 1
    params->dA = (fupermod_float**)malloc(sizeof(fupermod_float*));
    params->dB = (fupermod_float**)malloc(sizeof(fupermod_float*));
    params->dC = (fupermod_float**)malloc(sizeof(fupermod_float*));
    
    if (PRINT)
      fprintf(stdout, "gemm_kernel_gpu_single_fit: M %d N %d K %d\n", M, N, K);
    CUDA_SAFE_CALL(cudaMalloc((void**)&params->dA[0], sizeof(fupermod_float) * M * K));
    CUDA_SAFE_CALL(cudaMalloc((void**)&params->dB[0], sizeof(fupermod_float) * K * N));
    CUDA_SAFE_CALL(cudaMalloc((void**)&params->dC[0], sizeof(fupermod_float) * M * N));
    //CUDA_SAFE_CALL(cudaMemset(params->dC[0], 0, sizeof(fupermod_float) * M * N));
    
    // C = alpha*AxB + beta*C
    if (beta > 1e-4) {
        CUDA_SAFE_CALL(cudaMemcpy(params->dC[0], C, sizeof(fupermod_float) * M * N, cudaMemcpyHostToDevice));
    }
  }
    
  // transfer matrices A and B to device memory, note lda may change
  CUDA_SAFE_CALL(cudaMemcpy2D(params->dA[0], sizeof(fupermod_float) * K, A, sizeof(fupermod_float) * lda, sizeof(fupermod_float) * K, M, cudaMemcpyHostToDevice));
  CUDA_SAFE_CALL(cudaMemcpy2D(params->dB[0], sizeof(fupermod_float) * N, B, sizeof(fupermod_float) * ldb, sizeof(fupermod_float) * N, K, cudaMemcpyHostToDevice));
  
  //Note: cublas is column-major, so have to swap matrices A and B
  CUBLAS_SAFE_CALL(CUBLAS_GEMM(params->handles[0], trans_cblas2cublas(TransB), trans_cblas2cublas(TransA), N, M, K, &alpha, params->dB[0], N, params->dA[0], K, &beta, params->dC[0], N));

  //CUDA_SAFE_CALL(cudaEventRecord(stop, 0));
  CUDA_SAFE_CALL(cudaDeviceSynchronize());
   
  /*if (PRINT)
    fprintf(stdout, "idx_outer %d\n", params->idx_outer);*/
  params->idx_outer++;
  
  /*float time = 0.0f;
  CUDA_SAFE_CALL(cudaEventElapsedTime(&time, start, stop));
  gemm->time += time;*/

  // at the last iteration, copy result back and free device memory, destroy handles
  if ((params->num_iteration) && (params->idx_outer >= params->num_iteration)) {
    if (PRINT)
      fprintf(stdout, "idx_outer %d\nparams->num_iteration %d, Copy result back, and free dA/B/C\n", params->idx_outer, params->num_iteration);
    CUDA_SAFE_CALL(cudaMemcpy(C, params->dC[0], sizeof(fupermod_float) * M * N, cudaMemcpyDeviceToHost));
    CUDA_SAFE_CALL(cudaFree(params->dA[0])); free(params->dA); params->dA = NULL;
    CUDA_SAFE_CALL(cudaFree(params->dB[0])); free(params->dB); params->dB = NULL;
    CUDA_SAFE_CALL(cudaFree(params->dC[0])); free(params->dC); params->dC = NULL;
    CUBLAS_SAFE_CALL(cublasDestroy(params->handles[0])); free(params->handles); params->handles = NULL;
  }
}
#endif
