/*!
 * \page convert1dto2d_mxm_dist Distribution Converter
 * Converts a 1D distribution to a 2D distribution for matrix multiplication.
 * 1D distribution defines the area that should be partitioned to each processor.
 * 2D distribution calculated to minimise the total volume of communication. 
 * This is a serial programme that simply reads in one file and outputs another.
 */		
#include "config.h"
#include "fupermod/fupermod.h"
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "datapart_2d/mm_col_distr.h"
#include "datapart_2d/partition.h"
#include "datapart_2d/group_based_partition.h"

int main(int argc, char** argv) {

  char dist_path1D[256] = "./partition.dist";
  char dist_path2D[256] = "./partition.2ddist";
  char conf_file[256] = "./conf_file";
  //char nodetype_path[256] = "./nodetype";
  /* Here N is matrix dimension in elemtents. 
   * Total matrix size is N x N. 
   * There are N/kb x N/kb blocks. */
  int N = 0; // matrix size
  int kb = 0;

  int opt;
  int version=0;
  while ((opt = getopt(argc, argv, "hvk:N:p:q:m:")) != -1)
    switch (opt) {
      case 'h':
        fprintf(
            stderr,
            "Converts a 1D distribution to a 2D distribution for matrix multiplication in two different ways.\n"
            "Original 2D distribution and group based distribution.\n"
            "1D distribution defines the area that should be partitioned to each processor.\n"
            "2D distribution calculated to minimise the total volume of communication.\n"
            "This is a serial programme that simply reads in one file and outputs another.\n"
            "Usage: 1dto2d_mxm_dist [options]\n"
            "	-h	help\n"
            //"	-l S	shared library (required: a full path or relative to LD_LIBRARY_PATH)\n"
            //"	-o S	suboptions (comma separated options for the shared library)\n"
            " -v   Group based partitioning \n"
            "	-k I block size kb (required)\n"
            "	-N I Total matrix size (required, should be divisible by kb)\n"
            "	-p S path to input 1D distribution (default: %s)\n"
            "	-q S path to output 2D distribution (default: %s)\n"
            " -m S path to conf_file, specify machines to use. (default: %s)\n" 
            , dist_path1D, dist_path2D, conf_file);
        return 0;
        break;
      case 'v':
        version=1;
        break;
      case 'N':
        N = atoi(optarg);
        break; 
      case 'k':
        kb = atoi(optarg);
        break;
      case 'p':
        strcpy(dist_path1D, optarg);
        break;
      case 'q':
        strcpy(dist_path2D, optarg);
        break;
      case 'm':
        strcpy(conf_file, optarg);
        break;
      default:
        fprintf(stderr, "Unknown option %s\n", optarg);
        break;
    }

  if (N < 1 || kb < 1){
    fprintf(stderr, "Error: N and k must be given. N:%d k:%d\n", N, kb);
    exit(2);
  }
  if ( N % kb != 0){
    N = (N / kb) * kb;
    fprintf(stderr, "Warning!\n\tkb does not devide N evenly, rounding to N=%d\n", N);
  }

  // Get host names
  int size; //total number of devices.
  fupermod_process_conf* confs = fupermod_get_conf_all(conf_file, &size);

  if (size < 1) { fprintf(stderr, "Error: size < 1\n"); exit(1);}

  fupermod_dist* dist = fupermod_dist_alloc(0, size);
  FILE *file;
  if ((file = fopen(dist_path1D, "r")) != NULL) {
    if (fupermod_dist_read(file, dist, confs, size) != 0)
      exit(3); 
    fclose(file);
  } else {
    fprintf(stderr, "Error: Could not open dist file: %s\n", dist_path1D);
    exit(4);
  }
  if ((N / kb) != sqrt(dist->D)) {
    fprintf(stderr, "Warning: (N/kb) != sqrt(D) N:%d D:%lld.\nWrong N given here, or wrong D given in partitioner\nGoing to continue anyway with N.\n", N, dist->D);
  }


  double* areas = (double*)malloc(sizeof(double) * size);	
  int i;
  for (i = 0; i < size; i++)
    areas[i] = dist->parts[i].d;
  mm_col_i_distrib* idist = mm_col_dist(size, N/kb, kb, areas);
  free(areas); areas = NULL;

  /* USE TO GET NO OF COLUMNS OF PARTITIONING*/
  int max_j=0;
  for (i = 0; i < idist->nproc; i++) {
    if (idist->d[i].j > max_j) {
      max_j = idist->d[i].j;
    }
  }

  /*!
   * \If group based partitioning option is selected then program take group_id value from conf_file.
   * Generate the optimal partitioning based on groups. 
   * and write partition.2ddist file with one extra column added for Group ID
   */	
  if (version==1) {
    mm_col_i_group *col_group = (mm_col_i_group*)malloc(idist->nproc *sizeof(mm_col_i_group));

    for (i = 0; i < idist->nproc; i++) {
      col_group[i].hostname=confs[i].hostname;
      col_group[i].core_id =confs[i].rank_intra;
      col_group[i].d.i=idist->d[i].i; 
      col_group[i].d.j=idist->d[i].j;
      col_group[i].d.x=idist->d[i].x; 
      col_group[i].d.y=idist->d[i].y;
      col_group[i].d.w=idist->d[i].w;
      col_group[i].d.h=idist->d[i].h;
      if(dist!=NULL)
        col_group[i].time= dist->parts[i].t;
      char* subopts = strdup(confs[i].subopts);
      char* tokens[] = {"g_id", NULL};
      char* value;
      while (*subopts != '\0') {
        switch (getsubopt(&subopts, tokens, &value)) {
          case 0:
            col_group[i].group_id= atoi(value);
            break;
        }
      }    
    }
    group_based_partition(col_group,idist->nproc);
    file = fopen(dist_path2D, "w");
    mm_col_i_group_write(file,idist,col_group);
    fclose(file);
    /* else perform original partitioning */
  } else 
  {
    file = fopen(dist_path2D, "w");
    mm_col_dist_write(file, idist, dist, confs);
    fclose(file);
  }

  mm_col_i_dist_free(idist);
  fupermod_dist_free(dist);
  return 0;
}


