#ifndef FUPERMOD_BENCHMARK_H_
#define FUPERMOD_BENCHMARK_H_

#include "fupermod_kernel.h"

#ifdef __cplusplus
extern "C" {
#endif
  /*!
   * \defgroup fupermod_benchmark_group Benchmarking
   * This module provides API for benchmarking. The key data structure is \ref fupermod_benchmark :
   * it encapsulates the kernel to benchmark and provides the measurement method.
   * There are two methods implemented:
   * - basic parallel benchmark
   * (use \ref fupermod_benchmark_basic_alloc and \ref fupermod_benchmark_basic_free)
   * - emulation via provided FPMs
   * (use \ref fupermod_benchmark_emulate_alloc and \ref fupermod_benchmark_emulate_free)
   * \{
   */

  /*! Verbose level (default: 0 - none) */
  extern int fupermod_verbose;

  /*! Maximum time (30 years) */
#define FUPERMOD_MAX_TIME 1e9

  /*! 
   * Precision of measurement. Used as an input argument of benchmark functions. 
   * To provide reliable results, the computation experiments in each benchmark are repeated either fixed or
   * variable number of times. This data structure allows the user to control the accuracy and efficiency of benchmarking.
   * - Assigning to \ref reps_min and \ref reps_max the same values results in the fixed number of repetitions of
   * the computation operation, with the \ref cl and \ref eps arguments being ignored
   * (this allows the user to control the efficiency of benchmarking).
   * - If \ref reps_min < \ref reps_max, the experiments are repeated until a confidence interval, \ref fupermod_point::ci,
   * found with the confidence level, \ref cl = \f$ Pr(|\bar T - \mu| < ci) \f$, satisfies
   * \f$ \displaystyle\frac{ci}{\bar T} < \f$ \ref eps, or the number of repetitions reaches its maximum,
   * \ref reps_max (this allows the user to control the accuracy of benchmarking).
   * - If total benchmarking time exceeds \ref time_max_rep then no more repetitions are attempted. 
   * - If \ref zero_speed > 0, pthreads are used which allows the benchmark to be killed after a time interval x, 
   *   where  x = min(complexity(d)/\ref zero_speed, time_max_kill).
   */
  typedef struct fupermod_precision {
    /*! Minimum number of repetitions */
    int reps_min;
    /*! Maximum number of repetitions */
    int reps_max;
    /*! Confidence level */
    double cl;
    /*! Relative error */
    double eps;
    /*! If benchmark has exceeded this time do not do another repetition */
    double time_max_rep;
    /*! If benchmark slower then this speed, kill it with pthreads, if zero_speed <= 0, do not use pthreads */
    double zero_speed;
    /*! If zero_speed > 0, and benchmark has exceeded this time kill it with pthreads */
    double time_max_kill;
  } fupermod_precision;

  /*! Returns default precision settings */
  fupermod_precision fupermod_precision_defaults();

  /*!
   * Returns a confidence interval that contains the average execution time with a certain probability:
   * \f$ Pr(|\bar T - \mu| < ci) = cl \f$.
   * \note We assume that the execution times form an independent sample from a normally distributed population,
   * and use t distribution to estimate confidence interval.
   * \param cl confidence level
   * \param reps number of measurements (should be > 1)
   * \param t array of reps measurement results
   * \return confidence interval
   */
  double fupermod_ci(double cl, int reps, double* t);

  /*! Data point */
  typedef struct fupermod_point {
    /*! Problem size */
    long long int d;
    /*! Time */
    double t;
    /*! Number of repetitions the measurement has actually taken */
    int reps;
    /*!
     * Confidence interval, \f$ |\bar T - \mu| < ci \f$.
     * The \ref fupermod_ci function estimates confidence interval, using t distribution.
     */
    double ci;
  } fupermod_point;

  /*! Reads data point from file */
  int fupermod_point_read(FILE* stream, fupermod_point* point);

  /*! Writes data point to file */
  int fupermod_point_write(FILE* stream, fupermod_point* point, fupermod_complexity complexity);

  /*!
   * Calculates the data point of a node from the data points of the processes running on the node
   * \param point data point of a process
   * \param point_node data point of a multiprocessor/multicore (significant only at root of intra communicator)
   * \param comm_intra intra communicator
   * \param root_intra root of intra communicator
   */
  int fupermod_point_node(fupermod_point* point, fupermod_point* point_node, MPI_Comm comm_intra, int root_intra);

  /*! Benchmark */
  typedef struct fupermod_benchmark {
    /*! Computational kernel */
    fupermod_kernel* kernel;
    /*! Process hostname, rank_intra, device type (cpu or gpu etc),
     * process specific suboptions*/
    fupermod_process_conf conf;
    /*!
     * Execute the benchmark
     * \param benchmark benchmark
     * \param comm_sync MPI communicator used for synchronisation (probably should be comm_iter or MPI_COMM_SELF).
     * \param d problem size
     * \param precision precision
     * \param point data point
     * \return error status
     */
    int (*execute)(struct fupermod_benchmark* benchmark, MPI_Comm comm_sync, long long int d, fupermod_precision precision, fupermod_point* point);
  } fupermod_benchmark;

  /*! Deallocates the benchmark */
  typedef void (*fupermod_benchmark_free)(fupermod_benchmark* benchmark);

  /*!
   * Allocates the benchmark that executes the kernel at all processes.
   * \param kernel Kernels are synchronised and executed the same number of times.
   * \param conf Describes the type of device benchmark is executed on (cpu or gpu)
   * The result, the data point, is returned separately for each process.
   */
  fupermod_benchmark* fupermod_benchmark_basic_alloc(fupermod_kernel* kernel, fupermod_process_conf conf);


  /*! Deallocates the basic benchmark */
  void fupermod_benchmark_basic_free(fupermod_benchmark* benchmark);

  struct fupermod_model;

  /*! Allocates the benchmark that looks up the existing models at all processes */
  fupermod_benchmark* fupermod_benchmark_emulate_alloc(fupermod_kernel* kernel, struct fupermod_model* model);

  /*! Deallocates the emulate benchmark */
  void fupermod_benchmark_emulate_free(fupermod_benchmark* benchmark);

  /*!
   * \}
   */
#ifdef __cplusplus
}
#endif

#endif /*FUPERMOD_BENCHMARK_H_*/
