#include "fupermod_timer.h"
#include "fupermod_debug.h"
#include "fupermod_benchmark.h"
#include <stdlib.h>
#include <pthread.h>
#include <errno.h>

int fupermod_benchmark_basic_execute(struct fupermod_benchmark* benchmark, MPI_Comm comm, long long int d, fupermod_precision precision, fupermod_point* point) {
  // initialise
  int rank_global;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank_global);
  void* params;
  int rc = benchmark->kernel->initialize(d, &params);
  if (rc == FUPERMOD_ALLOC_FAIL) {
    point->d = d;
    point->t = FUPERMOD_MAX_TIME;
    point->reps = 0;
    point->ci = 0;
    return 0;
  }
  if (rc != FUPERMOD_SUCCESS) {
    fprintf(stderr, "%s: kernel initialization failed at proc %d\n", __func__, rank_global);
    return -1;
  }
  // global flag: 0 (continue), 1 (success), -1 (timeout), -100 (error)
  int stop = 0;
  double sum = 0;
  int reps = 0;
  double ci = 0;
  double* t = (double*)malloc(sizeof(double) * precision.reps_max);
  while (reps < precision.reps_max && !stop) {
    // local flag
    int res = 0;
    // synchronise the processes
    MPI_Barrier(comm);
    pthread_cond_t cond  = PTHREAD_COND_INITIALIZER;
    pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
    int rc;
    double time;
    struct timespec start, end;
    
    void* kernel_thread() {
      pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
      pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
      get_time(&start);
      rc = benchmark->kernel->execute(&mutex, params);
      get_time(&end);
      pthread_cond_signal(&cond);
      return NULL;
    }
    if (precision.zero_speed <= 0) {
      // no pthreads used
      kernel_thread();
      time = get_timediff(&start, &end);
    } else {
      int interval = benchmark->kernel->complexity(d, NULL)/ precision.zero_speed;
      if (interval < 20)
        interval = 20; // minimum 20 seconds.
      if (interval > precision.time_max_kill) 
        interval = precision.time_max_kill;
      struct timespec begin;
      get_time(&begin);
      struct timespec future;
      future.tv_sec = begin.tv_sec + interval;
      future.tv_nsec = begin.tv_nsec;
      pthread_mutex_lock(&mutex);
      pthread_t thread;
      // kernel theread started here.
      pthread_create(&thread, NULL, kernel_thread, NULL);
      int rct = pthread_cond_timedwait(&cond, &mutex, &future);
      pthread_mutex_unlock(&mutex);
      if (rct == ETIMEDOUT) {
        // time = kernel->complexity(d)/ (precision.zero_speed / 10000);
        time = 1e9;
        fprintf(stderr, "%s: P%d Notice: Killing benchmark because of timeout:%dseconds\n", __func__, rank_global, interval);
        int s = pthread_cancel(thread);
        if (s != 0)
          fprintf(stderr, "Error with pthread_cancel:%d\n", s);
        void *res;
        s = pthread_join(thread, &res);
        if (s != 0)
          fprintf(stderr, "Error with with pthread_join after pthread_cancel:%d\n", s);
        if (res == PTHREAD_CANCELED)
          fprintf(stderr, "thread was canceled\n");
        else
          fprintf(stderr, "thread wasn't canceled (shouldn't happen!)\n");
        rc = FUPERMOD_SUCCESS; // need to set since we canceled
      } else {
        int s;
        void *res;
        s = pthread_join(thread, &res);
        if (s != 0)
          fprintf(stderr, "Error with with pthread_join:%d\n", s);

        time = get_timediff(&start, &end);
      }
    }
    if (rc != FUPERMOD_SUCCESS) {
      fprintf(stderr, "%s: kernel execution failed at proc %d\n", __func__, rank_global);
      res = -100;
    }
    sum += t[reps] = time;
    // exit if time exceeds
    if (t[reps] > precision.time_max_rep)
      res = -1;
    // exit if statistics
    if (reps >= precision.reps_min - 1 && reps >= 2 &&
        (ci = fupermod_ci(precision.cl, reps, t)) * reps / sum < precision.eps)
      res = 1;
    MPI_Allreduce(&res, &stop, 1, MPI_INT, MPI_MIN, comm);
    reps++;
  }
  free(t);
  // failure
  if (stop == -100)
    return -1;
  point->d = d;
  point->t = sum / reps;
  point->reps = reps;
  point->ci = ci;
  // finalise
  if (benchmark->kernel->finalize(params) != FUPERMOD_SUCCESS) {
    fprintf(stderr, "%s: kernel finalization failed at proc %d\n", __func__, rank_global);
    return -1;
  }
  return 0;
}

fupermod_benchmark* fupermod_benchmark_basic_alloc(fupermod_kernel* kernel, fupermod_process_conf conf) {
  fupermod_benchmark* benchmark = (fupermod_benchmark*)malloc(sizeof(fupermod_benchmark));
  benchmark->kernel = kernel;
  benchmark->conf = conf;
  benchmark->execute = fupermod_benchmark_basic_execute;
  return benchmark;
}

void fupermod_benchmark_basic_free(fupermod_benchmark* benchmark) {
  free(benchmark);
}
