#ifndef FUPERMOD_DYNAMIC_H_
#define FUPERMOD_DYNAMIC_H_

#include "fupermod_partition.h"
#include <sys/time.h>

#ifdef __cplusplus
extern "C" {
#endif
  /*!
   * \defgroup fupermod_dynamic_group Dynamic data partitioning and load balancing
   * \{
   */

  /*!
   * Dynamic partitioner/balancer
   * Code snippet for dynamic data partitioning
   * \code
   * fupermod_dynamic partitioner;
   * fupermod_benchmark* benchmark = fupermod_benchmark_X_alloc();
   * int status = 0;
   * while (!status) {
   *   fupermod_partitioner_iterate(&partitioner, comm, root, precision, benchmark);
   *   if (rank == root)
   *     status = fupermod_dist_test(partitioner.dist, eps);
   *   MPI_Bcast(&status, 1, MPI_INT, root, comm);
   * }
   * fupermod_benchmark_X_free(benchmark);
   * \endcode
   * Code snippet for dynamic load balancing
   * \code
   * fupermod_dynamic balancer;
   * loop {
   *   long long int d;
   *   fupermod_dist_scatter(balancer.dist, &d, comm, root);
   *   ... // redistribute data according to balancer.dist
   *
   *   struct timeval start;
   *   gettimeofday(&start, NULL);
   *   ... // execute an iteration of the routine
   *   fupermod_balancer_iterate(&balancer, comm, root, start);
   * }
   * \endcode
   */
  typedef struct fupermod_dynamic {
    /*! Partitioning algorithm */
    fupermod_partition partition;
    /*! Number of processors */
    int size;
    /*! Current functional performance models */
    fupermod_model** models;
    /*! Current distribution */
    fupermod_dist* dist;
  } fupermod_dynamic;

  /*!
   * Iterate dynamic partitioner
   * \param partitioner dynamic partitioner (significant only at root)
   * \param comm MPI communicator
   * \param root rank of the root processor
   * \param precision precision
   * \param benchmark benchmark
   * \param eps relative error
   */
  int fupermod_partitioner_iterate(fupermod_dynamic* partitioner, MPI_Comm comm, int root, fupermod_precision precision, fupermod_benchmark* benchmark, double eps);

  /*!
   * Iterate dynamic partitioner one - does a benchmark of size one
   */
  int fupermod_partitioner_iterate_one(fupermod_dynamic* partitioner, MPI_Comm comm, int root, fupermod_precision precision, fupermod_benchmark* benchmark, double eps);
  /*!
   * Iterate dynamic load balancer
   * \param balancer dynamic load balancer (significant only at root)
   * \param comm MPI communicator
   * \param root rank of the root processor
   * \param start start time
   */
  int fupermod_balancer_iterate(fupermod_dynamic* balancer, MPI_Comm comm, int root, struct timespec start);

  /*!
   * \}
   */
#ifdef __cplusplus
}
#endif

#endif /* FUPERMOD_DYNAMIC_H_ */
