#ifndef FUPERMOD_KERNEL_H_
#define FUPERMOD_KERNEL_H_

#include <pthread.h>
#include "fupermod_conf.h"

#ifdef __cplusplus
extern "C" {
#endif
  /*!
   * \defgroup fupermod_kernel_group Computational kernel of a routine
   * A routine is a standalone parallel application supplied with a shared library
   * implementing its computational kernel, \ref fupermod_kernel.
   * The fupermod tools load the shared library and search for the \ref FUPERMOD_KERNEL_INSTANCE symbol,
   * the instance of the \ref fupermod_kernel data structure.
   * \see \ref template_kernel
   * \{
   */

  /*! Success return value for kernel */
#define FUPERMOD_SUCCESS 0

  /*! Failure return value for kernel */
#define FUPERMOD_FAIL -1

  /*! Memory failure return value for kernel. Sometimes this may be considered as normal behaviour. */
#define FUPERMOD_ALLOC_FAIL -2

  /*!
   * Computational complexity of the kernel. In general, linearly dependent on the problem size.
   * \param d problem size
   * \return complexity
   */
  typedef double (*fupermod_complexity)(long long int d, void* params);

  /*! Computational kernel of a routine */
  typedef struct fupermod_kernel {
    /*!
     * Loads the library. Parses suboptions at root and broadcasts the routines settings in communicator.
     * Some tools are non-MPI, therefore MPI_COMM_NULL should be handled.
     * \param comm MPI communicator
     * \param conf process configuration
     * \return return code
     */
    int (*load)(MPI_Comm comm, fupermod_process_conf* conf);

    /*!
     * Unloads the library. Some tools are non-MPI, therefore MPI_COMM_NULL should be handled.
     * \param comm MPI communicator
     * \return return code
     */
    int (*unload)(MPI_Comm comm);

    /*! Complexity */
    fupermod_complexity complexity;

    /*!
     * Initializes the kernel. Allocates data for benchmarking, which should be stored in global variables.
     * \param d problem size
     * \param params kernel parameters (if used, should be allocated here)
     * \return return code
     */
    int (*initialize)(long long int d, void** params);

    /*!
     * Executes the kernel. Uses the global variables of the kernel.
     * Kernel can be executed in a separate thread to enable termination in case of long run.
     * \param mutex a mutex to protect some resources:
     * \code
     * pthread_mutex_lock(mutex);
     * // use of resources
     * pthread_mutex_unlock(mutex);
     * \endcode
     * \param params kernel parameters
     * \return return code
     */
    int (*execute)(pthread_mutex_t* mutex, void* params);

    /*!
     * Finalizes the kernel. Free the memory for the kernel.
     * \param params kernel parameters (if used, should be deallocated here)
     * \return return code
     */
    int (*finalize)(void* params);
  } fupermod_kernel;

  /*! Kernel kernel instance */
#define FUPERMOD_KERNEL_INSTANCE "fupermod_kernel_instance"

  /*! Shared library implementing the computational kernel of a routine */
  typedef struct fupermod_library {
    /*! Path to the library */
    char* path;

    /*!
     * Process configuration
     * \note Use the same configuration for both the routine and its kernel.
     */
    fupermod_process_conf* conf;

    /*! Shared library handle */
    void* handle;

    /*! Computational kernel of a routine */
    fupermod_kernel* kernel;
  } fupermod_library;

  /*!
   * Loads a library implementing the computation kernel of a routine
   * \param comm MPI communicator, can be MPI_COMM_NULL
   * \param path path to the shared library
   * \param conf process configuration
   * \param library the shared library data structure
   * \return error code
   */
  int fupermod_library_load(MPI_Comm comm, char* path, fupermod_process_conf* conf, fupermod_library* library);

  /*!
   * Unloads a library implementing the computation kernel of a routine
   * \param comm MPI communicator, can be MPI_COMM_NULL
   * \param library the shared library data structure
   * \return error code
   */
  int fupermod_library_unload(MPI_Comm comm, fupermod_library* library);

  /*!
   * \}
   */
#ifdef __cplusplus
}
#endif

#endif /* FUPERMOD_KERNEL_H_ */
