#ifndef FUPERMOD_MODEL_H_
#define FUPERMOD_MODEL_H_

#include "fupermod_data.h"

#ifdef __cplusplus
extern "C" {
#endif
  /*!
   * \defgroup fupermod_model_group Approximations of a time function
   * API for approximation of the time function for a kernel.
   * The main structure is \ref fupermod_model : it encapsulates
   * the data points and the complexity function of the kernel
   * and provides the approximation of the time function and its derivative.
   * There are three models implemented:
   * - Constant model based on the last observation
   * - Constant model based on the approximated value for a selected point
   * (for approximation, the functional model based on Akima spline interpolation is used)
   * - Functional model based on Akima spline interpolation
   * - Piecewise linear model
   * \{
   */

  /*!
   * Model: approximation of time (based on data points). Must be defined at [0; D].
   * Abstract data structure used in data partitioning algorithms.
   * Encapsulates the data points provided by \ref fupermod_data.
   * Ensure data integrity (data points <-> model):
   * after a new data point is added by \ref fupermod_data_add,
   * the model must be updated by \ref update.
   */
  typedef struct fupermod_model {
    /*! Data */
    fupermod_data* data;
    /*! Complexity */
    fupermod_complexity complexity;
    /*! Maximum problem size */
    long long int D;
    /*! Evaluates the function. Must return positive values. */
    double (*t)(struct fupermod_model* model, double x);
    /*! Evaluates the derivative (required for the numerical data partitioners only) */
    double (*dt)(struct fupermod_model* model, double x);
    /*! Updates the model */
    int (*update)(struct fupermod_model* model, fupermod_point point);
    /*! Application spicific parameters (extra args for complexity) */
    void* params;
  } fupermod_model;

  /*! Allocates the model */
  typedef fupermod_model* (*fupermod_model_alloc)(fupermod_data* data, fupermod_complexity complexity, long long int D, void* params, int);
  /*! Deallocates the model */
  typedef void (*fupermod_model_free)(fupermod_model* model);

  /*! Returns estimated speed */
  double fupermod_model_speed(fupermod_model* model, long long int d);

  /*! Prints models */
  void fupermod_print_model(FILE* stream, fupermod_library* library, fupermod_process_conf conf, fupermod_model* model, long long int D);

  /* Prints the data points that make up a model */
  void fupermod_print_models_data(FILE* stream, const int size, fupermod_model** models);

  /*!
   * Allocates the constant model, using a data point from observations
   * \param data observations
   * \param complexity computational complexity
   * \param D maximum problem size
   * \param params unused variable
   * \param smooth unused variable
   * \return model
   */
  fupermod_model* fupermod_model_constant_alloc(fupermod_data* data, fupermod_complexity complexity, long long int D, void* params, int smooth);

  /*! Frees the model */
  void fupermod_model_constant_free(fupermod_model* model);

  /*!
   * Allocates the constant model. For use with full models, fits spline interpolation to all data so that a single value speed can be found for "bench_size".
   * \param data observations
   * \param complexity computational complexity
   * \param D maximum problem size
   * \param params unused variable
   * \param bench_size a data point whose time approximation will be used as the value of the constant model
   * \return model
   */
  fupermod_model* fupermod_model_constant_interp_alloc(fupermod_data* data, fupermod_complexity complexity, long long int D, void* params, int bench_size);

  /*! Frees the model */
  void fupermod_model_constant_interp_free(fupermod_model* model);

  /*!
   * Allocates piecewise model and fixes data such that it is monotonically increasing, concave and then monotonically decreasing.
   * \param data data points
   * \param complexity computational complexity
   * \param D maximum problem size
   * \param params unused variable
   * \param smooth unused variable
   */
  fupermod_model* fupermod_model_piecewise_alloc(fupermod_data* data, fupermod_complexity complexity, long long int D, void* params, int smooth);

  /*! Frees the model */
  void fupermod_model_piecewise_free(fupermod_model* model);

  /*! Prints piecewise models */
  void fupermod_print_models_piecewise(FILE* stream, int size, fupermod_model** models, long long int D);

  /*!
   * Allocates the model based on the Akima spline interpolation.
   * Designed for models with few data points (partial models)
   * \param data data points
   * \param complexity computational complexity
   * \param D maximum problem size
   * \param params unused variable
   * \param smooth smooth parameter for the Akima spline interpolation of full functional performance models (0 - no smoothing)
   * \result model
   */
  fupermod_model* fupermod_model_interp_alloc(fupermod_data* data, fupermod_complexity complexity, long long int D, void* params, int smooth);

  /*! Frees the model */
  void fupermod_model_interp_free(fupermod_model* model);

  /*! Returns pointer to allocated memory containing the name of a data file in the form: "path.hostname.rank_intra.device_type.model". 
   * Expect the tool to already have populated conf with appropriate info.
   * User must free the allocated memory manually
   * \param path initial path
   * \param conf configuration to generate the rest of the file name
   */
  char* fupermod_model_filename(const char* path, fupermod_process_conf conf);

  /*!
   * \}
   */
#ifdef __cplusplus
}
#endif

#endif /*FUPERMOD_MODEL_H_*/
