#include "fupermod_model_interp.h"
#include "fupermod_debug.h"

double fupermod_model_interp_t(struct fupermod_model* model, double x) {
  fupermod_model_interp* interp = (fupermod_model_interp*)model;
  double t = gsl_spline_eval(interp->spline, x, interp->accel);
  return t;
}

double fupermod_model_interp_dt(struct fupermod_model* model, double x) {
  fupermod_model_interp* interp = (fupermod_model_interp*)model;
  return gsl_spline_eval_deriv(interp->spline, x, interp->accel);
}

int fupermod_model_interp_update(fupermod_model* model, fupermod_point point) {
  if (point.d > 0){
    fupermod_data_add(model->data, point);
  }
  int count = model->data->count;
  if (count > 0) {
    // model
    fupermod_model_interp* interp = (fupermod_model_interp*)model;
    if (interp->accel != NULL)
      gsl_interp_accel_free(interp->accel);
    if (interp->spline != NULL)
      gsl_spline_free(interp->spline);

    int D = model->D;
    int extra_end, extra_start = 3;
    if (model->data->points[count - 1].d > D) {
      // we dont need extra end points.
      extra_end = 0;
    } else {
      // we need a few end points.
      extra_end = 3;
    }

    interp->accel = gsl_interp_accel_alloc();
    interp->spline = gsl_spline_alloc(gsl_interp_akima, count + extra_start + extra_end);
    size_t count_extra = count + extra_start + extra_end;
    double* xa = (double*)malloc(sizeof(double) * count_extra);
    double* ta = (double*)malloc(sizeof(double) * count_extra);
    double interval = ((double)model->data->points[0].d) * 0.95 / extra_start;
    int i;
    for (i = 0; i < extra_start; i++) {
      xa[i] = i * interval;
      ta[i] = i * model->data->points[0].t / extra_start;
    }
    for (i = 0; i < count; i++) {
      xa[i + extra_start] = (model->data->points[i].d);
      // Uncomment to under estimate speed:
      // xa[i + extra_start] = (model->data->points[i].d) * 0.95;
      ta[i + extra_start] = model->data->points[i].t;
    }
    if (extra_end > 0) {
      interval = (D - (model->data->points[count - 1].d) * 0.95) / extra_end;
      if (interval <= 0)
        interval = 1;
      for (i = 0; i < extra_end; i++) {
        double d_end = model->data->points[count - 1].d;
        double d =  d_end + (i + 1) * interval;
        xa[count + extra_start + i] = d;
        ta[count + extra_start + i] = (d / d_end) * model->data->points[count - 1].t;
      }
    }
    // Fix points - time increasing.
    double t = 0;
    for (i = extra_start; i < extra_start + count; i++) {
      if(ta[i] < t){
        //		j = 1;
        //		while (ta[i + j] < t)
        //			j++;
        ta[i] = t;
      } else 
        t = ta[i];
    }

    // Smooth
    if (interp->smooth > 0 && count > 10) {
      int avg;
      int j;
      double* ta_s = (double*)malloc(sizeof(double) * (count + extra_start + extra_end));
      for (i = 0; i < count + extra_start + extra_end; i++){
        double sum = 0;
        avg = 0;
        for (j = -interp->smooth / 2; j < interp->smooth / 2; j++){
          if ((i + j) > 0 && (i + j) < count + extra_start + extra_end) {
            sum += ta[i + j];
            avg++;
          } else {
            sum += ta[i];
            avg++;
          }
        }
        ta_s[i] = sum / avg;
      }
      free(ta);
      ta = ta_s;
    }
    gsl_spline_init(interp->spline, xa, ta, count + extra_start + extra_end);
    free(xa);
    free(ta);
  }
  return 0;
}

fupermod_model* fupermod_model_interp_alloc(fupermod_data* data, fupermod_complexity complexity, long long int D, void* params, int smooth) {
  NOT_USED(params);
  fupermod_model_interp* interp = (fupermod_model_interp*)malloc(sizeof(fupermod_model_interp));
  fupermod_model* model = (fupermod_model*)interp;
  model->data = data;
  model->complexity = complexity;
  model->D = D;
  model->t = fupermod_model_interp_t;
  model->dt = fupermod_model_interp_dt;
  model->update = fupermod_model_interp_update;
  interp->smooth = smooth;
  interp->accel = NULL;
  interp->spline = NULL;
  fupermod_model_interp_update(model, (fupermod_point){0, 0, 0, 0});
  return model;
}

void fupermod_model_interp_free(fupermod_model* model) {
  if (model) {
    fupermod_model_interp* interp = (fupermod_model_interp*)model;
    if (interp->accel != NULL)
      gsl_interp_accel_free(interp->accel);
    if (interp->spline != NULL)
      gsl_spline_free(interp->spline);
    free(model);
  }
}
