#include "fupermod_debug.h"
#include "fupermod_model_piecewise.h"
#include <stdlib.h>

/*!
 * Recursive function to find iterval of a given problem size, returns position in array.
 */
int divide_interval_data(fupermod_model_piecewise* model, int low, int high, int point){
  int mid = (low + high) / 2;
  if (mid == low)
    return low;
  if (model->d[mid] < point) {
    low = divide_interval_data(model, mid, high, point);
  } else if (model->d[mid] > point){
    low = divide_interval_data(model, low, mid, point);
  } else if (model->d[mid] == point){
    return mid;
  }
  return low;
}

double fupermod_model_piecewise_s(struct fupermod_model* model, double x) {
  // return speed.
  fupermod_model_piecewise* piecewise = (fupermod_model_piecewise*)model;
  int count = piecewise->count;
  if (x <= piecewise->d[0]) {
    // data is less then points, use horizontal line to zero
    return piecewise->s[0];
  }
  if (x >= piecewise->d[count - 1]) {
    // data is greater then points, use horizontal line to infinity
    return piecewise->s[count - 1];
  }
  int pos = divide_interval_data(piecewise, 0, count - 1, x);
  if (x == piecewise->d[pos])
    return piecewise->s[pos];
  double m = (piecewise->s[pos+1] - piecewise->s[pos]) / (piecewise->d[pos+1] - piecewise->d[pos]);
  double b = piecewise->s[pos] - m * piecewise->d[pos];
  return m * x + b;
}

double fupermod_model_piecewise_t(struct fupermod_model* model, double x) {
  // return time
  return model->complexity(x, NULL) / fupermod_model_piecewise_s(model, x);
}
double fupermod_model_piecewise_dt(struct fupermod_model* model, double x) {
  //todo: return true derivative
  NOT_USED(model);
  NOT_USED(x);
  return 0;
}

/*! Data points structure for fixing data */
typedef struct fupermod_piecewise_points {
  /*! Problem size */
  double* d;
  /*! speed */
  double* s;
  /*! next point in linked list */
  int* next;
  /*! Previous point in linked list */
  int* prev;
  /*! First point in linked list */
  int first;
  /*! Last point in linked list */
  int last;
} fupermod_piecewise_points;

void remove_point(fupermod_piecewise_points* p, int i) {
  if (p->d[i] > 0 && (p->next[i] >= 0 || p->prev[i] >=0)) {
    p->d[i] = 0;
    if (p->prev[i] >= 0)
      p->next[p->prev[i]] = p->next[i];
    else 
      p->first = p->next[i];

    if (p->next[i] >= 0)
      p->prev[p->next[i]] = p->prev[i];
    else
      p->last = p->prev[i];

    p->next[i] = p->prev[i] = -1;
  }
}

int fupermod_model_piecewise_update(fupermod_model* model, fupermod_point point) {
  /* From "Data Partitioning with a Functional Performance Model of 
   * Heterogeneous Processors" 2007
   *
   * On the interval [0, x], the function is:
   *    monotonically increasing and
   *    concave.
   *
   * On the interval [x, ∞), the function is monotonically decreasing.
   *
   * Hence any straight line coming through the origin of the coordinate 
   * system intersects the graph of the function in no more than one point.
   * ie. Ensure single solution.
   *
   * Code below fixes data points (removes them) when they do not obey rules 
   * above.
   */

  if (point.d > 0){
    fupermod_data_add(model->data, point);
  }
  fupermod_model_piecewise* piecewise = (fupermod_model_piecewise*)model;
  int count = model->data->count;
  if (count > 0) {
    fupermod_piecewise_points p;
    p.d = (double*)malloc(sizeof(double) * (count));
    p.s = (double*)malloc(sizeof(double) * (count));
    p.next = (int*)malloc(sizeof(int) * (count));
    p.prev = (int*)malloc(sizeof(int) * (count));
    p.first = 0;
    p.last = count -1;
    int i;
    int maxi = 0;
    for (i = 0; i < count; i++) {
      //underestimate function a little.
      p.d[i] = (model->data->points[i].d) * 0.95;
      p.s[i] = model->complexity(model->data->points[i].d, NULL) / model->data->points[i].t;
      p.next[i] = i + 1;
      p.prev[i] = i - 1;
      // find highest point x for intervals [0,x], [x, ∞)
      if (p.s[i] > p.s[maxi]){
        maxi = i;
      }
    }
    p.next[count -1] = -1;

    double m_max = p.s[maxi] / p.d[maxi];
    double current_max = p.s[0];
    // monotonically increasing up to x (maxi)
    i = p.first;
    while(i < maxi && i >= 0) {
      int next_i = p.next[i];
      if (p.s[i] / p.d[i] < m_max) {
        // point lies under line to max, remove it.
        remove_point(&p, i);
      } else {
        if (p.s[i] < current_max) {
          // not increasing
          remove_point(&p, i);
        } else {
          current_max = p.s[i];
        }
      }
      i = next_i;
    }

    //and concave up to x (maxi).
    double* m = (double*)malloc(sizeof(double) * (count));
    i = p.first;
    while(i < maxi && i >= 0) {
      m[i] = (p.s[p.next[i]] - p.s[i]) / (p.d[p.next[i]] - p.d[i]);
      if (p.prev[i] >= 0 && m[i] > m[p.prev[i]]) {
        // remove current point and go back one step
        int tmp_i = p.prev[i];
        remove_point(&p, i);
        i = tmp_i;
      } else
        i = p.next[i];
    }
    free(m); m = NULL;

    // monotonically decreasing in range [x, ∞).
    current_max = p.s[maxi];
    i = maxi;
    while(i <= p.last && i >= 0) {
      if (p.s[i] > current_max) {
        p.s[i] = current_max;
      } else {
        current_max = p.s[i];
      }
      i = p.next[i];
    }

    int countn = 0;
    i = p.first;
    while(i <= p.last && i >= 0) {
      i = p.next[i];
      countn++;
    }

    piecewise->s = (double *)malloc(sizeof(double) * countn);
    piecewise->d = (double *)malloc(sizeof(double) * countn);
    int n = p.first;
    for (i = 0; i < countn; i++) {
      piecewise->d[i] = p.d[n];
      piecewise->s[i] = p.s[n];
      n = p.next[n];
    }
    piecewise->count = countn;

    free(p.d); 		p.d 		= NULL;
    free(p.s); 		p.s 		= NULL;
    free(p.next); p.next 	= NULL;
    free(p.prev); p.prev 	= NULL;
  }
  return 0;
}

fupermod_model* fupermod_model_piecewise_alloc(fupermod_data* data, fupermod_complexity complexity, long long int D, void* params, int smooth) {
  NOT_USED(smooth);
  fupermod_model_piecewise* piecewise = (fupermod_model_piecewise*) malloc(sizeof(fupermod_model_piecewise));
  fupermod_model* model = (fupermod_model*)piecewise;
  model->data = data;
  model->complexity = complexity;
  model->D = D;
  model->t = fupermod_model_piecewise_t;
  model->dt = fupermod_model_piecewise_dt;
  model->update = fupermod_model_piecewise_update;
  model->params = params;
  piecewise->count = 0;
  piecewise->d = NULL;
  piecewise->s = NULL;
  fupermod_model_piecewise_update(model, (fupermod_point){0, 0, 0, 0});
  return model;
}

void fupermod_model_piecewise_free(fupermod_model* model) {
  if (model) {
    fupermod_model_piecewise* piecewise = (fupermod_model_piecewise*)model;
    free(piecewise->d);
    free(piecewise->s);
    free(model);
  }
}

void fupermod_print_models_piecewise(FILE* stream, int size, fupermod_model** models_, long long int D) {
  fupermod_model_piecewise** models = (fupermod_model_piecewise**)models_;
  int row = 0; 
  int continue_ = 1;

  int i;
  for (i = 0; i < size; i++) 
    fprintf(stream, "d%d          \ts%d          \t", i, i); 
  fprintf(stream, "\n");
  for (i = 0; i < size; i++) 
    fprintf(stream, "%e\t%le\t", 0.0, models[i]->s[0]); 
  fprintf(stream, "\n");
  while (continue_) {
    continue_ = 0;
    for (i = 0; i < size; i++) {
      if (row < models[i]->count && models[i]->d[row] < D) {
        fprintf(stream, "%e\t%le\t", models[i]->d[row], models[i]->s[row]); 
      } else {
        fprintf(stream, ".  \t.            \t");
      }
    }
    fprintf(stream, "\n");
    row++;
    for (i = 0; i < size; i++) 
      if (row < models[i]->count && models[i]->d[row] < D) 
        continue_ = 1;
  }
  for (i = 0; i < size; i++) 
    fprintf(stream, "%e\t%le\t", (double)D, models[i]->s[models[i]->count - 1]); 
  fprintf(stream, "\n\n\n");
}
