#include "fupermod.h"
#include "fupermod_partition.h"
#include <stdlib.h>
#include <math.h>
#include <string.h>

fupermod_dist* fupermod_dist_alloc(long long int D, int size) {
  fupermod_dist* dist = (fupermod_dist*)malloc(sizeof(fupermod_dist));
  dist->D = D;
  dist->size = size;
  dist->parts = (fupermod_part*)calloc(size, sizeof(fupermod_part));
  int i;
  for (i = 0; i < size; i++)
    dist->parts[i].d = D / size + ((i < D % size) ? 1 : 0);
  return dist;
}

void fupermod_dist_free(fupermod_dist* dist) {
  if (dist != NULL) {
    free(dist->parts);
    free(dist);
  }
}

/*! Reads fupermod_dist from file*/
int fupermod_dist_read(FILE* stream, fupermod_dist* dist, fupermod_process_conf* confs, int maxsize) {
  int size = 0;
  int D = 0;
  if (fscanf(stream, "#size %d\n", &size) != 1) {
    fprintf(stderr, "Error: Error, failed to read dist file in correct format (#size)\n");
    return -1;
  }
  if (size > maxsize) {
    fprintf(stderr, "Error: size in dist file greater than max size\n");
    return -1;
  }
  if (fscanf(stream, "#D %d\n", &D) != 1) {
    fprintf(stderr, "Error: Error, failed to read dist file in correct format (#D)\n");
    return -1;
  }
  if (D < 0) {
    fprintf(stderr, "Error: Reading total problem size\n");
    return -1;
  }
  dist->D = D;
  dist->size = size;

  int* check = (int*)calloc(sizeof(int), maxsize);

  int i;
  char *line = NULL;
  size_t len = 0;
  getline(&line, &len, stream); //skip column headers
  for (i = 0; i < size; i++){
    char hostname[MPI_MAX_PROCESSOR_NAME];
    int rank_intra;
    long long int d;
    double t;
    int _i;
    long long int x;
    if (getline(&line, &len, stream) < 2) {
        fprintf(stderr, "Error reading in next line %d %s\n", i, __func__);
        return -1;
    }
    if (sscanf(line, "%s\t%d\t%lld\t%le\t%d\t%lld", hostname, &rank_intra, &d, &t, &_i, &x) != 6) {
      fprintf(stderr, "Error reading line, format not recognised\n");
      fprintf(stderr, "Line %d: %s\n", i, line);
      return -1;
    }

    int j;
    for (j = 0;; j++){
      if (j == maxsize){
        fprintf(stderr, "Error: Hostname in dist file not matched with hostname in this MPI instance\n");
        fprintf(stderr, "Line %d: %s\n", i, line);
        return -1;
      }
      if (check[j] == 0 && strcmp(hostname, confs[j].hostname) == 0 && confs[j].rank_intra == rank_intra) {
        check[j] = 1;
        dist->parts[j].d = d;
        dist->parts[j].t = t;
        dist->parts[j].i = _i;
        dist->parts[j].x = x;
        break;
      }
    }
  }
  // Now zero remaining nodes
  for (i = 0; i < maxsize; i++){
    if (check[i] == 0) {
      check[i] = 1;
      dist->parts[i].d = 0;
      dist->parts[i].t = 0;
      dist->parts[i].i = 0;
      dist->parts[i].x = 0;
    }
  }
  free(check);
  return 0;	
}

/*! Writes fupermod_part to file*/
int fupermod_dist_write(FILE* stream, fupermod_dist* dist, fupermod_library* library, fupermod_process_conf* confs) {
  fprintf(stream, "#size %d\n", dist->size);
  fprintf(stream, "#D %lld\n", dist->D);
  fprintf(stream, "#host\trank_intra\td\t%12s\ti\tx\t%12s\t%12s\n", "time", "complexity", "speed");
  int i;
  for (i = 0; i < dist->size; i++){
    long long int d = dist->parts[i].d;
    double t = dist->parts[i].t;
    double c = ((fupermod_kernel*)(library->kernel))->complexity(d, NULL);
    double s = t > 0 ? c / t : 0;
    fprintf(stream, "%s\t%d\t%lld\t%le\t%d\t%lld\t%le\t%le\n", confs[i].hostname, confs[i].rank_intra, d, 
        t, dist->parts[i].i, dist->parts[i].x, c, s);
  }
  return 0;
}

long long int fupermod_dist_total(fupermod_dist* dist) {
  long long D = 0;
  int i;
  for (i = 0; i < dist->size; i++)
    D += dist->parts[i].d;
  return D;
}

int fupermod_dist_test_total(fupermod_dist* dist) {
  return dist->D == fupermod_dist_total(dist);
}

int fupermod_dist_test_times(fupermod_point* points, int size, double eps) {
  double err = 0;
  int i;
  for (i = 0; i < size; i++) {
    int j;
    for (j = 0; j < size; j++){
      double diff = fabs((points[i].t - points[j].t) / points[i].t);
      if (diff > err)
        err = diff;
    }
  }
  return err < eps;
}

int fupermod_dist_scatter(fupermod_dist* dist, long long int* d, int root, MPI_Comm comm) {
  // compare distribution and communication sizes
  int err = 0;
  int rank;
  MPI_Comm_rank(comm, &rank);
  int size;
  MPI_Comm_size(comm, &size);
  /*	if (rank == root)
      err = (dist->size != size);
      MPI_Bcast(&err, 1, MPI_INT, root, comm);
      if (err) {
      if (rank == root)
      fprintf(stderr, "%s: distribution and communicator sizes mismatch\n", __func__);
      return -1;
      }
      */
  // scatter dist->part[i].d
  int* counts = (int*)malloc(sizeof(int) * size);
  int* displs = (int*)malloc(sizeof(int) * size);
  int i;
  for (i = 0; i < size; i++) {
    counts[i] = sizeof(long long int);
    displs[i] = sizeof(fupermod_part) * i;
  }
  err = MPI_Scatterv(rank == root ? &dist->parts[0].d : NULL, counts, displs, MPI_CHAR, d, 1, MPI_LONG_LONG_INT, root, comm);
  free(counts);
  free(displs);
  return err;
}

fupermod_partitioner partitioner_switch(int algorithm, int smooth, int benchsize){
  fupermod_partitioner p = {0,0,0,0};
  switch (algorithm) {
    case 0: // Homogeneous distribution
      {
        p.partition = fupermod_partition_homogeneous;
        p.model_alloc = fupermod_model_constant_alloc;
        p.model_free = fupermod_model_constant_free;
        p.arg = benchsize;
      }
      break;
    case 1: // Constant performance model 
      {
        p.partition = fupermod_partition_basic;
        p.model_alloc = fupermod_model_constant_alloc;
        p.model_free = fupermod_model_constant_free;
        p.arg = smooth;
      }
      break;
    case 3: // FPM Geometric partitioning
      {
        p.partition = fupermod_partition_geometric;
        p.model_alloc = fupermod_model_piecewise_alloc;
        p.model_free = fupermod_model_piecewise_free;
        p.arg = smooth;
      }
      break;
    case 4: // FPM multiroot solver
      {
        p.partition = fupermod_partition_multiroot;
        p.model_alloc = fupermod_model_interp_alloc;
        p.model_free = fupermod_model_interp_free;
        p.arg = smooth;
      }
      break;
  }
  return p;
}

