#ifndef FUPERMOD_PARTITION_H_
#define FUPERMOD_PARTITION_H_

#include "fupermod_model.h"

#ifdef __cplusplus
extern "C" {
#endif
  /*!
   * \defgroup fupermod_partition Data partitioning
   * \{
   */

  /*! Partitioning information of each processor */
  typedef struct fupermod_part {
    /*! Partitioned problem size */
    long long int d;
    /*! Time */
    double t;
    /*! Location of processor (ordering) */
    int i;
    /*! Position of partitioned rectangle (origin) */
    long long int x;
  } fupermod_part;

  /*! Data distribution */
  typedef struct fupermod_dist {
    /*! Total problem size */
    long long int D;
    /*! Number of processors */
    int size;
    /*! Partitioning information of each processor */
    fupermod_part* parts;
  } fupermod_dist;

  /*! Allocates an equal distribution */
  fupermod_dist* fupermod_dist_alloc(long long int D, int size);

  /*! Frees the distribution */
  void fupermod_dist_free(fupermod_dist* dist);

  /*! Reads distribution from file*/
int fupermod_dist_read(FILE* stream, fupermod_dist* dist, fupermod_process_conf* confs, int maxsize);

  /*! Writes distribution to file*/
  int fupermod_dist_write(FILE* stream, fupermod_dist* dist, fupermod_library* library, fupermod_process_conf* confs);

  /*!
   * Returns the total problem size
   * \param dist distribution
   * \return total problem size
   */
  long long int fupermod_dist_total(fupermod_dist* dist);

  /*!
   * Tests the problem sizes in the distribution
   * \param dist distribution
   * \return true if sum of the parts is equal to the total problem size
   */
  int fupermod_dist_test_total(fupermod_dist* dist);

  /*!
   * Tests the times in the distribution
   * \param points array of size size.
   * \param size number of proc.
   * \param eps maximum relative difference between times
   * \return true if the relative difference between times is less than eps
   */
  int fupermod_dist_test_times(fupermod_point* points, int size, double eps);

  /*!
   * Scatters the distribution
   * \param dist distribution (significant only at root)
   * \param d problem size
   * \param root root
   * \param comm MPI communicator
   */
  int fupermod_dist_scatter(fupermod_dist* dist, long long int* d, int root, MPI_Comm comm);

  /*!
   * Partitioning algorithm. All partitioning algorithms must implement this interface.
   * \param dist optimal data distribution (input/output parameter)
   * \param size number of processors
   * \param models time functions
   */
  typedef int (*fupermod_partition)(fupermod_dist* dist, int size, fupermod_model** models);

  /*!
   * Returns an homogeneous distribution (for experimental comparison)
   */
  int fupermod_partition_homogeneous(fupermod_dist* dist, int size, fupermod_model** models);
  /*!
   * Proportional data partitioning. Uses models of any type
   * (with the constant models provided by \ref fupermod_model_constant_alloc,
   * this function will perform the traditional data partitioning algorithm)
   */
  int fupermod_partition_basic(fupermod_dist* dist, int size, fupermod_model** models);

  /*!
   * Geometrical data partitioning. Requires models providing the particular shape of the speed
   * (for example, the piecewise model provided by \ref fupermod_model_piecewise_alloc)
   */
  int fupermod_partition_geometric(fupermod_dist* dist, int size, fupermod_model** models);

  /*!
   * Finds the optimal partition.
   * Solves a system of non-linear equations \f$ f(\mathbf{x}) = 0 \f$, where \f$ f(\mathbf{x}) = \begin{cases}
   * n - \displaystyle\sum\limits_{i=1}^{p}x_i &\\
   * \displaystyle\frac{x_i}{s_i(x_i)} - \displaystyle\frac{x_1}{s_1(x_1)} & 2 \le i \le p
   * \end{cases} \f$ by the least squares minimization of the residual of the system.
   * Uses the HYBRJ algorithm (a modified version of Powell's Hybrid method) implemented in GSL.
   * The iterative solution starts with \f$ \mathbf{x_0} \f$ given by the dist argument.
   * dist is the resulting distribution.
   */
  int fupermod_partition_multiroot(fupermod_dist* dist, int size, fupermod_model** models);

#ifdef HAVE_LIBNLOPT_CXX
  /*! Finds the optimal partition. */
  int fupermod_partition_nlopt(fupermod_dist* dist, int size, fupermod_model** models);

  /*! Finds the optimal partition. */
  int fupermod_partition_nlopt_minmax(fupermod_dist* dist, int size, fupermod_model** models);
#endif

  /*! Provides pointers to partition and model functions. */
  typedef struct fupermod_partitioner {
    /*! Partition function */
    fupermod_partition partition;
    /*! Allocate model function */
    fupermod_model_alloc model_alloc;
    /*! Free model function */
    fupermod_model_free model_free;
    /*! todo:*/
    int arg;
  } fupermod_partitioner;

  /*! Sets up pointers to appropriate model and partition functions 
   * /param algorithm Number between 0 4.
   *	0: Homogeneous distribution
   *	1: Constant performance model 1
   *	2: Constant performance model 2
   *	3: Geometric algorithm FFPM
   *	4: FPM multiroot solver
   */
  fupermod_partitioner partitioner_switch(int algorithm, int smooth, int benchsize);



  /*!
   * \}
   */
#ifdef __cplusplus
}
#endif

#endif /*FUPERMOD_PARTITION_H_*/
