#include "fupermod.h"
#include "fupermod_model_piecewise.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define PRINT_ANGLES 0

/*!
 * Finds the intersection of a line from the origin with the function. returns problem size (x).
 */
double intersect_line_with_piecewise(fupermod_model_piecewise* model, double m);

/*!
 * Recursive function to find iterval of intersection of line from origin with function,
 * returns position in array.
 */
int divide_interval_slope(fupermod_model_piecewise* model, int above, int below, double m);

int fupermod_partition_geometric(fupermod_dist* dist, int size, fupermod_model** models_) {
  fupermod_model_piecewise** models = (fupermod_model_piecewise**)models_;
  if (dist->D < size) {
    int i;
    for (i = 0; i < size; i++) {
      dist->parts[i].d = i < dist->D ? 1 : 0;
      dist->parts[i].t = models_[i]->complexity(dist->parts[i].d, models_[i]) / models[i]->s[0];
    }
    return 0;
  }
  double max = 0;
  double min = 0;
  int i;
  for (i = 0; i < size; i++) {
    double speed = models_[i]->complexity(dist->D / size, models_[i]) / models_[i]->t(models_[i], dist->D / size);
    if (i == 0)
      max = min = speed;
    else if (speed > max)
      max = speed;
    else if (speed < min)
      min = speed;
  }
  double L_m = min / ((double)dist->D / size);
  double U_m = max / ((double)dist->D / size);
  double* L_x = (double *)malloc(sizeof(double) * size);
  double* U_x = (double *)malloc(sizeof(double) * size);
  double* M_x = (double *)malloc(sizeof(double) * size);
  double continue_ = 0;

  FILE* angles;
  if (PRINT_ANGLES){
    angles = fopen("angles.dat", "w");
    fprintf(angles,"0\t0\n%lld\t%le\n\n\n", dist->D, L_m * dist->D);
    fprintf(angles,"0\t0\n%lld\t%le\n\n\n", dist->D, U_m * dist->D);
  }
  double sum_x_max = 0, sum_x_min = 0;
  for (i = 0; i < size; i++) {
    sum_x_max += L_x[i] = intersect_line_with_piecewise(models[i], L_m);
    sum_x_min += U_x[i] = intersect_line_with_piecewise(models[i], U_m);
    if (L_x[i] - U_x[i] >= 1)
      continue_ = 1;
  }
  if (fupermod_verbose > 1) fprintf(stdout, "sum max:%lf sum min:%lf\n", sum_x_max, sum_x_min);
  if (size > 1 && (sum_x_max < dist->D || sum_x_min > dist->D)) {
    fprintf(stderr, "Error, partitioning will not converge\n"
        "sum max:%lf sum min:%lf D:%lld\n", sum_x_max, sum_x_min, dist->D);
    return -1;
  }

  while(continue_) {
    continue_ = 0;
    double N1 = 1 / sqrt(L_m * L_m + 1);
    double N2 = 1 / sqrt(U_m * U_m + 1);
    double M_m = (N1 * L_m + N2 * U_m) / (N1 + N2);

    if (PRINT_ANGLES){
      fprintf(angles,"0\t0\n%lld\t%le\n\n\n", dist->D, M_m * dist->D);
    }

    double sum_x = 0;
    for (i = 0; i < size; i++) {
      sum_x += M_x[i] = intersect_line_with_piecewise(models[i], M_m);
    }
    if (fupermod_verbose > 1) {
      fprintf(stdout, "M_m: %lf sum:%lf, Dist: ", M_m, sum_x);
      for (i = 0; i < size; i++) 
        fprintf(stdout, "%1.2f ", M_x[i]);
      fprintf(stdout, "\n");
    }
    if (sum_x <= dist->D) {
      double* tmp = U_x;
      U_x = M_x;
      M_x = tmp;
      U_m = M_m;
    } else {
      double* tmp = L_x;
      L_x = M_x;
      M_x = tmp;
      L_m = M_m;
    }
    for (i = 0; i < size; i++) {
      if (L_x[i] - U_x[i] >= 1)
        continue_ = 1;
    }
    if ((U_m - L_m) / U_m < 1e-15)
      continue_ = 0;
  }
  long long int sum_d = 0;
  for (i = 0; i < size; i++) {
    sum_d += dist->parts[i].d = (int)U_x[i];
  }

  if (PRINT_ANGLES){
    fclose(angles);
  }
  free(L_x); L_x = NULL;
  free(U_x); L_x = NULL;
  free(M_x); L_x = NULL;

  //Hack to give a processor at least 1 unit.
  //{
  int extra = 0;
  for (i = 0; i < size; i++) {
    if (dist->parts[i].d == 0) {
      dist->parts[i].d = 1;
      extra++;
      fprintf(stderr, "Warning: Was partitioning P%d zero, now rounding up to 1\n", i);
    }
  }
  i = 0;
  while (extra > 0) {
    if (dist->parts[i].d > 3) {
      dist->parts[i].d--;
      extra--;
      fprintf(stderr, "Warning: Removing one unit from P%d now:%lld\n", i, dist->parts[i].d);
    }
    i++;
    if (i == size)
      i = 0;
  }
  sum_d = 0;
  for (i = 0; i < size; i++) {
    sum_d += dist->parts[i].d;
  }
  //}

  sum_d = dist->D - sum_d;
  if (sum_d > size * 3) // Print warning if distributing too many remainders relative to size.
    fprintf(stderr, "Warning, distributing more remainders then P (r:%lld P:%d, D:%lld)\n", sum_d, size, dist->D);

  if (fupermod_verbose > 1) {
    printf("d: ");
    for (i = 0; i < size; i++)
      printf("%lld ", dist->parts[i].d);
    printf(" , remainder: %lld\n", sum_d);
  }

  i = 0;
  while (sum_d > 0){
    dist->parts[i++].d++;
    sum_d--;
    if (i == size) {
      i = 0;
    }
  }

  // result
  for (i = 0; i < size; i++)
    dist->parts[i].t = models_[i]->t(models_[i], dist->parts[i].d);

  return 0;
}

double intersect_line_with_piecewise(fupermod_model_piecewise* model, double m){
  int count = model->count;
  // check fist point is above and last below
  double r = model->s[0] / model->d[0];
  if (r <= m) { // line is steeper then all points. 
    return model->s[0] / m;
  } 
  r = model->s[count-1] / model->d[count - 1];
  if (r >= m) { // line passes under all points 
    return model->s[count - 1] / m;
  } 
  // intercept is between 0 and count-1.
  int pos = divide_interval_slope(model, 0, count - 1, m);
  double m2 = (model->s[pos+1] - model->s[pos]) / (model->d[pos+1] - model->d[pos]);
  double b = model->s[pos] - m2 * model->d[pos];

  return b / (m - m2);
}

int divide_interval_slope(fupermod_model_piecewise* model, int above, int below, double m){
  int mid = (above + below) / 2;
  if (mid == above)
    return above;
  double r = model->s[mid] / model->d[mid];

  if (r > m) {
    above = divide_interval_slope(model, mid, below, m);
  } else if (r < m){
    above = divide_interval_slope(model, above, mid, m);
  } else if (r == m){
    return mid;
  }
  return above;
}
