#include "fupermod.h"
#include <stdio.h>
#include <gsl/gsl_multiroots.h>
#include <gsl/gsl_sort_vector.h>

/*! Parameters for system of equations (for GSL) */
typedef struct fupermod_params {
  /*! Problem size to partition */
  long long int D;
  /*! Number of processors */
  int size;
  /*! Models */
  fupermod_model** models;
} fupermod_params;

/*! Copies and fixes the values of the variables */
double* fupermod_x2d(const gsl_vector* x, fupermod_params* params) {
  double* d = (double*)malloc(sizeof(double) * params->size);
  gsl_vector_view dst = gsl_vector_view_array(d, params->size);
  gsl_vector_const_view src = gsl_vector_const_subvector(x, 0, params->size);
  gsl_vector_memcpy(&dst.vector, &src.vector);
  return d;
}

/*! System of equations (for GSL) */
int fupermod_f(const gsl_vector* x, void* params_, gsl_vector* f) {
  fupermod_params* params = (fupermod_params*)params_;
  // variables
  double* d = fupermod_x2d(x, params);
  if (fupermod_verbose > 1)
    fprintf(stderr, "* Using %s here\n", __func__);

  // D = d0 + ... + d_n-1 (1 equation)
  double y = params->D;
  int i;
  for (i = 0; i < params->size; i++) {
    if (d[i] < 0) {
      // d[i] negative, so make y bigger.
      y -= 3 * params->D * d[i];
      fprintf(stderr, "%s warning: d%d less than 0, %f -> 0\n", __func__, i, d[i]);
      d[i] = 0;
    } else {
      y -= d[i];
    }
    if (d[i] > 2 * params->D) {
      fprintf(stderr, "%s warning: d%d greater than D, %f -> %lld\n", __func__, i, d[i], 2 * params->D);
      d[i] = 2 * params->D;
    }
  }
  if (fupermod_verbose > 1) {
    fprintf(stderr, "D - Sum(d_i): %le\n", y);
    int i;
    for (i = 0; i < params->size; i++) {
      fprintf(stderr, "%1.2le ", d[i]);
    }
    fprintf(stderr, "\n");
  }
  gsl_vector_set(f, 0, y);

  // t0 = ti (size - 1 eqiations)
  double t0 = params->models[0]->t(params->models[0], d[0]);
  if (fupermod_verbose > 1)
    fprintf(stderr, "solver times: %1.2le ", t0);
  for (i = 1; i < params->size; i++) {
    double ti = params->models[i]->t(params->models[i], d[i]);
    if (fupermod_verbose > 1)
      fprintf(stderr, "%1.2le ", ti);
    gsl_vector_set(f, i, t0 - ti);
  }
  if (fupermod_verbose > 1) {
    fprintf(stderr, "\n");
  }
  free(d);
  return GSL_SUCCESS;
}

/*! Jacobian matrix for system of equations (for GSL) */
int fupermod_df(const gsl_vector* x, void* params_, gsl_matrix* J) {
  if (fupermod_verbose > 1) {
    fprintf(stderr, "= Using %s here\n", __func__);
  }
  fupermod_params* params = (fupermod_params*)params_;
  // variables
  double* d = fupermod_x2d(x, params);
  int i;
  for (i = 0; i < params->size; i++) {
    if (d[i] < 0) {
      fprintf(stderr, "%s warning: d%d less than 0, %f -> 0\n", __func__, i, d[i]);
      d[i] = 0;
    }
    if (d[i] > 2 * params->D) {
      fprintf(stderr, "%s warning: d%d greater than D, %f -> %lld\n", __func__, i, d[i], 2 * params->D);
      d[i] = 2 * params->D;
    }
  }
  // 1 equation
  int j;
  for (j = 0; j < params->size; j++)
    gsl_matrix_set(J, 0, j, -1);

  // size - 1 equations
  double dt0 = params->models[0]->dt(params->models[0], d[0]);
  for (i = 1; i < params->size; i++) {
    gsl_matrix_set(J, i, 0, dt0);
    int j;
    for (j = 1; j < params->size; j++)
      if (j == i) {
        double dti = params->models[i]->dt(params->models[i], d[i]);
        gsl_matrix_set(J, i, j, -dti);
      } else
        // J(i, j) = 0
        gsl_matrix_set(J, i, j, 0);
  }

  free(d);
  return GSL_SUCCESS;
}

/*! Invokes \ref fupermod_f and \ref fupermod_df (for GSL) */
int fupermod_fdf(const gsl_vector* x, void* params, gsl_vector* f, gsl_matrix* J) {
  fupermod_f(x, params, f);
  fupermod_df(x, params, J);
  return GSL_SUCCESS;
}

int fupermod_partition_multiroot(fupermod_dist* dist, int size, fupermod_model** models) {
  // solver and x0
  fupermod_params params = { dist->D, size, models };
  gsl_multiroot_function_fdf fdf = { fupermod_f, fupermod_df, fupermod_fdf, size, &params };
  gsl_vector* x0 = gsl_vector_alloc(size);
  int i;
  for (i = 0; i < size; i++) {
    gsl_vector_set(x0, i, dist->D / size);
  }
  gsl_multiroot_fdfsolver* solver = gsl_multiroot_fdfsolver_alloc(gsl_multiroot_fdfsolver_hybridsj, size);
  gsl_multiroot_fdfsolver_set(solver, &fdf, x0);
  gsl_vector_free(x0);

  // iterative solution
  const int iters = 100;
  const double eps = 1e-6;
  int iter = 0;
  int status = GSL_CONTINUE;
  while (iter < iters && status == GSL_CONTINUE) {
    if (fupermod_verbose > 1) {
      fprintf(stderr, "%s: iteration %d\n", __func__, iter);
    }
    int err = gsl_multiroot_fdfsolver_iterate(solver);
    if (err) {
      if (fupermod_verbose > 1) {
        fprintf(stderr, "%s has hit an error: status = %s\n", __func__, gsl_strerror(err));
      }
      break;
    }
    if (fupermod_verbose > 1) {
      for (i = 0; i < size; i++) {
        fprintf(stderr, "%f ", gsl_vector_get(solver->x, i));
      }
      fprintf(stderr, "\n");
    }
    status = gsl_multiroot_test_residual(solver->f, eps);
    iter++;
  }
  if (fupermod_verbose > 0) {
    fprintf(stderr, "%s: status: %s iter: %d\n", __func__, gsl_strerror(status), iter);
  }
  // fix the solution and distribute the remainders
  double* d = fupermod_x2d(solver->x, &params);
  long long int D_sum = 0;
  for (i = 0; i < size; i++) {
    if (d[i] < 0) {
      fprintf(stderr, "Error: d[%d] < 0!", i);
      return -1;
    }
    D_sum += d[i] = round(d[i]);
  }
  gsl_permutation* p = gsl_permutation_alloc(size);
  gsl_sort_vector_index(p, solver->x);
  // remainder D - D_sum < size if it worked.
  if (abs(dist->D - D_sum) > size) {
    fprintf(stderr, "Root finder failed, Sum d[i] != (D +/- size)\n");
    return -1;
  }
  //-------------------------------------------------------------------------------------
  // add extra to the biggest or subtract from smallest
  i = size - 1;
  while (D_sum < dist->D) {
    d[p->data[i--]]++;
    D_sum++;
  }
  i = 0;
  while (D_sum > dist->D) {
    if (d[p->data[i]] >= 1) {
      d[p->data[i]]--;
      D_sum--;
    }
    i++;
  }
  //--------------------------------------------------------------------------------------
  gsl_permutation_free(p);
  gsl_multiroot_fdfsolver_free(solver);

  D_sum = 0;
  for (i = 0; i < size; i++) {
    D_sum += d[i];
    if (d[i] < 0 || d[i] > dist->D) {
      fprintf(stderr, "Error: d[i]=%d out of range [0,%lld]\n", (int)d[i], dist->D);
      return -1;
    }
  }
  if (D_sum != dist->D) {
    fprintf(stderr, "Error: Sum d[i] != D, sum:%lld D:%lld\n", D_sum, dist->D);
    return -1;
  }
  // result
  dist->size = size;
  for (i = 0; i < size; i++) {
    dist->parts[i].d = d[i];
    dist->parts[i].t = models[i]->t(models[i], d[i]);
  }
  free(d);
  return 0;
}
