#include "fupermod.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <nlopt.h>

/*! Data for objective function */
typedef struct f_data {
  fupermod_model** models;
} f_data;

/*! Objective function */
double objective_f_minmax(unsigned dim, const double* x, double* grad, void* data) {
  fupermod_model** models = ((f_data*)data)->models;
  double max = 0;
  double min = HUGE_VAL;
  int maxi = 0;
  int mini = 0;
  unsigned i;
  for (i = 0; i < dim; i++) {
    double time = models[i]->t(models[i], x[i]);
    if (time > max){
      max = time;
      maxi = i;
    }
    if (time < min){
      min = time;
      mini = i;
    }
  }
  //fprintf(stderr, "Max:%le P%d\n", max, maxi);

  if (grad) {
    //fprintf(stderr, "Looking for obj grad!\n");
    for (i = 0; i < dim; i++) {
      grad[i] = 0;
    }
    grad[maxi] = models[maxi]->dt(models[maxi], x[maxi]);
    grad[mini] = -models[mini]->dt(models[mini], x[mini]);
    //		fprintf(stderr, "d:%d dt[%d]:%le\n", (int)x[maxi], maxi, grad[maxi]);
  }
  {
    fprintf(stderr, "D\t");
    for (i = 0; i < dim; i++) {
      fprintf(stderr, "%d\t", (int)x[i]);
    }
    fprintf(stderr, "\nT\t");
    for (i = 0; i < dim; i++) {
      fprintf(stderr, "%1.1le\t", models[i]->t(models[i], x[i]));
    }
    if (grad) {
      fprintf(stderr, "\nG\t");
      for (i = 0; i < dim; i++) {
        fprintf(stderr, "%1.1le\t", grad[i]);
      }
    }
    fprintf(stderr, "\n");

  }
  return max - min;
}

/*! Data for constraints */
typedef struct fc_data {
  double D;
} fc_data;

/*! Constraints */
double constraints_minmax(unsigned dim, const double* x, double* grad, void* data) {
  double D = ((fc_data*)data)->D;
  if (grad) {
    //fprintf(stderr, "Looking for constraint grad!\n");
    unsigned i;
    for (i = 0; i < dim; i++){
      grad[i] = 1;
    }
  }
  unsigned i;
  double sum = -D;
  for (i = 0; i < dim; i++)
    sum += x[i];

  //fprintf(stderr, "%lf %lf\n", D, sum);
  return sum;
}

int fupermod_partition_nlopt_minmax(fupermod_dist* dist, int size, fupermod_model** models) {
  //Also try NLOPT_GD_STOGO, NLOPT_LD_SLSQP, NLOPT_GN_ISRES
  //nlopt_opt opt = nlopt_create(NLOPT_AUGLAG, size); /* algorithm and dimensionality */
  nlopt_opt opt = nlopt_create(NLOPT_LD_SLSQP, size); /* algorithm and dimensionality */
  nlopt_set_lower_bounds1(opt, 0);
  nlopt_set_upper_bounds1(opt, dist->D);

  f_data data;
  data.models = models;
  nlopt_set_min_objective(opt, objective_f_minmax, &data);

  fc_data c_data;
  c_data.D = dist->D;
  nlopt_add_equality_constraint(opt, constraints_minmax, &c_data, 100); // tolerence of +-1

  nlopt_set_stopval(opt,1e-3);
  //	nlopt_set_xtol_rel(opt, 1e-7);
  //	nlopt_set_xtol_abs1(opt, 1e-7);
  //	nlopt_set_maxtime(opt, 40); // timeout - help debug never ending loop.

  double* x = (double*)malloc(sizeof(double) * size); /* some initial guess */
  int i;
  for (i = 0; i < size; i++)
    x[i] = dist->D / size;

  //nlopt_opt local_opt = nlopt_create(NLOPT_GD_STOGO, size); /* algorithm and dimensionality */
  //nlopt_opt local_opt = nlopt_create(NLOPT_LD_SLSQP, size); /* algorithm and dimensionality */
  //nlopt_set_xtol_rel(local_opt, 1e-7);
  //nlopt_set_xtol_abs1(local_opt, 1e-7);
  nlopt_set_maxtime(opt, 40); // timeout - help debug never ending loop.
  //nlopt_set_local_optimizer(opt, local_opt);

  double min_result; /* the minimum objective value, upon return */
  int rc;
  if ((rc = nlopt_optimize(opt, x, &min_result)) < 0) {
    printf("nlopt failed! error:%d\n", rc);
  }
  else {
    printf("found minimum result of %le\n",min_result);
    printf("size:%d d: ", size);
    for (i = 0; i < size; i++)
      printf("%1.1lf\t", x[i]);
    printf("\n");
  }

  int* d = (int*)malloc(sizeof(int) * size);
  for (i = 0; i < size; i++)
    d[i] = (int)x[i];
  free(x);
  nlopt_destroy(opt);

  long long int D_sum = 0;
  for (i = 0; i < size; i++) {
    if (d[i] < 0) {
      fprintf(stderr, "Error: d[%d] < 0!", i);
      return -1;
    }
    D_sum += d[i];
  }
  // add extra to the biggest or subtract from smallest
  i = 0;
  while (D_sum < dist->D) {
    d[i++]++;
    D_sum++;
    if (i == size)
      i = 0;
  }
  i = 0;
  while (D_sum > dist->D) {
    if (d[i] >= 1) {
      d[i]--;
      D_sum--;
    }
    i++;
    if (i == size)
      i = 0;
  }

  D_sum = 0;
  for (i = 0; i < size; i++) {
    D_sum += d[i];
    if (d[i] < 0 || d[i] > dist->D) {
      fprintf(stderr, "Error: d[i]=%d out of range [0,%lld]\n", (int)d[i], dist->D);
      return -1;
    }
  }
  if (D_sum != dist->D) {
    fprintf(stderr, "Error: Sum d[i] != D, sum:%lld D:%lld\n", D_sum, dist->D);
    return -1;
  }

  // update distribution
  for (i = 0; i < size; i++) {
    dist->parts[i].d = d[i];
    dist->parts[i].t = models[i]->t(models[i], d[i]);
  }
  free(d);
  return 0;
}
