#ifndef FUPERMOD_2D_KERNEL_H_
#define FUPERMOD_2D_KERNEL_H_

#ifdef HAVE_CONFIG_H
#include "fupermod/fupermod_kernel.h"
#else
#include "fupermod_kernel.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif
  /*!
   * \defgroup fupermod_2d_kernel_group 2D computational kernel of a routine
   * A 2D routine is a standalone parallel application supplied with a shared library
   * implementing its computational kernel, \ref fupermod_2d_kernel.
   * The fupermod tools load the shared library and search for the \ref FUPERMOD_LOADER_INSTANCE symbol,
   * the instance of the \ref fupermod_loader data structure.
   * \see \ref template_kernel
   * \{
   */

  /*!
   * Computational complexity of the kernel
   * \param m problem size
   * \param n problem size
   * \return complexity
   */
  typedef double (*fupermod_2d_complexity)(long long int m, long long int n, void* param);

  /*! 2D computational kernel of a routine */
  typedef struct fupermod_2d_kernel {
    /*! Complexity from two parameters*/
    fupermod_2d_complexity complexity_2d;

    /*!
     * Initializes the kernel. Allocates data for benchmarking, which are stored in global variables.
     * \param comm MPI communicator
     * \param m problem size
     * \param n problem size
     * \param params kernel parameters (if used, should be allocated here)
     * \return return code
     */
    int (*initialize)(long long int m, long long int n, void** params);

    /*!
     * Executes the kernel. Uses the global variables of the kernel.
     * Kernel can be executed in a separate thread to enable termination in case of long run.
     * \param mutex a mutex to protect some resources:
     * \code
     * pthread_mutex_lock(mutex);
     * // use of resources
     * pthread_mutex_unlock(mutex);
     * \endcode
     * \return return code
     */
    int (*execute)(pthread_mutex_t* mutex, void* params);

    /*!
     * Finalizes the kernel. Free the memory for the kernel.
     * \return return code
     */
    int (*finalize)(void* params);

    /*! Complexity from one variable and second parameter comes from void param */
    fupermod_complexity complexity_1d;
  } fupermod_2d_kernel;

  /*!
   * \}
   */
#ifdef __cplusplus
}
#endif

#endif /* FUPERMOD_2D_KERNEL_H_ */
