#include "mxm_2d.h"
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <memory.h>

#define PRINT_TIMES 0
#define VERBOSE_MATRIX 0

fupermod_dgemm_workspace* fupermod_dgemm_alloc(MPI_Comm comm, FILE* stream, fupermod_process_conf* confs, fupermod_gemm* gemm) {
  fupermod_dgemm_workspace* w = (fupermod_dgemm_workspace*)malloc(sizeof(fupermod_dgemm_workspace));
  w->gemm = gemm;

  int rank;
  MPI_Comm_rank(comm, &rank);
  int size; 
  MPI_Comm_size(comm, &size);

  //TODO check the size after the read
  w->idist = mm_col_dist_read(stream, size, confs);
  if (w->idist == NULL)
    return NULL;
  int kb = w->idist->kb;

  // print total volume of communication.
  {
    int tvc = 0;
    if (rank == 0){
      int i;
      for (i = 0; i < size; i++) {
        tvc += w->idist->d[i].w;
        tvc += w->idist->d[i].h;
      }
      fprintf(stdout, "tvc: %d\n", tvc);
    }
  }

  int myw = w->idist->d[rank].w;
  int myh = w->idist->d[rank].h;
  // working space:
  w->WA = fupermod_malloc(sizeof(fupermod_float) * myh * kb * kb, confs+rank);
  w->WB = fupermod_malloc(sizeof(fupermod_float) * myw * kb * kb, confs+rank);

  if (w->WA == NULL || w->WB == NULL) {
    fprintf(stderr, "P%d Failed to allocate matrix\n"
        "P%d myw = %d, myh = %d, kb = %d\n"
        ,rank, rank, myw, myh, kb);
    fupermod_dgemm_free(w, confs+rank);
    return NULL;
  }

  return w;
}

void fupermod_dgemm_free(fupermod_dgemm_workspace* w, fupermod_process_conf* conf){
  fupermod_free(w->WA, conf);
  fupermod_free(w->WB, conf);

  mm_col_i_dist_free(w->idist);
  free(w); w = NULL;
}

// todo: should be someting like: 
/* int fupermod_dgemm(enum CBLAS_ORDER Order, enum CBLAS_TRANSPOSE TransA, enum CBLAS_TRANSPOSE TransB, const int M, const int N, const int K,
   double alpha, double *A, const int lda, double *B, const int ldb, double beta, double *C, const int ldc, MPI_Comm comm, mm_col_i_distribution idist) */
int fupermod_dgemm(fupermod_float *A, fupermod_float *B, fupermod_float *C, MPI_Comm comm, fupermod_dgemm_workspace* w) {
  // my matrix block
  int rank;
  MPI_Comm_rank(comm, &rank);
  mm_col_i_zone* d = w->idist->d;
  int myx = d[rank].x;
  int myy = d[rank].y;
  int myw = d[rank].w;
  int myh = d[rank].h;
  int myi = d[rank].i;
  int myj = d[rank].j;

  FILE* time_file = NULL;
  if(PRINT_TIMES){
    char filename[256];
    sprintf(filename, "times.mxmcol.%d.dat", rank);
    time_file = fopen(filename, "w");
    char hostname[MPI_MAX_PROCESSOR_NAME];
    int len;
    MPI_Get_processor_name(hostname, &len);
    fprintf(time_file, "# host: %s, myw:%e myh:%le myi:%d myj:%d N:%le\n", hostname, (double)myw, (double)myh, myi, myj, (double)w->idist->N);
  }

  // progress report
  struct timeval start, inter1, inter2;
  gettimeofday(&start, NULL);
  gettimeofday(&inter2, NULL);
  double total_time_dgemm = 0;

  // order ranks by columns for A-communications
  int size;
  MPI_Comm_size(comm, &size);
  int col_count = 0; // number of columns
  int* row_counts = NULL; // numbers of rows in each column
  int** ranks = NULL; // ranks by columns
  int i;
  for (i = 0; i < size; i++) {
    int col = d[i].j;
    if (col >= col_count) {
      row_counts = (int*)realloc(row_counts, sizeof(int) * (col + 1));
      ranks = (int**)realloc(ranks, sizeof(int*) * (col + 1));
      int j;
      for (j = col_count; j <= col; j++) {
        row_counts[j] = 0;
        ranks[j] = NULL;
      }
      col_count = col + 1;
    }
    int row = d[i].i;
    if (row >= row_counts[col]) {
      row_counts[col] = row + 1;
      ranks[col] = (int*)realloc(ranks[col], sizeof(int) * (row + 1));
    }
    ranks[col][row] = i;
  }
  // find prev/next in ring in my column for B-communications
  int prev = ranks[myj][(myi - 1 + row_counts[myj]) % row_counts[myj]];
  int next = ranks[myj][(myi + 1) % row_counts[myj]];

  // for non-blocking communications
  int req_count = 0;
  MPI_Request* requests = (MPI_Request*)malloc(sizeof(MPI_Request) * size);
  if (requests == NULL) {
    fprintf(stderr, "P%d Failed to allocate MPI objects\n", rank);
    return -1;
  }

  // Main k loop - moving pivrow & pivcol
  int kb = w->idist->kb;
  int k;
  for (k = 0; k < w->idist->N; k++) {
    // progress report
    if (rank == 0 && k > 0) {
      gettimeofday(&inter1, NULL);
      if ( inter1.tv_sec - inter2.tv_sec > 60) {
        double time = ((double)(inter1.tv_sec - start.tv_sec) + (inter1.tv_usec - start.tv_usec) * 1.0e-6);
        double expected = time / (double)k * w->idist->N;
        fprintf(stderr, "loop %d of %d, expected time: %le\n", k, w->idist->N, expected);
        gettimeofday(&inter2, NULL);
      }
    }

    // Find pivot column and row
    int pivcol_j = -1; // j index of processes holding pivot column
    int pivrow_proc = -1; // rank of a process holding pivot row in my column
    for (i = 0; i < size; i++) {
      if (pivcol_j < 0 && k >= d[i].x && k < d[i].x+d[i].w)
        pivcol_j = d[i].j;
      if (pivrow_proc < 0 && d[i].j == myj && k >= d[i].y && k < d[i].y+d[i].h)
        pivrow_proc = i;
    }
    if (pivcol_j < 0) {
      fprintf(stderr, "P%d: Error, Problem with pivcol\n", rank);
      return -1;
    }
    if (pivrow_proc < 0) {
      fprintf(stderr, "P%d: Error, Problem with pivrow_proc\n", rank);
      return -1;
    }

    // if I don't hold pivcol of A, recv it from the previous column
    if (myj != pivcol_j) {
      int j = (myj - 1 + col_count) % col_count; // previous column in ring
      int i;
      for (i = 0; i < row_counts[j]; i++) {
        int src = ranks[j][i];
        // if there is an overlap between my proc and src
        if (myy < d[src].y + d[src].h && myy + myh > d[src].y) {
          // where to recv to
          int start = d[src].y - myy;
          if (start < 0)
            start = 0;
          // how much to recv
          int length = d[src].y + d[src].h - myy;
          if (length > myh)
            length = myh - start;
          else
            length = length - start;
          int rc = MPI_Irecv(&w->WA[start * kb * kb], length * kb * kb, FUPERMOD_MPI_FLOAT, src, 0, comm, &requests[req_count++]);
          if (rc != MPI_SUCCESS) {
            fprintf(stderr, "Error Recving A\n");
            return -1;
          }
        }
      }
    }
    // if I don't hold pivrow of B, recv it from the previous in column
    if (rank != pivrow_proc) {//recv
      int rc = MPI_Irecv(w->WB, myw * kb * kb, FUPERMOD_MPI_FLOAT, prev, 0, comm, &requests[req_count++]);
      if (rc != MPI_SUCCESS){
        fprintf(stderr, "Error Recving B\n");
        return -1;
      }
    }
    MPI_Waitall(req_count, requests,MPI_STATUSES_IGNORE);
    req_count= 0;

    // part of A to send and multiply
    fupermod_float* A_pt = w->WA;
    int lda = kb;
    // Am I pivcol?
    if (myj == pivcol_j) {
      A_pt = &A[(k - myx) * kb]; // point to local A.
      lda = myw * kb; // lda changes.
    }
    // send pivcol of A to the next column, if it doesn't hold it
    int j = (myj + 1) % col_count; // next column in ring
    if (j != pivcol_j) {
      int i;
      for (i = 0; i < row_counts[j]; i++) {
        int src = ranks[j][i];
        if (myy < d[src].y + d[src].h && myy + myh > d[src].y) { // there is an overlap between my proc and src
          //start send from
          int start = d[src].y - myy;
          if (start < 0)
            start = 0;
          int length = d[src].y + d[src].h - myy;
          if (length > myh)
            length = myh - start;
          else
            length = length - start;
          // possibly non-contiguous data (if A_pt is A)
          MPI_Datatype fupermod_col;
          MPI_Type_vector(length * kb, kb, lda, FUPERMOD_MPI_FLOAT, &fupermod_col);
          MPI_Type_commit(&fupermod_col);
          int rc = MPI_Isend(&A_pt[start*kb*lda], 1, fupermod_col, src, 0, comm, &requests[req_count++]);
          if (rc != MPI_SUCCESS) {
            fprintf(stderr, "Error Sending A\n");
            return -1;
          }
          MPI_Type_free(&fupermod_col);
        }
      }
    }
    // part of B to send and multiply
    fupermod_float* B_pt = w->WB;
    int ldb = myw * kb;
    if (rank == pivrow_proc)
      B_pt = &B[(k - myy) * kb * kb * myw]; // Set for local dgemm
    // send pivrow of B
    if (next != pivrow_proc) {
      int rc = MPI_Isend(B_pt,  myw * kb * kb, FUPERMOD_MPI_FLOAT, next, 0, comm,&requests[req_count++]);
      if (rc != MPI_SUCCESS) {
        fprintf(stderr, "Error Sending B\n");
        return -1;
      }
    }
    MPI_Waitall(req_count, requests, MPI_STATUSES_IGNORE);

    if (VERBOSE_MATRIX && rank == 0) {
      int j;
      printf("P%d A(%d,%d):\n", rank, myh * kb, kb);
      for (i = 0; i < myh * kb; i++) {
        for (j = 0; j < kb; j++) {
          printf("  %3.1f", A_pt[i * lda + j]);
        }
        printf("\n");
      }
      printf("P%d B(%d,%d):\n", rank, myh * kb, kb);
      for (i = 0; i < kb; i++) {
        for (j = 0; j < ldb; j++) {
          printf("  %3.1f", B_pt[i * ldb + j]);
        }
        printf("\n");
      }
    }
    struct timeval start_dgemm, end_dgemm;
    gettimeofday(&start_dgemm, NULL);
    fupermod_gemm_execute(w->gemm, CblasRowMajor, CblasNoTrans, CblasNoTrans, myh * kb, myw * kb, kb, 1, A_pt, lda, B_pt, ldb, 1, C, myw * kb);
    gettimeofday(&end_dgemm, NULL);

    double time = ((double)(end_dgemm.tv_sec - start_dgemm.tv_sec) + (end_dgemm.tv_usec - start_dgemm.tv_usec) * 1.0e-6);
    total_time_dgemm += time;
    if (PRINT_TIMES)
      fprintf(time_file, "%le\n", time);

    if (VERBOSE_MATRIX) {
      if (rank == 0) {
        printf("P%d C(%d,%d):\n", rank, myh * kb, myw * kb);
        int j;
        for (i = 0; i < myw * kb; i++) {
          for (j = 0; j < myh * kb; j++) {
            printf("  %3.3f", C[i * myw * kb + j]);
          }
          printf("\n");
        }
      }
    }
  }// end k loop

  struct timeval end;
  gettimeofday(&end, NULL);
  double total_time = ((double)(end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec) * 1.0e-6);
  double total_comm = total_time - total_time_dgemm;
  if(PRINT_TIMES){
    fprintf(time_file, "#P%d: total: %le dgemm: %le comm: %le ratio dgemm/comm : %4.2lf\n", rank, total_time, total_time_dgemm, total_comm, total_time_dgemm / total_comm);
    fclose(time_file);
  }
  fprintf(stdout, "#P%d: total: %le dgemm: %le comm: %le ratio dgemm/comm : %4.2lf\n", rank, total_time, total_time_dgemm, total_comm, total_time_dgemm / total_comm);
#ifdef GREEN_COMPUTING
  FILE* green_file = NULL;
  char green_fname[256];
  sprintf(green_fname, "green.mxmcol.%d.%d.dat", rank, w->idist->N);
  green_file = fopen(green_fname, "w");
  if(green_file !=NULL) {
    char hostname[MPI_MAX_PROCESSOR_NAME];
    int len;
    MPI_Get_processor_name(hostname, &len);
    fprintf(green_file, "# host: %s, myw:%e myh:%le myi:%d myj:%d N:%le\n", hostname, (double)myw, (double)myh, myi, myj, (double)w->idist->N);
    double total_time = ((double)(end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec) * 1.0e-6);
    double total_comm = total_time - total_time_dgemm;
    fprintf(green_file, "#P%d: total: %le dgemm: %le comm: %le ratio dgemm/comm : %4.2lf start : %lf end : %lf\n", rank, total_time, total_time_dgemm, total_comm, total_time_dgemm / total_comm, ((double)(start.tv_sec) + start.tv_usec * 1.0e-6), ((double)(end.tv_sec) + end.tv_usec * 1.0e-6));
    fclose(green_file);
  } else {
    fprintf(stderr, "Unable to open file for writing green computing data.\n");
  }
#endif

  free(requests);
  free(row_counts);
  for (i = 0; i < col_count; i++)
    free(ranks[i]);
  free(ranks);

  return 0;
}
