#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <getopt.h>
#include "datapart_2d/mm_col_distr.h"
#include "fupermod/fupermod_partition.h"
#include "fupermod/fupermod_conf.h"
#include <gsl/gsl_sort_int.h>

int convert_col_to_car(FILE* fmachine, FILE* fpartition, mm_col_i_distrib* idist, fupermod_process_conf* confs);

int main(int argc, char** argv) 
{
  // input
  int exit = 0;
  int kb = 16;
  int algorithm = 0;
  char machine_file[256]  = "./machinefile";
  char mfile_car[256] = "./machinefile.car";
  char dist_in[256]   = "./partition.dist";
  char dist_col[256]  = "./partition.dist.col";
  char dist_car[256]  = "./partition.dist.car";
  char app_car[256]   = "./appfile";
  char exe_car[256]   = "./parallel_mxm_car";

  int opt;
  while ((opt = getopt(argc, argv, "ha:k:e:")) != -1)
    switch(opt) {
      case 'h':
        fprintf(stderr,
            "cartesian partitin \n"
            "Usage: [options]\n"
            "  -h    help\n"
            "  -a I  Algorithm chosen for partitioning  (default %d)\n"
            "  -k I  Kb  (default %d)\n"
            "  -e S  Path of mxm routine (default %s)\n",
            algorithm, kb, exe_car);
        exit = 1;
        break;
      case 'a':
        algorithm = atoi(optarg);
        break;
      case 'k':
        kb = atoi(optarg);
        break;
      case 'e':
        strcpy(exe_car, optarg);
        break;
      default:
        fprintf(stderr, "Unknown option %s\n", optarg);
        break;
    }

  if (exit) return 0;

  int size; //total number of devices.
  fupermod_process_conf* confs = fupermod_get_conf_all(machine_file, &size);

  //todo: check size == MPI_comm_size

  // Load data points
  if (algorithm == 0) { 
    /* 
     * input original machinefile and partition.dist
     * use column_based algorithm to arrange nodes and it's area
     * output new machinefile.car and partition.dist.car (and also partition.dist.col), so that procs can do bcast in rows
     */
    // Overall problem size
    int N = 0; 
    fupermod_dist* dist = fupermod_dist_alloc(0, size);
    FILE* stream = NULL;
    if ((stream = fopen(dist_in, "r")) != NULL) {
      if (fupermod_dist_read(stream, dist, confs, size) != 0) {
        fclose(stream); 
        fupermod_dist_free(dist);
        fprintf(stderr, "Error: read dist file: %s\n", dist_in);
        return -1;
      }
    } else {
      fclose(stream); 
      fupermod_dist_free(dist);
      fprintf(stderr, "Error: open dist file: %s\n", dist_in);
      return -1;
    }
    fclose(stream);
    N = sqrt(dist->D);

    // column_based partition
    double *areas = (double *) malloc (sizeof(double) * size);
    int i;
    for (i = 0; i < size; i++) {
      areas[i] = dist->parts[i].d;
    }

    mm_col_i_distrib* idist = mm_col_dist(size, N, kb, areas);
    free(areas);

    FILE* fmachine_car = fopen(mfile_car, "w");
    FILE* fpartition_col = fopen(dist_col, "w");
    FILE* fpartition_car = fopen(dist_car, "w");
    FILE* fappfile_car = fopen(app_car, "w");

    // generate partition.dist.col
    mm_col_dist_write(fpartition_col, idist, dist, confs);
    fupermod_dist_free(dist);

    // generate partition.dist.car
    int np = convert_col_to_car(fmachine_car, fpartition_car, idist, confs);
    mm_col_i_dist_free(idist);

    // generate appfile for mpirun
    fprintf(fappfile_car, "-hostfile %s -np %d %s -k%d -m %s", mfile_car, np, exe_car, kb, mfile_car);

    fclose(fpartition_col);
    fclose(fmachine_car);
    fclose(fpartition_car);
    fclose(fappfile_car);
  } else if (algorithm == 1) {
    /* 
     * input machinefile and partition.dist.col after optimized according to communication cost
     * the order of nodes in machinefile and partition.dist.col should be the same
     * without changing the position of nodes
     * output new machinefile.car and partition.dist.car, so that can do bcast in rows
     */
    mm_col_i_distrib* idist = NULL;

    FILE* stream = NULL;
    if ((stream = fopen(dist_col, "r")) != NULL) {
      if (!(idist = mm_col_dist_read(stream, size, confs))) {
        mm_col_i_dist_free(idist);
        fprintf(stderr, "Error: read partition.dist.col file: %s\n", dist_col);
        return -1;
      }
    } else {
      fclose(stream); 
      mm_col_i_dist_free(idist);
      fprintf(stderr, "Error: open partition.dist.col file: %s\n", dist_col);
      return -1;
    }

    FILE* fmachine_car = fopen(mfile_car, "w");
    FILE* fpartition_car = fopen(dist_car, "w");
    FILE* fappfile_car = fopen(app_car, "w");

    // generate machine.car and partition.dist.car
    int np = convert_col_to_car(fmachine_car, fpartition_car, idist, confs);
    mm_col_i_dist_free(idist);

    // generate appfile
    fprintf(fappfile_car, "-hostfile %s -np %d %s -k%d -m %s\n", mfile_car, np, exe_car, kb, mfile_car);

    fclose(stream);
    fclose(fmachine_car);
    fclose(fpartition_car);
    fclose(fappfile_car);
  } else {

  }
  return 0;
}

/*! 
 * input mm_col_i_distrib, 
 * output machinefile.car, partition.dist.car file, 
 * return the num of procs 
 */
int convert_col_to_car(FILE* fmachine, FILE* fpartition, mm_col_i_distrib* idist, fupermod_process_conf* confs) {
  long long int N = idist->N;
  int size = idist->nproc;
  int kb = idist->kb;
  // cartesian partition
  int *ys = (int *) malloc(sizeof(int) * 2 * size);
  int i;
  for (i = 0; i < size; i++) {
    ys[i+size] = idist->d[i].y;
  }
  // sort
  gsl_sort_int(ys + size, 1, size);
  // remove equivalent values
  ys[0] = ys[size];
  int rows = 1; // num of rows
  for (i = 1; i < size; i++) {
    if (ys[i + size - 1] != ys[i + size]) {
      ys[rows] = ys[i + size];
      rows++;
    }
  }
  ys[rows] = N;

  fpos_t pos_start;
  fgetpos (fpartition, &pos_start); // reserve the position for '#size %d'
  fprintf(fpartition, "#size %d\n"
      "#N %d\n"
      "#kb %d\n"
      "#host\ti\tj\tx\ty\tw\th\tw*h\tmem(MB)\t#\n",
      0, idist->N, kb);
  printf("\n#host\ti\tj\tx\ty\tw\th\tw*h\tmem(MB)\t#\n");

  int j, k, np = 0;
  for (i = 0; i < size; i++) {
    int y = idist->d[i].y;
    int h = idist->d[i].h;
    int istart, iend;
    for (j = 0; j <= rows; j++) {
      if (y == ys[j]) {
        istart = j;
      }
      if ((y+h) == ys[j]) {
        iend = j;
        break;
      }
    }
    int nrows = iend - istart;
    for (k = 0; k < nrows; k++) {
      np++; // num of procs
      int _i = istart + k;
      int _j = idist->d[i].j;
      int  x = idist->d[i].x;
      int  y = ys[istart + k];
      int  w = idist->d[i].w;
      int  h = ys[istart + k + 1] - ys[istart + k];

      //TODO print rank_intra to fmachine too?
      fprintf(fmachine, "%s\n", confs[i].hostname);
      fprintf(fpartition, "%s\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%zu\t%d\n",
          confs[i].hostname, _i, _j, x, y, w, h, w*h, (size_t)(8 * (w * kb * h * kb)/1024/1024), i);
      // printout
      printf("%s\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%zu\t%d\n",
          confs[i].hostname, _i, _j, x, y, w, h, w*h, (size_t)(8 * (w * kb * h * kb)/1024/1024), i);
    }
  }
  fsetpos (fpartition, &pos_start);
  fprintf(fpartition, "#size %d\n", np);
  free(ys);
  return np;
}

