#include "mxm_hybrid.h"
#include "memory_hybrid.h"
#include "dgemm_hybrid.h"
#include "datapart_2d/fupermod_hybrid.h"
#include "cblas_wrappers/fupermod_cblas.h"
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>

#define VERBOSE_MATRIX 0
#define dx(i) idist->d[i].x
#define dy(i) idist->d[i].y
#define dw(i) idist->d[i].w
#define dh(i) idist->d[i].h
#define di(k) idist->d[k].i
#define dj(k) idist->d[k].j
#define PRINT_TIMES 1

int MPI_MSG_TAG_FOR_A = 0xf1;
int MPI_MSG_TAG_FOR_B = 0xf2;

void* cuda_malloc(size_t size){
	return malloc(size);	
}

fupermod_dgemm_workspace* fupermod_dgemm_alloc_read(MPI_Comm comm, FILE* stream, fupermod_process_conf* confs, fupermod_gemm* gemm) {
	fupermod_dgemm_workspace* w = (fupermod_dgemm_workspace*)malloc(sizeof(fupermod_dgemm_workspace));
	w->gemm = gemm;

	int rank;
	MPI_Comm_rank(comm, &rank);
	int size; 
	MPI_Comm_size(comm, &size);

	// compute idist.
  //TODO check the size after the read
	w->idist = mm_col_dist_read(stream, size, confs);
	if (w->idist == NULL)
		return NULL;
	int kb = w->idist->kb;

	// print total volume of communication.
	{
		int tvc = 0;
		if (rank == 0){
			int i;
			for (i = 0; i < size; i++) {
				tvc += w->idist->d[i].w;
				tvc += w->idist->d[i].h;
			}
			fprintf(stdout, "tvc: %d\n", tvc);
		}
	}

	int myw = w->idist->d[rank].w;
	int myh = w->idist->d[rank].h;
	// working space:
	if (HYBRID_MEMORY) {
		malloc_hybrid(sizeof(double) * myh * kb * kb, &w->WA, PINNED);
		malloc_hybrid(sizeof(double) * myw * kb * kb, &w->WB, PINNED);
	} else {
		w->WA = (double*) cuda_malloc(sizeof(double) * myh * kb * kb);
		w->WB = (double*) cuda_malloc(sizeof(double) * myw * kb * kb);
	}
	if (w->WA == NULL || w->WB == NULL) {
		fprintf(stderr, "P%d Failed to allocate matrix\n"
				"P%d myw = %d, myh = %d, kb = %d\n"
				,rank, rank, myw, myh, kb);
		fupermod_dgemm_free(w);
		return NULL;
	}

	return w;
}

void fupermod_dgemm_free(fupermod_dgemm_workspace* w){
	if (HYBRID_MEMORY) {
		free_hybrid(w->WA, PINNED);
		free_hybrid(w->WB, PINNED);
	} else {
		free(w->WA);
		free(w->WB);
	}
	mm_col_i_dist_free(w->idist);
	free(w); w = NULL;
}

// todo: should be someting like: 
/* int fupermod_dgemm(enum CBLAS_ORDER Order, enum CBLAS_TRANSPOSE TransA, enum CBLAS_TRANSPOSE TransB, const int M, const int N, const int K,
	 double alpha, double *A, const int lda, double *B, const int ldb, double beta, double *C, const int ldc, MPI_Comm comm, mm_col_i_distribution idist) */
//
//  ********* Column Major dgemm *******
//
int fupermod_dgemm(double *A, double *B, double *C, MPI_Comm comm, fupermod_dgemm_workspace* w, HybridConfig config) {
	int kb = w->idist->kb;
	double* WA = w->WA;
	double* WB = w->WB;

	int rank;
	MPI_Comm_rank(comm, &rank);
	int size; 
	MPI_Comm_size(comm, &size);
	mm_col_i_distrib* idist = w->idist;
	int myx, myy, myw, myh, myi, myj;
	myx = idist->d[rank].x;
	myy = idist->d[rank].y;
	myw = idist->d[rank].w;
	myh = idist->d[rank].h;
	myi = idist->d[rank].i;
	myj = idist->d[rank].j;

	int lda, ldb, ldc; 
	lda = ldb = ldc = myh * kb;
	// hybrid fuction
	int rc;
	rc = dgemm_initialize(myh * kb, myw * kb, kb, 0, NULL, 0, NULL, 0, 0, C, ldc, config);
	if (rc != 0) {
		fprintf(stderr, "P%d: Error with dgemm_initialize\n", rank);
		MPI_Abort(comm, 70);
	}

	MPI_Comm col_comm; 
	int col_size; 
	rc = MPI_Comm_split(comm, myj, myi, &col_comm);
	if (rc != MPI_SUCCESS) {
		fprintf(stderr, "Error splitting groups\n");
		return -1;
	}
	MPI_Comm_size(col_comm, &col_size);

	// For non blocking MPI_Irecv
	MPI_Request* requestsS = (MPI_Request*)malloc(sizeof(MPI_Request)*(size-1) * 2);
	MPI_Request* requestsR = (MPI_Request*)malloc(sizeof(MPI_Request)*(size-1) * 2);
	MPI_Status* statusesS=(MPI_Status*)malloc(sizeof(MPI_Status)*(size-1));
	MPI_Status* statusesR=(MPI_Status*)malloc(sizeof(MPI_Status)*(size-1));
	if (requestsS == NULL || requestsR == NULL || statusesS == NULL || statusesR == NULL){
		fprintf(stderr, "P%d Failed to allocate MPI objects\n", rank);
		return -1;
	}
	double* tmp = (double*) malloc(sizeof(double) * 10 * size);
	if ( tmp == NULL ){
		fprintf(stderr, "P%d Failed to allocate space for overhead\n", rank);
		return -1;
	}
	free(tmp);

	//non-blocking index 
	int recv_count, send_count = 0;

	// progress report
	struct timeval startt, inter1, inter2;
	gettimeofday(&startt, NULL);
	gettimeofday(&inter2, NULL);
	double total_time_dgemm = 0;

	FILE* time_file = NULL;
	if(PRINT_TIMES){
		char filename[256];
		sprintf(filename, "times.mxmcol.%d.dat", rank);
		time_file = fopen(filename, "w");
		char hostname[MPI_MAX_PROCESSOR_NAME];
		int len;
		MPI_Get_processor_name(hostname, &len);
		fprintf(time_file, "# host: %s, myw:%e myh:%le myi:%d myj:%d N:%le\n", hostname, (double)myw, (double)myh, myi, myj, (double)w->idist->N);
	}

	// Main k loop - moving pivrow & pivcol	
	int k;
	for (k = 0; k < w->idist->N; k++) {
		// progress report
		if (rank == 0 && k > 0) {
			gettimeofday(&inter1, NULL);
			if ( inter1.tv_sec - inter2.tv_sec > 60) {
				double time = ((double)(inter1.tv_sec - startt.tv_sec) + (inter1.tv_usec - startt.tv_usec) * 1.0e-6)/ (double)k * w->idist->N;
				fprintf(stderr, "loop %d of %d, expected time: %le\n", k, w->idist->N, time);
				gettimeofday(&inter2, NULL);
			}
		}

		double* A_pt = WA;
		recv_count = 0;
		send_count = 0;

		// Find j index of procs holding pivcol
		int pivcol_j = -1;
		int i;
		for (i = 0; i < size; i++) {
			if (k >= dx(i) && k < dx(i)+dw(i)) {
				pivcol_j = dj(i);
				break;
			}
		}
		// Am i pivcol?
		int pivcol = 0;
		if (myj == pivcol_j) {
			pivcol = 1;
			A_pt = &A[(k - myx) * kb * lda]; // point to local A.
		}
		int pivrow_proc = -1;
		for (i = 0; i < size; i++) {
			if ((dj(i) == myj) && k >= dy(i) && k < dy(i)+dh(i)) {
				pivrow_proc = i;
				break;
			}
		}
		if (pivcol_j < 0) {
			fprintf(stderr, "Error, Problem with pivcol\n");
		}
		if (pivrow_proc < 0) {
			fprintf(stderr, "P%d: Error, Problem with pivrow_proc\n", rank);
			return -1;
		}

		// Send/recv	pivcol of A
		// test intersection with each proc.
		for (i = 0; i < size; i++) {
			//MPI_Barrier(comm);
			if ((pivcol || dj(i) == pivcol_j) && i != rank && // either i'm pivcol or proc i is
					(myy < dy(i) + dh(i)) && (myy + myh > dy(i))) { // there is an overlap

				//start send/recv from
				int start = dy(i) - myy;
				if (start < 0) {
					start = 0;
				}
				int length;
				if (dy(i) + dh(i) > myy + myh) {
					length = myh - start;
				} else {
					length = dy(i) + dh(i) - myy - start;
				}
				MPI_Datatype	fupermod_col;
				MPI_Type_vector(kb, length * kb, lda, MPI_DOUBLE, &fupermod_col);
				MPI_Type_commit(&fupermod_col);
				if (pivcol) { // send
					rc = MPI_Isend(&A_pt[start * kb], 1, fupermod_col, i, MPI_MSG_TAG_FOR_A, comm, &requestsS[send_count++]);
					if (rc != MPI_SUCCESS) {
						fprintf(stderr, "Error Sending A\n");
						return -1;
					}
				} else { // recv
					rc = MPI_Irecv(&WA[start * kb], 1, fupermod_col, i, MPI_MSG_TAG_FOR_A, comm, &requestsR[recv_count++]);
					if (rc != MPI_SUCCESS) {
						fprintf(stderr, "Error Recving A\n");
						return -1;
					}
				}
				MPI_Type_free(&fupermod_col);
			}
		}
		//==== Send/Recv B ====
		double* B_pt;
		int lldb;
		if (pivrow_proc == rank) {
			B_pt = &B[(k-myy) * kb]; // Set for local dgemm
			lldb = ldb;
			MPI_Datatype	fupermod_col;
			MPI_Type_vector(myw * kb, kb, ldb, MPI_DOUBLE, &fupermod_col);
			MPI_Type_commit(&fupermod_col);
			// send B-part
			for (i = 0; i < col_size; i++) {
				if (i != myi) {
					rc = MPI_Isend(&B[(k-myy) * kb], 1, fupermod_col, i, MPI_MSG_TAG_FOR_B, col_comm, &requestsS[send_count++]);
					if (rc != MPI_SUCCESS) {
						fprintf(stderr, "Error Sending B\n");
						return -1;
					}
				}
			}
			MPI_Type_free(&fupermod_col);
		} else {
			B_pt = WB;
			lldb = kb;
			MPI_Datatype	fupermod_col;
			MPI_Type_vector(myw * kb, kb, kb, MPI_DOUBLE, &fupermod_col);
			MPI_Type_commit(&fupermod_col);
			rc = MPI_Irecv(WB, 1, fupermod_col, di(pivrow_proc), MPI_MSG_TAG_FOR_B, col_comm, &requestsR[recv_count++]);
			if (rc != MPI_SUCCESS) {
				fprintf(stderr, "Error receiving B\n");
				return -1;
			}
			MPI_Type_free(&fupermod_col);
		}
		//====

		// Wait for all recv's to complete.
		MPI_Waitall(recv_count, requestsR, statusesR);
		MPI_Waitall(send_count, requestsS, statusesS);

		if (VERBOSE_MATRIX && rank == 0) {
			int j;
			printf("P%d A(%d,%d):\n", rank, myh * kb, kb);
			for (i = 0; i < myh * kb; i++) {
				for (j = 0; j < kb; j++) {
					printf("  %3.1f", A_pt[i + j * lda]);
				}
				printf("\n");
			}
			printf("P%d B(%d,%d):\n", rank, myh * kb, kb);
			for (i = 0; i < kb; i++) {
				for (j = 0; j < myw * kb; j++) {
					printf("  %3.1f", B_pt[i + j * lldb]);
				}
				printf("\n");
			}
		}	

		if (VERBOSE_MATRIX){
			double* tmpC = (double*)malloc(sizeof(double) * myh * kb * myw * kb);
			int i;
			for (i = 0; i < myh * kb * myw * kb; i++)
				tmpC[i] = C[i];
			fupermod_gemm_execute(w->gemm, CblasColMajor, CblasNoTrans, CblasNoTrans, myh * kb, myw * kb, kb, 1, A_pt, lda, B_pt, lldb, 1, tmpC, ldc);
			printf("P%d tmpC(%d,%d):\n", rank, myh * kb, myw * kb);
			int j;
			for (i = 0; i < myw * kb; i++) {
				for (j = 0; j < myh * kb; j++) {
					printf("  %3.1f", tmpC[i + j * ldc]);
				}
				printf("\n");
			}
			free(tmpC);
		}
		struct timeval start_dgemm, end_dgemm;
		gettimeofday(&start_dgemm, NULL);
		//		printf("\n*****M %d N %d K %d lda %d ldb %d ldc %d \n", myh*kb, myw*kb, kb, lda, ldb, myw*kb);
		rc = dgemm_hybrid(myh * kb, myw * kb, kb, 1, A_pt, lda, B_pt, lldb, 1, C, ldc, config);
		if (rc != 0)
		{
			fprintf(stderr, "error with dgemm_hybrid\n");
			MPI_Abort(comm, 71);
		}
		//mxm_cblas_dgemm(CblasColMajor, CblasNoTrans, CblasNoTrans, myh * kb, myw * kb, kb, 1, A_pt, lda, B_pt, lldb, 1, C, ldc);
		gettimeofday(&end_dgemm, NULL);
		double time = ((double)(end_dgemm.tv_sec - start_dgemm.tv_sec) + (end_dgemm.tv_usec - start_dgemm.tv_usec) * 1.0e-6);
		total_time_dgemm += time;
		//		fprintf(time_file, "%le\n", time);

		if (VERBOSE_MATRIX) {
			if (rank == 0) {
				printf("P%d C(%d,%d):\n", rank, myh * kb, myw * kb);
				int j;
				for (i = 0; i < myw * kb; i++) {
					for (j = 0; j < myh * kb; j++) {
						printf("  %3.1f", C[i + j * ldc]);
					}
					printf("\n");
				}
			}
		}
	}// end k loop

	// hybrid function
	dgemm_finalize(myh * kb, myw * kb, kb, 0, NULL, 0, NULL, 0, 0, C, ldc, config);

	//finish all communication before exiting
	MPI_Barrier(comm);

	if(PRINT_TIMES){
		struct timeval endt;
		gettimeofday(&endt, NULL);
		double total_time = ((double)(endt.tv_sec - startt.tv_sec) + (endt.tv_usec - startt.tv_usec) * 1.0e-6);
		double total_comm = total_time - total_time_dgemm;
		fprintf(time_file, "#P%d: total: %le dgemm: %le comm: %le ratio dgemm/comm : %4.2lf\n", rank, total_time, total_time_dgemm, total_comm, total_time_dgemm / total_comm);
		fprintf(stdout, "#P%d: total: %le dgemm: %le comm: %le ratio dgemm/comm : %4.2lf\n", rank, total_time, total_time_dgemm, total_comm, total_time_dgemm / total_comm);
		fclose(time_file);
	}

	free(requestsS);
	free(statusesS);
	free(requestsR);
	free(statusesR);
	return 0;
}
