/*! Kernel of the \ref mxm_hybrid routine */

#include "config.h"
#include "fupermod/fupermod_debug.h"
#include "fupermod/fupermod_kernel.h"
#include "datapart_2d/fupermod_hybrid.h"
#include "cblas_wrappers/fupermod_cblas.h"
#include "mxm_hybrid.h"
#include "memory_hybrid.h"
#include "dgemm_hybrid.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#define VERBOSE_MATRIX 0
#define MEM_TYPE PAGEABLE

/*! Switch to choose using MPI_SEND/Recv to self or memcpy */
#define KERNEL_MPI_COPY 1

// Overall problem size
//int N = 50;
// Block size
int kb = -1;
// Number of iterations to benchmark
int iters = 1;
double *A, *B, *C;
fupermod_dgemm_workspace* workspace;
MPI_Comm comm = MPI_COMM_NULL;

int load(MPI_Comm comm_, fupermod_process_conf* conf) {
/*	char* tokens[] = {"N", "k", "i"};
	char* value;
	while (*subopts != '\0') {
		switch (getsubopt(&subopts, tokens, &value)) {
			case 0:
				N = atoi(value);
				break;
			case 1:
				kb = atoi(value);
				break;
			case 2:
				iters = atoi(value);
				break;
			default:
				fprintf(stderr, "%s: unknown suboption %s\n", __func__, value);
				break;
		}
	}*/
	comm = comm_;
	return FUPERMOD_SUCCESS;
}

int unload(MPI_Comm comm) {
   NOT_USED(comm);
	return FUPERMOD_SUCCESS;
}

double complexity(long long int area, void* param){
  NOT_USED(param);
	// Same as benchmark but without 10%
	if (kb < 0) {
		fprintf(stderr, "Error kb not set yet!\n");
		MPI_Abort(comm, 12);
	}
	int myw, myh;
	myw = myh = sqrt(area);
	return 2.0 * (myh * kb) * (myw * kb) * (iters * kb);
}

int myw, myh;
double* WA;
double* WB;

int initialize_2d(int _myh, int _myw, void** params){
	myh = _myh;
	myw = _myw;
	mxm_col_params* param = (mxm_col_params*)*params;
	kb = param->kb;

	if (myh == 0 || myw == 0) {
		fprintf(stderr, "**==** Warning: zero size! myh =%d myw =%d\n", myh, myw);
	}
	//cublas initialize
	//mxm_cblas_initialize();

	// initialise A, B, C, WA, WB
	unsigned long long int mem_size = sizeof(double) * (myw * kb * myh * kb);
#ifdef __LP64__
	if (mem_size > 8580000000) { //Manual check to keep under 24GB
		int rank;
		MPI_Comm_rank(comm, &rank);
		fprintf(stderr, "P%d Not allocating matrix because of manual check\n"
				"P%d mem_size = %lu, myw = %d, myh = %d, kb = %d\n"
				"p%d %s returning %d\n",rank, rank, (long unsigned)mem_size, myw, myh, kb, rank, __func__, FUPERMOD_ALLOC_FAIL );
		return FUPERMOD_ALLOC_FAIL;
	}
#endif
	if (HYBRID_MEMORY) {
		malloc_hybrid(mem_size, &A, MEM_TYPE);
		malloc_hybrid(mem_size, &B, MEM_TYPE);
		malloc_hybrid(mem_size, &C, MEM_TYPE);
		malloc_hybrid(sizeof(double) * myh * kb * kb, &WA, MEM_TYPE);
		malloc_hybrid(sizeof(double) * myw * kb * kb, &WB, MEM_TYPE);
	} else {
		A = (double*) malloc(mem_size);
		B = (double*) malloc(mem_size);
		C = (double*) malloc(mem_size);
		WA = (double*) malloc(sizeof(double) * myh * kb * kb);
		WB = (double*) malloc(sizeof(double) * myw * kb * kb);
	}
	// Did it work?
	if (A == NULL || B == NULL || C == NULL || WA == NULL || WB == NULL) {
		int rank;
		MPI_Comm_rank(comm, &rank);
		fprintf(stderr, "P%d Failed to allocate matrix\n"
				"P%d myw = %d, myh = %d, kb = %d\n"
				"p%d %s returning %d\n",rank, rank, myw, myh, kb, rank, __func__, FUPERMOD_ALLOC_FAIL );
		//*time = FUPERMOD_MAX_TIME;
	if (HYBRID_MEMORY) {
		free_hybrid(A, PINNED); A = NULL;
		free_hybrid(B, PINNED); B = NULL;
		free_hybrid(C, PINNED); C = NULL;
		free_hybrid(WA, PINNED); WA = NULL;
	} else {
		free(A);  A = NULL;
		free(B);  B = NULL;
		free(C);  C = NULL;
		free(WA); WA = NULL;
		free(WB); WB = NULL;
	}
		return FUPERMOD_ALLOC_FAIL;
	}
	// Fill A, B. Zero C
	int i;
	for (i = 0; i < myh * kb * myw * kb; i++) {
		A[i] = 1.1;
		B[i] = 2.5;
		C[i] = 0.0;
	}
	// Pull WA, WB into main memory, possibly push out A,B,C
	for (i = 0; i < myh * kb * kb; i++) {
		WA[i] = 0.0;
	}
	for (i = 0; i < myw * kb * kb; i++) {
		WB[i] = 0.0;
	}

	return FUPERMOD_SUCCESS;
}

int initialize(long long int area, void** params){
	int _myh = sqrt(ROUNDUP * area);
	return initialize_2d(_myh, _myh, params);
}

int execute(pthread_mutex_t *mutex, void* params){
   NOT_USED(mutex);
	// hybrid dgemm initialize
	mxm_col_params* param = (mxm_col_params*)params;
	param->time = -1;
	int ldc = myh * kb;
	int rc_dgemm_init = 0;
	rc_dgemm_init = dgemm_initialize(myh * kb, myw * kb, kb, 0, NULL, 0, NULL, 0, 0, C, ldc, param->hybridconf);

	if (myh == 0 || myw == 0) {
		param->time = -1;
		return 0;
	}
	if (rc_dgemm_init != 0) {
		int rank;
		MPI_Comm_rank(MPI_COMM_WORLD, &rank);
		if (rc_dgemm_init == -2) {
			fprintf(stderr, "P%d: Warning with dgemm_initialize, but going to continue\n", rank);
			param->time = 1e9;
		} else {
			fprintf(stderr, "P%d: Error with dgemm_initialize\n", rank);
			fprintf(stderr, "\twith: myh:%d myw:%d kb:%d\n", myh, myw, kb);
			MPI_Abort(comm, 72);
		}
	}

	// For hybrid - reset all matrix so that paging is shown!
	// Fill A, B. Zero C
	int i;
	for (i = 0; i < myh * kb * myw * kb; i++) {
		A[i] = 1.1;
		B[i] = 2.5;
		C[i] = 0.0;
	}
	// Pull WA, WB into main memory, possibly push out A,B,C
	for (i = 0; i < myh * kb * kb; i++) {
		WA[i] = 0.0;
	}
	for (i = 0; i < myw * kb * kb; i++) {
		WB[i] = 0.0;
	}

	struct timeval start, end;
	gettimeofday(&start, NULL);

	MPI_Status statusesR;
	MPI_Request requestsS, requestsR;
	int k;
	int rc_dgemm_hybrid = 0;
	for (k = 0; k < iters; k++) {
		//make the benchmark as close to the real routine as possible:
		//MPI communications on the same process
		if (KERNEL_MPI_COPY) {
			MPI_Datatype fupermod_col;
			MPI_Type_vector(myh * kb, kb, myw * kb, MPI_DOUBLE, &fupermod_col);
			MPI_Type_commit(&fupermod_col);

			int rc;
			rc = MPI_Isend(&A[k * kb], 1, fupermod_col, 0, 0xf1, MPI_COMM_SELF, &requestsS);
			if (rc != MPI_SUCCESS) {
				printf("Error Sending A\n");
				return FUPERMOD_FAIL;
			}
			MPI_Type_free(&fupermod_col);
			rc = MPI_Irecv(WA, myh * kb * kb, MPI_DOUBLE, 0, 0xf1, MPI_COMM_SELF, &requestsR);
			if (rc != MPI_SUCCESS) {
				printf("Error Recving A\n");
				return FUPERMOD_FAIL;
			}
			MPI_Waitall(1, &requestsR, &statusesR);
		} else { // If you want to avoid mpi copy do this:
			int i;	
			for (i = 0; i < myh * kb ; i++) {
				memcpy(&WA[i * kb], &A[myw * kb * i +(k * kb)], sizeof(double) * kb);
			}
		}
		double* B_local = &B[(k * kb) * (myw * kb)];
		memcpy(WB, B_local, sizeof(double) * myw * kb * kb);

		// multiply WA x WB = C
		// mxm_cblas_xgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans, myh * kb, myw * kb, kb, 1, WA, kb, WB, myw * kb, 1, C, myw * kb);
		//printf("\n-----running dgemm\n");
		//printf("dgemm_hybrid(%d, %d, %d, 1, WA, %d, WB, %d, 1, C, %d, hybridconf)\n", myh * kb, myw * kb, kb, myh * kb, kb, myh * kb);
		if (myh != 0 && myw != 0) {
			rc_dgemm_hybrid = dgemm_hybrid(myh * kb, myw * kb, kb, 1, WA, myh * kb, WB, kb, 1, C, myh * kb, param->hybridconf);
		}
	}
	gettimeofday(&end, NULL);
	if (myh != 0 && myw != 0) {
		dgemm_finalize(myh * kb, myw * kb, kb, 0, NULL, 0, NULL, 0, 0, C, ldc, param->hybridconf);
	}
	if (rc_dgemm_hybrid == -1) {
		int rank;
		MPI_Comm_rank(MPI_COMM_WORLD, &rank);
		fprintf(stderr, "P%d: Error with dgemm_hybrid returned -1", rank);
		MPI_Abort(comm, 74);
	}
	if (rc_dgemm_hybrid != 0 || rc_dgemm_init != 0) {
		param->time = 1e9;
		return -2;
	} 
	param->time = ((double)(end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec) * 1.0e-6) / iters;
	return 0;
}

int finalize(void* params){
   NOT_USED(params);
	if (HYBRID_MEMORY) {
		free_hybrid(A, PINNED); A = NULL;
		free_hybrid(B, PINNED); B = NULL;
		free_hybrid(C, PINNED); C = NULL;
		free_hybrid(WA, PINNED); WA = NULL;
	} else {
		free(A);  A = NULL;
		free(B);  B = NULL;
		free(C);  C = NULL;
		free(WA); WA = NULL;
		free(WB); WB = NULL;
	}
	//cublas finalize
	//mxm_cblas_finalize();
	return FUPERMOD_SUCCESS;
}

fupermod_kernel kernel = {
    load,
    unload,
	complexity,
	initialize,
	execute,
	finalize,
};
