/*!
 * \page mxm_col Matrix multiplication col
 * Implements the model-based BR algorithm
 * todo: to rewrite
 */
#include "config.h"
#include "fupermod/fupermod_partition.h"
#include "datapart_2d/fupermod_hybrid.h"
#include "mxm_hybrid.h"
#include "memory_hybrid.h"
#include "cblas_wrappers/fupermod_cblas.h"
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define VERBOSE_MATRIX 0
#define LOCAL_TEST 0

/*! Fill matrix with increasing numbers - For testing purpose. */
void fill_matrix_testing(int rank, fupermod_dgemm_workspace* workspace, double *A, double *B, double *C);
/*! Fill matrix with constant numbers. */
void fill_matrix_constants(int rank, fupermod_dgemm_workspace* workspace, double *A, double *B, double *C);
/*! Test: compare solution with a serial mxm. */
void compare_solution(int rank, fupermod_dgemm_workspace* workspace, double *C, MPI_Comm comm);

int main(int argc, char** argv) {
	MPI_Init(&argc, &argv);
	MPI_Comm comm = MPI_COMM_WORLD;

	// exit routine
	int exit = 0; 
	// Test final result.
	int test_result = 0;
	//partition file path
	char dist_path[256] = "./partition.dist";
	char nodedist_path[256] = "./nodedist";
	char nodetype_path[256] = "./nodetype";

	int rank;
	MPI_Comm_rank(comm, &rank);
	int size;
	MPI_Comm_size(comm, &size);
	int root = 0;

	//if (!rank) getc(stdin); MPI_Barrier(MPI_COMM_WORLD);

	if (rank == root) {
		int opt;
		while ((opt = getopt(argc, argv, "hk:tp:q:r:")) != -1)	
			switch(opt) {
				case 'h':
					fprintf(stderr,
							"routine\n"
							"Usage: [options]\n"
							"	-h	help\n"
							"	-t	Test final result (default none)\n"
							"	-p S path to input inter-node distribution (default: %s)\n"
							"	-q S path to input node distribution (default: %s)\n"
							"	-r S path to read node type info (default: %s)\n" 
							, dist_path, nodedist_path, nodetype_path);
					exit = 1;
					break;
				case 't':
					test_result = 1;
					break;
				case 'p':
					strcpy(dist_path, optarg);
					break;
				case 'q':
					strcpy(nodedist_path, optarg);
					break;
				case 'r':
					strcpy(nodetype_path, optarg);
					break;
				default:
					fprintf(stderr, "Unknown option %s\n", optarg);
					break;
			}
	}

	MPI_Bcast(&exit, 1, MPI_INT, root, comm);
	if (exit) {
		MPI_Finalize();
		return 0;
	}

	MPI_Bcast(&test_result, 1, MPI_INT, root, comm);
	MPI_Bcast(&dist_path, 256, MPI_CHAR, root, comm);
	MPI_Bcast(&nodedist_path, 256, MPI_CHAR, root, comm);
	MPI_Bcast(&nodetype_path, 256, MPI_CHAR, root, comm);

	//initialize
	fupermod_gemm* gemm = fupermod_gemm_alloc(NULL);

	// read from dist file and allocate working space
	fupermod_dgemm_workspace* workspace = NULL;
	char** hostnames = fupermod_allgather_hostnames(comm);
	FILE* stream = NULL;

  fupermod_process_conf* confs; //todo: get conf for here. merge old hybridconf with new fupermod_process_conf.
	if ((stream = fopen(dist_path, "r")) != NULL) {
		workspace = fupermod_dgemm_alloc_read(comm, stream, confs, gemm);
		if (workspace == NULL)
			MPI_Abort(comm, 51); 
		fclose(stream);
	} else {
		fprintf(stderr, "Error: Could not open dist file: %s\n", dist_path);
		MPI_Abort(comm, 50);
	}
	free(hostnames);

	// Get hybrid variables and dist.
	HybridConfig* hybridconf = (HybridConfig*)malloc(sizeof(HybridConfig));
	int rc;
	rc = hybrid_nodetype_read(nodetype_path, hybridconf);
	if (rc != 0){
		fprintf(stderr, "P%d: Error reading nodetype with path:%s\n", rank, nodetype_path);
		MPI_Abort(comm, 40);
	}
	rc = hybrid_nodedist_read(nodedist_path, hybridconf);
	if (rc != 0){
		fprintf(stderr, "P%d: Error reading node dist with path:%s\n", rank, nodedist_path);
		MPI_Abort(comm, 41);
	}

	//mxm_cblas_initialize();

	if (rank == 0)
		mm_col_print(workspace->idist);

	int myw = workspace->idist->d[rank].w;
	int myh = workspace->idist->d[rank].h;
	int kb = workspace->idist->kb;

	// initialise A, B, C, WA, WB
	size_t mem_size = sizeof(double) * (myw * kb * myh * kb);
	double *A, *B, *C;

	if (HYBRID_MEMORY) {
		malloc_hybrid(mem_size, &A, PINNED);
		malloc_hybrid(mem_size, &B, PINNED);
		malloc_hybrid(mem_size, &C, PINNED);
	} else {
		A = (double*) malloc(mem_size);
		B = (double*) malloc(mem_size);
		C = (double*) malloc(mem_size);
	}

	// Did it work?
	if (A == NULL || B == NULL || C == NULL) {
		fprintf(stderr, "P%d Failed to allocate matrix\n"
				"P%d myw = %d, myh = %d, kb = %d\n"
				"P%d A:%p B:%p C:%p\n"
				"p%d Exiting\n",rank, rank, myw, myh, kb, rank, A, B, C, rank);

		MPI_Abort(comm, 11);
	}
	// Fill A, B. Zero C
	if (test_result)
		fill_matrix_testing(rank, workspace, A, B, C);
	else
		fill_matrix_constants(rank, workspace, A, B, C);

	//fupermod_dgemm
	fupermod_dgemm(A, B, C, comm, workspace, *hybridconf);

	if (test_result)
		compare_solution(rank, workspace, C, comm);

	fupermod_dgemm_free(workspace);
	if (HYBRID_MEMORY) {
		free_hybrid(A, PINNED); A = NULL;
		free_hybrid(B, PINNED); B = NULL;
		free_hybrid(C, PINNED); C = NULL;
	} else {
		free(A);  A = NULL;
		free(B);  B = NULL;
		free(C);  C = NULL;
	}

	//cublas finalize
	fupermod_gemm_free(gemm);

	MPI_Finalize();
	return 0;
}

void fill_matrix_testing(int rank, fupermod_dgemm_workspace* workspace, double *A, double *B, double *C){

	if (rank == 0) 
		mm_col_print(workspace->idist);
	int kb =  workspace->idist->kb;
	int kw = kb * workspace->idist->d[rank].w;
	int kh = kb * workspace->idist->d[rank].h;
	int kx = kb * workspace->idist->d[rank].x;
	int ky = kb * workspace->idist->d[rank].y;
	int kN = kb * workspace->idist->N;

	int i, j;
	for (j = 0; j < kw; j++) {
		for (i = 0; i < kh; i++) {
			A[i + j * kh] = (i + ky) * kN + (j + kx);
			B[i + j * kh] = (i + ky) * kN + (j + kx) + 5;
		}
	}

	for (i = 0; i < kh* kw; i++) {
		C[i] = 0.0;
	}
	if (VERBOSE_MATRIX && rank == 0) {
		printf("P%d A(%d,%d):\n", rank, kh, kw);
		for (i = 0; i < kh; i++) {
			for (j = 0; j < kw; j++) {
				printf("  %3.1f", A[i + j * kh]);
			}
			printf("\n");
		}
		printf("P%d B(%d,%d):\n", rank, kh, kw);
		for (i = 0; i < kh; i++) {
			for (j = 0; j < kw; j++) {
				printf("  %3.1f", B[i + j * kh]);
			}
			printf("\n");
		}
	}	


}

void fill_matrix_constants(int rank, fupermod_dgemm_workspace* workspace, double *A, double *B, double *C){
	int kb =  workspace->idist->kb;
	int myw = workspace->idist->d[rank].w;
	int myh = workspace->idist->d[rank].h;

	int i;
	for (i = 0; i < myh * kb * myw * kb; i++) {
		A[i] = 1.1;
		B[i] = 2.5;
		C[i] = 0.0;
	}
}

void compare_solution(int rank, fupermod_dgemm_workspace* workspace, double *C, MPI_Comm comm){
	int kb =  workspace->idist->kb;
	int kw = kb * workspace->idist->d[rank].w;
	int kh = kb * workspace->idist->d[rank].h;
	int kx = kb * workspace->idist->d[rank].x;
	int ky = kb * workspace->idist->d[rank].y;
	int kN = kb * workspace->idist->N;

	double* At = (double*) malloc(sizeof(double) * (kN * kN));
	double* Bt = (double*) malloc(sizeof(double) * (kN * kN));
	double* Ct = (double*) malloc(sizeof(double) * (kN * kN));

	int i, j;
	for (j = 0; j < kN; j++) {
		for (i = 0; i < kN; i++) {
			At[i + j * kN] = i * kN + j;
			Bt[i + j * kN] = i * kN + j + 5;
		}
	}
	for (i = 0; i < kN * kN; i++) {
		Ct[i] = 0.0;
	}
	fupermod_gemm_execute(workspace->gemm, CblasColMajor, CblasNoTrans, CblasNoTrans, kN, kN, kN, 1, At, kN, Bt, kN, 0, Ct, kN);

	double diff = 0;
	for (i = 0; i < kh; i++) {
		for (j = 0; j < kw; j++) {
			double tmp;
			tmp = fabs(Ct[(i + ky) + (j + kx) * kN] - C[i + j * kh]);
	//		if (tmp > 1e-6)
	//			fprintf(stderr, "Error:%le\n", tmp);
			diff += tmp;
		}
	}
	/*if (VERBOSE_MATRIX && rank == 0) {
		printf("P%d C(%d,%d):\n", rank, kh, kw);
		for (i = 0; i < kh; i++) {
		for (j = 0; j < kw; j++) {
		printf("  %3.1f", C[i + j * kh]);
		}
		printf("\n");
		}
		printf("P%d C - testing(%d,%d):\n", rank, kN, kN);
		for (i = 0; i < kN; i++) {
		for (j = 0; j < kN; j++) {
		printf("  %3.1f", Ct[i + j * kN]);
		}
		printf("\n");
		}
		}*/

	double sum;
	MPI_Reduce(&diff, &sum, 1, MPI_DOUBLE, MPI_SUM, 0, comm);
	if (rank == 0) {
		if (sum < 1E-6)
			fprintf(stderr, "Matrix multiplication was successful, diff = %le\n", sum);
		else 
			fprintf(stderr, "***\nError: Matrix multiplication failed, diff = %le\n***\n", sum);
	}

	free(At); At = NULL;
	free(Bt); Bt = NULL;
	free(Ct); Ct = NULL;
}
