/*!
 * Builds 2 level hybrid full functional performance models.
 */

#include "config.h"
#include "fupermod/fupermod_data.h"
#include "fupermod/fupermod_dynamic.h"
#include "datapart_2d/fupermod_hybrid.h"
#include <getopt.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

int main(int argc, char** argv) {
  MPI_Init(&argc, &argv);
  MPI_Comm comm = MPI_COMM_WORLD;

  int exit = 0;
  char path[256] = "";
  char subopts[256] = "";
  char data_dir[256] = "./fpm";
  int max_steps = 3;
  int n_upper = 100;
  int n_lower = 1;
  int increment = 1;
  int method = 0;
  int debug = 0;
  double granularity = 1;
  fupermod_precision precision = fupermod_precision_defaults();

  int rank;
  MPI_Comm_rank(comm, &rank);
  int root = 0;
  //if (rank == root) {
  int opt;
  while ((opt = getopt(argc, argv, "hl:f:o:m:L:U:g:c:s:r:R:T:i:e:z:Z:d:")) != -1)
    switch (opt) {
      case 'h':
        fprintf(
            stderr,
            "Builds full functional performance model\n"
            "Usage: builder [options]\n"
            "	-h	help\n"
            "	-l S	shared library (required: a full path or relative to LD_LIBRARY_PATH)\n"
            "	-f S	path to data file (default: %s)\n"
            "	-o S	suboptions (comma separated options for the shared library)\n"
            "	-m I	method (default %d)\n"
            "	     0 Fixed number of points set [L, U, s] \n"
            "	     1 Adaptive, set [L, U, g, c]\n"
            "	-L I	lower problem size (default:%d)\n"
            "	-U I	upper problem size (default:%d)\n"
            "	-g D	granularity of measurement (default %f)\n"
            "	-c I	initial increment (default %d)\n"
            "	-s I	number of steps in the model (default %d)\n"
            "	-r I	minimum number of repetitions to average over (>=1) (default %d)\n"
            "	-R I	maximum number of repititions (defaule %d)\n"
            "	-T D	Maximum time benchmarking a point after which no more repititions are started (defaule %f)\n"
            "	-i D	confidence level (default %f)\n"
            "	-e D	relative error (default %f)\n"
            "	-z D	zero speed, below which benchmark is killed with pthreads, disabled with 0 (default %le).\n"
            "	-Z D	max time, which benchmark is killed with pthreads (defaule %le).\n"
            "	-d  	debug\n",
          data_dir, method, n_lower, n_upper, granularity, increment, max_steps,
          precision.reps_min, precision.reps_max, precision.time_max_rep, precision.cl, precision.eps, precision.zero_speed, precision.time_max_kill);
        exit = 1;
        break;
      case 'l':
        strcpy(path, optarg);
        break;
      case 'f':
        strcpy(data_dir, optarg);
        break;
      case 'o':
        strcpy(subopts, optarg);
        break;
      case 'm':
        method = atoi(optarg);
        break;
      case 'L':
        n_lower = atoi(optarg);
        break;
      case 'U':
        n_upper = atoi(optarg);
        break;
      case 'g':
        granularity = atoi(optarg);
        break;
      case 'c':
        increment = atoi(optarg);
        break;
      case 's':
        max_steps = atoi(optarg);
        break;
      case 'r':
        precision.reps_min = atoi(optarg);
        break;
      case 'R':
        precision.reps_max = atoi(optarg);
        break;
      case 'T':
        precision.time_max_rep = atof(optarg);
        break;
      case 'z':
        precision.zero_speed = atof(optarg);
        break;
      case 'Z':
        precision.time_max_kill = atof(optarg);
        break;
      case 'i':
        precision.cl = atof(optarg);
        break;
      case 'e':
        precision.eps = atof(optarg);
        break;
      case 'd':
        debug = 1;
        break;
      default:
        fprintf(stderr, "Unknown option %s\n", optarg);
        break;
    }
  if (!strcmp(path, "")) {
    fprintf(stderr, "Error: routine shared library not specified\n");
    exit = 1;
  }
  //}
  MPI_Bcast(&exit, 1, MPI_INT, root, comm);
  if (exit) {
    MPI_Finalize();
    return 0;
  }
  //MPI_Bcast(&path, 256, MPI_CHAR, root, comm);
  MPI_Bcast(&data_dir, 256, MPI_CHAR, root, comm);
  MPI_Bcast(&subopts, 256, MPI_CHAR, root, comm);
  MPI_Bcast(&method, 1, MPI_INT, root, comm);
  MPI_Bcast(&n_lower, 1, MPI_INT, root, comm);
  MPI_Bcast(&n_upper, 1, MPI_INT, root, comm);
  MPI_Bcast(&granularity, 1, MPI_DOUBLE, root, comm);
  MPI_Bcast(&increment, 1, MPI_INT, root, comm);
  MPI_Bcast(&max_steps, 1, MPI_INT, root, comm);
  MPI_Bcast(&debug, 1, MPI_INT, root, comm);
  MPI_Bcast(&precision, sizeof(fupermod_precision), MPI_CHAR, root, comm);

  if (debug) {
    if (!rank) 
      getc(stdin);
    MPI_Barrier(MPI_COMM_WORLD);
  }
  fupermod_library library;
  fupermod_library_load(comm, path, NULL, &library);
  printf("Kernel:%s %s\n", library.path, library.conf->subopts);


  HybridConfig* hybridconf = (HybridConfig*)malloc(sizeof(HybridConfig));
  int rc;
  //todo: make command line args of following:
  char nodetype_path[256] = "./nodetype";
  int node_iters = 100;
  double node_eps = 0.001;
  int kb = 128;
  fupermod_partitioner node_partitioner = partitioner_switch(3, 0, 0);

  rc = hybrid_nodetype_read(nodetype_path, hybridconf);
  if (rc != 0){
    fprintf(stderr, "P%d: Error reading nodetype with path:%s\n", rank, nodetype_path);
    MPI_Abort(comm, 40);
  }

  fupermod_benchmark_hybrid* benchmark = fupermod_benchmark_hybrid_alloc(library.kernel, hybridconf, n_upper, kb, node_iters, node_eps, node_partitioner);
  fupermod_benchmark* bench = (fupermod_benchmark*)benchmark;

  // Data file for a process
  char hostname[MPI_MAX_PROCESSOR_NAME];
  int len;
  MPI_Get_processor_name(hostname, &len);
  MPI_Comm comm_intra;
  fupermod_comm_intra(comm, &comm_intra);
  int rank_intra;
  MPI_Comm_rank(comm_intra, &rank_intra);
  

  // get the host names
  // TODO: UPDATE HERE.
  char conf_file[256] = "conf_file";
  if (!strcmp(conf_file, "")) {
    fprintf(stderr, "Error: must specify conf_file\n");
    return -1;
  }

  fupermod_process_conf conf = fupermod_get_conf(MPI_COMM_WORLD, conf_file);
  char* filename = fupermod_data_filename(data_dir, conf);
  FILE* outfile = fopen(filename, "w");
  if (outfile == NULL)
    MPI_Abort(comm, -1);
  fupermod_data_write_head(outfile, &library);
  // Data file for a processor
  /*	fupermod_data_filename_node(data_dir, hostname, filename);
      FILE* outfile_node = NULL;
      if (rank_intra == 0) {
      outfile_node = fopen(filename, "w");
      if (outfile_node == NULL)
      MPI_Abort(comm, -1);
      fupermod_data_write_head(outfile_node, &library);
      }*/
  free(filename); filename = NULL;
  if (max_steps < 2 || n_upper <= n_lower)
    increment = n_upper;
  else if (method == 0)
    increment = (int) (n_upper - n_lower) / (max_steps - 1);
  if (increment < 1)
    increment = 1;

  int n;
  for (n  = n_lower; n <= n_upper; n += increment) {
    // Data point for a process
    fupermod_point point;
    bench->execute(bench, comm_intra, n, precision, &point);
    fupermod_point_write(outfile, &point, library.kernel->complexity);
    // Data point for a processor
    fupermod_point node_point;
    fupermod_point_node(&point, &node_point, comm_intra, 0);
    /*if (rank_intra == 0) {
      fupermod_point_write(outfile_node, &node_point, kernel->complexity);
      fflush(outfile_node);
      } */

    if (method == 1 && point.t / increment > granularity)
      increment *= 2;
  }
  fclose(outfile);
  /*if (rank_intra == 0)
    fclose(outfile_node);*/
  MPI_Comm_free(&comm_intra);

  fupermod_benchmark_hybrid_free(benchmark);
  fupermod_library_unload(comm, &library);

  MPI_Finalize();
  return 0;
}
