/*!
 * \page dyparter Dynamic partitioner
 * Dynamically builds partial functional performance models
 */

#include "config.h"
#include "fupermod/fupermod_conf.h"
#include "fupermod/fupermod_dynamic.h"
#include <getopt.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>

#define VERBOSE 1

int main(int argc, char** argv) {
  MPI_Init(&argc, &argv);
  MPI_Comm comm = MPI_COMM_WORLD;

  int exit = 0;
  char path[256] = "";
  char data_dir[256] =  "./";
  char* conf_file =  "./conf_file";
  char dist_path[256] = "./partition.dist";
  int print_models = 1;
  int print_models_data = 1;
  int algorithm = 3;
  char subopts[256] = "";
  long long int D = 20;
  int iters = 10;
  double eps = 0.001;
  //todo: dont want defaults here.
  fupermod_precision precision; // = fupermod_precision_defaults();

  precision.reps_min = 1;
  precision.reps_max = 1;
  precision.cl = 0.95;
  precision.eps = 0.025;
  precision.time_max_rep = 60; // 1 minute
  precision.zero_speed = 1e7; // 1e7 flops.
  precision.time_max_kill = 1200; // 20 minutes
  int smooth = 0;

  int rank;
  MPI_Comm_rank(comm, &rank);
  int root = 0;
  if (rank == root) {
    int opt;
    while ((opt = getopt(argc, argv, "ha:l:o:p:d:i:e:")) != -1)
      switch (opt) {
        case 'h':
          fprintf(
              stderr,
              "Dynamic partitioner: dynamically builds partial functional performance models\n"
              "Usage: dyparter [options]\n"
              "  -h    help\n"
              "-a I	Balancing algorithm. (default: %d)\n"
              "\t\t0: Homogeneous distribution\n"
              "\t\t1: Constant performance model 1 (small benchmark)\n"
              "\t\t2: Constant performance model 2 (homogeneous benchmark)\n"
              "\t\t3: Geometric algorithm FFPM\n"
              "\t\t4: FPM multiroot solver\n"
              "  -l S  shared library (required: a full path or relative to LD_LIBRARY_PATH)\n"
              "  -o S  suboptions (comma separated options for the shared library)\n"
              "  -p S  path to output inter-node distribution (default: %s)\n"
              "  -d I  problem size to partition (default %lld)\n"
              "  -i I  number of inter-node iterations (default: %d)\n"
              "  -e I  Stopping condition eps (default: %lf)\n"
              , algorithm, dist_path, D, iters, eps);
          exit = 1;
          break;
        case 'a':
          algorithm = atoi(optarg);
          break;
        case 'l':
          strcpy(path, optarg);
          break;
        case 'o':
          strcpy(subopts, optarg);
          break;
        case 'p':
          strcpy(dist_path, optarg);
          break;
        case 'd':
          D = atoi(optarg);
          break;
        case 'i':
          iters = atoi(optarg);
          break;
        case 'e':
          eps = atof(optarg);
          break;
        default:
          fprintf(stderr, "Unknown option %s\n", optarg);
          break;
      }
  }
  MPI_Bcast(&exit, 1, MPI_INT, root, comm);
  if (exit) {
    MPI_Finalize();
    return 0;
  }
  MPI_Bcast(&path, 256, MPI_CHAR, root, comm);
  MPI_Bcast(&subopts, 256, MPI_CHAR, root, comm);
  MPI_Bcast(&D, 1, MPI_INT, root, comm);
  MPI_Bcast(&iters, 1, MPI_INT, root, comm);
  MPI_Bcast(&eps, 1, MPI_DOUBLE, root, comm);

  // Read machine file and get device type (create if doesn't exist).
  /*TODO this won't works on a platform without nfs*/
  int file_exist = 0;
  if(rank == 0 && access(conf_file, R_OK) !=-1){
    file_exist = 1;
  }
  MPI_Bcast(&file_exist, 1, MPI_INT, root, comm);
  if(file_exist==0){
    FILE* file = fopen(conf_file, "w");
    MPI_Barrier(MPI_COMM_WORLD);
    if(file == NULL) {
      perror("can't create conf_file");
      MPI_Finalize();
      return 0;
    }
    fupermod_print_conf(MPI_COMM_WORLD, root, file, "cpu", subopts);
    fclose(file);
    fflush(file);
    MPI_Barrier(MPI_COMM_WORLD);
  }
  
  int size;
  MPI_Comm_size(comm, &size);

  int conf_file_size;
  fupermod_process_conf* confs = fupermod_get_conf_all_sorted(comm, conf_file, &conf_file_size);
  if (conf_file_size != size) {
    fprintf(stderr, "Error: conf_size != size %d != %d, %s\n", conf_file_size, size, __FUNCTION__);
    MPI_Abort(comm, 1);
  }

  fupermod_library library;
  fupermod_library_load(comm, path, &confs[rank], &library);

  fupermod_data** data = NULL;
  fupermod_model** models = NULL;
  fupermod_model_free model_free = NULL;
  fupermod_partition partition = NULL;

  //Initialise data and models on root
  if (rank == root) {
    data = (fupermod_data**)malloc(sizeof(fupermod_data*) * size);
    models = malloc(sizeof(fupermod_model*) * size);
    int i;
    for (i = 0; i < size; i++) {
      data[i] = fupermod_data_alloc();
    }
  }
  switch (algorithm) {
    case 0: // Homogeneous distribution
      {
        /*    p.partition = fupermod_partition_homogeneous;
              p.model_alloc = fupermod_model_constant_alloc;
              p.model_free = fupermod_model_constant_free;
              p.arg = benchsize; */
      }
      break;
    case 1: // Constant performance model 1 (small benchmark)
    case 2: // Constant performance model 2 (homogeneous benchmark)
      {
        partition = fupermod_partition_basic;

        int bench_size = (algorithm == 1 ? 1e2 : D / size);
        int i;
        for (i = 0; i < size; i++)
          models[i] = fupermod_model_constant_alloc(data[i], library.kernel->complexity, D, NULL, bench_size);
        model_free = fupermod_model_constant_free;
      }
      break;
    case 3: // FPM Geometric partitioning
      {
        partition = fupermod_partition_geometric;
        if (rank == root) {
          int i;
          for (i = 0; i < size; i++)
            models[i] = fupermod_model_piecewise_alloc(data[i], library.kernel->complexity, D, NULL, smooth);
          model_free = fupermod_model_piecewise_free;
        }
      }
      break;
      /*
         case 4: // FPM multiroot solver
      //todo

      { 
      partition = fupermod_partition_multiroot;
      int i;
      for (i = 0; i < size; i++)
      .model_alloc = fupermod_model_interp_alloc;
      p.model_free = fupermod_model_interp_free;
      p.arg = smooth;
      }
      case 4: // FPM multiroot solver
      {
      partition = fupermod_partition_multiroot;
      int i;
      for (i = 0; i < size; i++)
      models[i] = fupermod_model_interp_alloc(data[i], kernel->complexity, D, NULL, smooth);
      model_free = fupermod_model_interp_free;
      }
      break; */
      break;
  }
  fupermod_dynamic partitioner = {partition, size, models, fupermod_dist_alloc(D, size)};
  fupermod_benchmark* benchmark = fupermod_benchmark_basic_alloc(library.kernel, confs[rank]);

  int iter = 0;
  int status = 0;
  // Main loop

  while (iter++ < iters && status == 0) {
    if (VERBOSE && rank == root) {
      fprintf(stderr, "** iter:%d of %d: ", iter, iters);
      int i;
      for (i = 0; i < size; i++) {
        fprintf(stderr, "%lld ", partitioner.dist->parts[i].d);
      }
      fprintf(stderr, "\n");
    }
    status = fupermod_partitioner_iterate(&partitioner, comm, root, precision, benchmark, eps);
  }

  // Output and free
  if (rank == root) {
    // Output the distribution
    FILE* distout = fopen(dist_path, "w");
    fupermod_dist_write(distout, partitioner.dist, &library, confs);
    fclose(distout);
  }
  if (rank == root && print_models) {
    int i;
    for (i = 0; i < size; i++) {
      char* filename = fupermod_model_filename(data_dir, confs[i]);
      FILE* outfile = fopen(filename, "w");
      free(filename); filename = NULL;
      if (outfile == NULL)
        MPI_Abort(comm, -1);
      fupermod_print_model(outfile, &library, confs[i], models[i], D);
      fclose(outfile);
    }
  }
  if (rank == root && print_models_data) {
    char *filename = NULL;
    asprintf(&filename, "%spoints_all.dat", data_dir);
    FILE* file = fopen(filename, "w");
    fprintf(file, "#Kernel:%s %s, D:%lld, machines:", library.path, library.conf->subopts, D);
    int i;
    for (i = 0; i < size; i++) {
      fprintf(file, "%s.%d.%s ", confs[i].hostname, confs[i].rank_intra, confs[i].device_type);
    }
    fprintf(file, "\n");
    fupermod_print_models_data(file, size, models);
    fclose(file);
  }
  if (rank == root) {
    fupermod_dist_free(partitioner.dist);
    int i;
    for (i = 0; i < size; i++) {
      model_free(models[i]);
      fupermod_data_free(data[i]);
    }
    free(models);
    free(data);
  }

  fupermod_benchmark_basic_free(benchmark);

  MPI_Finalize();
  return 0;
}
