/*!
 * \page hybrid_parter Hybrid partitioner
 * Dynamically builds partial functional performance models
 */

#include "config.h"
#include "fupermod/fupermod_dynamic.h"
#include "datapart_2d/fupermod_hybrid.h"
#include <getopt.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define VERBOSE 1

int main(int argc, char** argv) {
  MPI_Init(&argc, &argv);
  MPI_Comm comm = MPI_COMM_WORLD;

  int exit = 0;
  char path[256] = "";
  char subopts[256] = "";
  char nodetype_path[256] = "./nodetype";
  char dist_path[256] = "./partition.dist";
  char nodedist_path[256] = "./nodedist";
  int algorithm = 4;
  int node_algorithm = 4;
  int print_models = 1;
  int print_models_data = 1;
  int N = 64; // matrix size
  int kb = 4;
  int iters = 100;
  int node_iters = 100;
  double eps = 0.001;
  double node_eps = 0.001;
  fupermod_precision precision = fupermod_precision_defaults();
  int smooth = 0;
  int benchsize = 0;

  int rank;
  MPI_Comm_rank(comm, &rank);

  //if (!rank) getc(stdin); MPI_Barrier(comm);
  int root = 0;
  if (rank == root) {
    int opt;
    while ((opt = getopt(argc, argv, "ha:b:l:o:N:i:j:E:e:k:p:q:r:")) != -1)
      switch (opt) {
        case 'h':
          fprintf(
              stderr,
              "Dynamic partitioner: dynamically builds partial functional performance models\n"
              "Usage: hybrid_parter [options]\n"
              "	-h	help\n"
              "	-a I	Inter-node Balancing algorithm. (default: %d)\n"
              "	\t0: Homogeneous distribution\n"
              "	\t1: Constant performance model\n"
              "	\t3: Geometric algorithm FFPM\n"
              "	\t4: FPM multiroot solver\n"
              "	-b I	Node balancing algorithm. (default: %d)\n"
              "	\t0: Homogeneous distribution\n"
              "	\t1: Constant performance model\n"
              "	\t3: Geometric algorithm FFPM\n"
              "	\t4: FPM multiroot solver\n"
              "	-l S	shared library (required: a full path or relative to LD_LIBRARY_PATH)\n"
              "	-o S	suboptions (comma separated options for the shared library)\n"
              "	-N I matrix size to partition (default: %d, should be divisible by kb)\n"
              "	-i I number of inter-node iterations (default: %d)\n"
              "	-j I number of iterations within node(default: %d)\n"
              "	-E F Epsilon error for stopping inter-node (default: %lf)\n"
              "	-e F Epsilon error for stopping within node (default: %lf)\n"
              "	-k I block size kb (default: %d)\n"
              "	-p S path to output inter-node distribution (default: %s)\n"
              "	-q S path to output node distribution (default: %s)\n"
              "	-r S path to read node type info (default: %s)\n"
              , algorithm, node_algorithm, N, iters, node_iters, eps, node_eps, kb, dist_path, nodedist_path, nodetype_path);
          exit = 1;
          break;
        case 'a':
          algorithm = atoi(optarg);
          break;
        case 'b':
          node_algorithm = atoi(optarg);
          break;
        case 'l':
          strcpy(path, optarg);
          break;
        case 'o':
          strcpy(subopts, optarg);
          break;
        case 'N':
          N = atoi(optarg);
          break;
        case 'i':
          iters = atoi(optarg);
          break;
        case 'j':
          node_iters = atoi(optarg);
          break;
        case 'E':
          eps = atof(optarg);
          break;
        case 'e':
          node_eps = atof(optarg);
          break;
        case 'k':
          kb = atoi(optarg);
          break;
        case 'p':
          strcpy(dist_path, optarg);
          break;
        case 'q':
          strcpy(nodedist_path, optarg);
          break;
        case 'r':
          strcpy(nodetype_path, optarg);
          break;
        default:
          fprintf(stderr, "Unknown option %s\n", optarg);
          break;
      }
    if ( N % kb != 0){
      N = (N / kb) * kb;
      fprintf(stderr, "Warning!\n\tkb does not devide N evenly, rounding to N=%d\n", N);
    }
  }
  MPI_Bcast(&exit, 1, MPI_INT, root, comm);
  if (exit) {
    MPI_Finalize();
    return 0;
  }
  MPI_Bcast(&path, 256, MPI_CHAR, root, comm);
  MPI_Bcast(&nodedist_path, 256, MPI_CHAR, root, comm);
  MPI_Bcast(&nodetype_path, 256, MPI_CHAR, root, comm);
  MPI_Bcast(&kb, 1, MPI_INT, root, comm);
  MPI_Bcast(&N, 1, MPI_INT, root, comm);
  MPI_Bcast(&iters, 1, MPI_INT, root, comm);
  MPI_Bcast(&node_iters, 1, MPI_INT, root, comm);
  MPI_Bcast(&eps, 1, MPI_DOUBLE, root, comm);
  MPI_Bcast(&node_eps, 1, MPI_DOUBLE, root, comm);
  MPI_Bcast(&algorithm, 1, MPI_INT, root, comm);
  MPI_Bcast(&node_algorithm, 1, MPI_INT, root, comm);
  MPI_Bcast(&subopts, 256, MPI_CHAR, root, comm);

  // Load the kernel library
  if (!strcmp(path, "")) {
    fprintf(stderr, "Routine shared library must be specified\n");
    MPI_Abort(comm, 5);
  }
  fupermod_library library;
  fupermod_library_load(comm, path, NULL, &library);

  // pick the partitioning algorithm to use
  fupermod_partitioner partitioner = partitioner_switch(algorithm, smooth, benchsize);
  fupermod_partitioner node_partitioner = partitioner_switch(node_algorithm, smooth, benchsize);



  HybridConfig* hybridconf = (HybridConfig*)malloc(sizeof(HybridConfig));
  int rc;
  rc = hybrid_nodetype_read(nodetype_path, hybridconf);
  if (rc != 0){
    fprintf(stderr, "P%d: Error reading nodetype with path:%s\n", rank, nodetype_path);
    MPI_Abort(comm, 40);
  }
  fupermod_benchmark_hybrid* benchmark = fupermod_benchmark_hybrid_alloc(library.kernel, hybridconf, N, kb, node_iters, node_eps, node_partitioner);
  int N_blocked = N / kb;
  long long int D = N_blocked * N_blocked;
  int size;
  MPI_Comm_size(comm, &size);
  fupermod_data** data = NULL;
  fupermod_model** models = NULL;
  if (rank == root) {
    data = (fupermod_data**)malloc(sizeof(fupermod_data*) * size);
    models = (fupermod_model**)malloc(sizeof(fupermod_model*) * size);
    int i;
    for (i = 0; i < size; i++) {
      data[i] = fupermod_data_alloc();
      models[i] = partitioner.model_alloc(data[i], library.kernel->complexity, D, NULL, partitioner.arg);
    }
  }
  fupermod_dynamic partitioner_dyn = {partitioner.partition, size, models, fupermod_dist_alloc(D, size)};

  int iter = 0;
  int status = 0;
  // main loop
  fupermod_partitioner_iterate_one(&partitioner_dyn, comm, root, precision, (fupermod_benchmark*)benchmark, eps);
  while (iter < 50) {
    if (VERBOSE && rank == root) {
      fprintf(stderr, "** iter:%d of %d inter-node dist: ", iter, iters);
      int i;
      for (i = 0; i < size; i++) {
        fprintf(stderr, "%lld ", partitioner_dyn.dist->parts[i].d);
      }
      fprintf(stderr, "\n");
    }
    status = fupermod_partitioner_iterate(&partitioner_dyn, comm, root, precision, (fupermod_benchmark*)benchmark, eps);
    fprintf(stderr, "P%d Iter: %d of %d Status: %d\n", rank, iter, iters, status);

    iter++;
    if (iter >= iters && status == 0)
      break;
    if (iters == 2 && iter >= iters) //special case for CPM1 - break even if status isn't 0.
      break;
  }

  // Output the distribution
  fupermod_process_conf* confs; //todo: get conf for here. merge old hybridconf with new fupermod_process_conf.
  char** hostnames = fupermod_allgather_hostnames(comm);
  if (rank == root) {
    FILE* distout = fopen(dist_path, "w");
    mm_col_dist_write(distout, benchmark->mm_col_dist, partitioner_dyn.dist, confs);
    fclose(distout);
  }
  hybrid_nodedist_write(nodedist_path, hybridconf);

  // Print models
  if (rank == root && print_models) {
    int i;
    for (i = 0; i < size; i++) {
      char* filename = fupermod_model_filename("./", confs[i]);
      FILE* outfile = fopen(filename, "w");
      free(filename); filename = NULL;
      if (outfile == NULL)
        MPI_Abort(comm, -1);
      fupermod_print_model(outfile, &library, confs[i], models[i], D);
      fclose(outfile);
    }
  }
  if (rank == root && print_models_data) {
    FILE* file = fopen("points_all.dat", "w");
    fprintf(file, "#Kernel:%s %s, D:%lld, machines:", library.path, library.conf->subopts, D);
    int i;
    for (i = 0; i < size; i++) {
      fprintf(file, "%s ", hostnames[i]);
    }
    fprintf(file, "\n");
    fupermod_print_models_data(file, size, models);
    fclose(file);
  }

  if (rank == root) {
    fupermod_dist_free(partitioner_dyn.dist);
    int i;
    for (i = 0; i < size; i++) {
      partitioner.model_free(models[i]);
      fupermod_data_free(data[i]);
    }
    free(models);
    free(data);
  }
  fupermod_benchmark_hybrid_free(benchmark);
  free(hybridconf);
  if (rank == root) {
    fprintf(stderr, "\nPartitioing all done for N = %d !\n", N);
  }

  MPI_Finalize();
  return 0;
}
