/*!
 * \page partitioner Partitioner
 * Partitions 1D workloads.
 * Takes conf_file and total problem size as input. 
 * Reads in performance model data files. Using 1 of 7 algorithms partitions workload to balance load.
 * Outputs a text file <em>partition.dist</em> with hostname, device rank and distribution for later input into routine (or fupermod_1dto2d_mxm_dist).
 *
 * Algorithms:<ol start=0>
 * <li>Partial functional performance model</li>
 * <li>Full functional performance model</li>
 * <li>Constant performance model 1 (small kernel)</li>
 * <li>Constant performance model 2 (homogeneous kernel)</li>
 * <li>Homogeneous distribution</li>
 * <li>Manual distribution</li>
 * <li>Geometric Partitioning</li>
 * </ol>
 *
 */
#include "config.h"
#include "fupermod/fupermod_partition.h"
#include "fupermod/fupermod_conf.h"
#include <getopt.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

int main(int argc, char** argv) {

  char path[256]      = "";
  char subopts[256]   = "";
  char dist_path[256]     = "./partition.dist";
  char data_dir[256]      = "./";
  char conf_file[256]  = "./conf_file";
  int D            = 20;
  int smooth       = 0;
  int algorithm    = 0;
  int print_models = 0;

  int opt;
  while ((opt = getopt(argc, argv, "ha:l:o:D:s:m:f:cp:")) != -1)
    switch (opt) {
      case 'h':
        fprintf(
            stderr,
            "Partitions data, building functional performance models\n"
            "Usage: partitioner [options]\n"
            "-h	help\n"
            "-a I	Balancing algorithm. (default: %d)\n"
            "\t\t0: Homogeneous distribution\n"
            "\t\t1: Constant performance model 1 (small benchmark)\n"
            "\t\t2: Constant performance model 2 (homogeneous benchmark)\n"
            "\t\t3: Geometric algorithm FFPM\n"
            "\t\t4: FPM multiroot solver\n"
            "-l S shared library (required: a full path or relative to LD_LIBRARY_PATH)\n"
            "-o S suboptions (comma separated options for the shared library)\n"
            "-D I problem size to partition (required)\n"
            "-s I smoothing, number of points to average over(default: %d)\n"
            "-f S path to data file (default: %s)\n"
            "-m S path to machine file (default: %s)\n"
            "-p S path to output distribution (default: %s)\n"
            "-c   print out fitted models to file (default: No)\n"
            ,algorithm, smooth, data_dir, conf_file, dist_path 
            );
        return 0;
        break;
      case 'a':
        algorithm = atoi(optarg);
        break;
      case 'l':
        strcpy(path, optarg);
        break;
      case 'o':
        strcpy(subopts, optarg);
        break;
      case 'D':
        D = atoi(optarg);
        break;
      case 's':
        smooth = atoi(optarg);
        break;
      case 'f':
        strcpy(data_dir, optarg);
        break;
      case 'm':
        strcpy(conf_file, optarg);
        break;
      case 'p':
        strcpy(dist_path, optarg);
        break;
      case 'c':
        print_models = 1;
        break;
      default:
        fprintf(stderr, "Unknown option %s\n", optarg);
        break;
    }

  // Load the kernel library
  if (!strcmp(path, "")) {
    fprintf(stderr, "Routine shared library must be specified\n");
    return 1;
  }

  // Get host names
  int size; //total number of devices.
  fupermod_process_conf* confs = fupermod_get_conf_all(conf_file, &size);
  
  // assuming kb is the same for all processes
  fupermod_library library;
  if(fupermod_library_load(MPI_COMM_NULL, path, &confs[0], &library) != 0) {
    fprintf(stderr,"Error: opening shared library\n");
    return 1;
  }
  
  // Load data points
  fupermod_data** data = (fupermod_data**)malloc(sizeof(fupermod_data*) * size);
  int i;
  for (i = 0; i < size; i++) {
    char* filename = fupermod_data_filename(data_dir, confs[i]);

    FILE* data_file = fopen(filename, "r");
    if (data_file == NULL){
      fprintf(stderr,"Error: could not open: %s\n", filename);
      return 1;
    }
    fprintf(stderr,"Info: Reading data from: %s\n", filename);

    data[i] = fupermod_data_alloc();
    fupermod_data_read(data_file, data[i]);
    if (data[i]->count == 0){
      fprintf(stderr,"Error: failed to read any valid data from: %s\n",
          filename);
      return 1;
    }
    fclose(data_file);
    free(filename); filename = NULL;
  }

  // Allocate the partitioning algorithm and models
  fupermod_partition partition = NULL;
  fupermod_model** models = (fupermod_model**)malloc(sizeof(fupermod_model*) * size);
  fupermod_model_free model_free = NULL;
  switch (algorithm) {
    case 0: // Homogeneous distribution
      partition = NULL;
      break;
    case 1: // Constant performance model 1 (small benchmark)
    case 2: // Constant performance model 2 (homogeneous benchmark)
      {
        partition = fupermod_partition_basic;
        int bench_size = (algorithm == 1 ? 1e2 : D / size);
        int i;
        for (i = 0; i < size; i++)
          models[i] = fupermod_model_constant_alloc(data[i], library.kernel->complexity,
              D, NULL, bench_size);

        model_free = fupermod_model_constant_free;
      }
      break;
    case 3: // FPM Geometric partitioning
      {
        partition = fupermod_partition_geometric;
        int i;
        for (i = 0; i < size; i++)
          models[i] = fupermod_model_piecewise_alloc(data[i], library.kernel->complexity,
              D, NULL, smooth);

        model_free = fupermod_model_piecewise_free;
      }
      break;
    case 4: // FPM multiroot solver
      {
        partition = fupermod_partition_multiroot;
        int i;
        for (i = 0; i < size; i++)
          models[i] = fupermod_model_interp_alloc(data[i], library.kernel->complexity,
              D, NULL, smooth);

        model_free = fupermod_model_interp_free;
      }
      break;
  }

  // Print models
  if (print_models) { 
    for (i = 0; i < size; i++) {
      char* filename = fupermod_model_filename(data_dir, confs[i]);
      FILE* outfile = fopen(filename, "w");
      free(filename); filename = NULL;
      if (outfile == NULL)
        MPI_Abort(MPI_COMM_WORLD, -1);
      fupermod_print_model(outfile, &library, confs[i], models[i], D);
      fclose(outfile);
    }
  }

  // Do the partitioning
  fupermod_dist* dist = fupermod_dist_alloc(D, size);
  if (partition)
    partition(dist, size, models);

  // Test if the algorithm worked
  if (!fupermod_dist_test_total(dist)) {
    fprintf(stderr,
        "Error: Partitioning algorithm failed, dist does not sum to D\n");
    return 1;
  }

  // Output the distribution
  FILE* distout = fopen(dist_path, "w");
  if (distout == NULL) {
    fprintf(stderr, "Error, couldn't open distout for writing: %s\n", dist_path);
  } else {
    fupermod_dist_write(distout, dist, &library, confs);
    fclose(distout);
  }
  // Output to stderr
  fprintf(stderr, "Distribution: ");
  for (i = 0; i < size; i++) {
    fprintf(stderr, "%lld ", dist->parts[i].d);
  }
  fprintf(stderr, "\n");

  fupermod_dist_free(dist);
  for (i = 0; i < size; i++) 
    fupermod_conf_free(confs[i]);
  free(confs);

  // Deallocate models
  if (model_free)
    for (i = 0; i < size; i++)
      model_free(models[i]);
  free(models);

  // Deallocate data and hostnames
  for (i = 0; i < size; i++)
    fupermod_data_free(data[i]);
  free(data);

  // Unload the kernel library
  fupermod_library_unload(MPI_COMM_NULL, &library);

  return 0;
}
